package main

import (
	"context"
	"fmt"
	"net"
	"os"
	"os/signal"
	"syscall"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/generic/metrics"

	account "gitlab.mashroom.com/mashroom/backend/account/api"
	closecom "gitlab.mashroom.com/mashroom/backend/closecom/api"
	"gitlab.mashroom.com/mashroom/backend/common/api/health"
	"gitlab.mashroom.com/mashroom/backend/common/api/process"
	"gitlab.mashroom.com/mashroom/backend/common/generic/addr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/grpc"
	"gitlab.mashroom.com/mashroom/backend/common/generic/tracer"
	loggerfactory "gitlab.mashroom.com/mashroom/backend/common/lib/logger/factory"
	loggerservice "gitlab.mashroom.com/mashroom/backend/common/lib/logger/service"
	extras "gitlab.mashroom.com/mashroom/backend/extras/api"
	invitation "gitlab.mashroom.com/mashroom/backend/invitation/api"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	apirto "gitlab.mashroom.com/mashroom/backend/offer/api/rto"
	"gitlab.mashroom.com/mashroom/backend/offer/impl"
	rtorepository "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/repository/mongo"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/service"
	servicerto "gitlab.mashroom.com/mashroom/backend/offer/internal/app/service/rto"
	servicertodaemon "gitlab.mashroom.com/mashroom/backend/offer/internal/app/service/rto/daemon"
	transportgrpc "gitlab.mashroom.com/mashroom/backend/offer/internal/app/transport/grpc"
	usecaserto "gitlab.mashroom.com/mashroom/backend/offer/internal/app/usecase/rto"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/connections"
	_ "gitlab.mashroom.com/mashroom/backend/offer/internal/connections" // initialize connections
	"gitlab.mashroom.com/mashroom/backend/offer/model"
	"gitlab.mashroom.com/mashroom/backend/offer/version"
	schedule "gitlab.mashroom.com/mashroom/backend/schedule/api"

	"github.com/oklog/oklog/pkg/group"
	opentracing "github.com/opentracing/opentracing-go"
	"go.mongodb.org/mongo-driver/mongo"
	"go.uber.org/dig"
)

var (
	container   *dig.Container
	log         loggerservice.Logger
	err         error
	ctx         context.Context
	cancel      context.CancelFunc
	serviceName string
)

func init() { //nolint
	container = dig.New()

	serviceName = "offer"

	log = new(loggerfactory.ZapFactory).CreateLogger()
	err = container.Provide(func() loggerservice.Logger {
		return log
	})
	if err != nil {
		panic(err)
	}

	ctx, cancel = context.WithCancel(context.Background())

	opentracingTracer := tracer.GlobalTracer()
	err = container.Provide(func() opentracing.Tracer {
		return opentracingTracer
	})
	if err != nil {
		panic(err)
	}

	model.InitRepo()

	// todo: tracing, logs, metrics middlewares after all was ready and work
	err = container.Provide(func() *mongo.Database {
		return model.DB
	})
	err = container.Provide(rtorepository.NewRepository)
	if err != nil {
		panic(err)
	}
	err = container.Provide(rtorepository.NewHistoryRepository)
	if err != nil {
		panic(err)
	}

	err = container.Provide(func() extras.ExtrasServiceClient {
		return connections.ExtrasService
	})
	if err != nil {
		panic(err)
	}
	err = container.Provide(func() account.AccountServiceClient {
		return connections.AccountService
	})
	if err != nil {
		panic(err)
	}
	err = container.Provide(func() notification.NotificationServiceClient {
		return connections.NotificationService
	})
	if err != nil {
		panic(err)
	}
	err = container.Provide(func() listing.ListingServiceClient {
		return connections.ListingService
	})
	if err != nil {
		panic(err)
	}
	err = container.Provide(func() invoice.InvoiceServiceClient {
		return connections.InvoiceService
	})
	if err != nil {
		panic(err)
	}
	err = container.Provide(func() closecom.ClosecomServiceClient {
		return connections.ClosecomService
	})
	if err != nil {
		panic(err)
	}

	err = container.Provide(service.NewExtrasService)
	if err != nil {
		panic(err)
	}
	err = container.Provide(servicerto.NewSecurityService)
	if err != nil {
		panic(err)
	}
	err = container.Provide(servicerto.NewNotificationService)
	if err != nil {
		panic(err)
	}
	err = container.Provide(servicerto.NewCloseComService)
	if err != nil {
		panic(err)
	}

	err = container.Provide(servicertodaemon.NewFinishService)
	if err != nil {
		panic(err)
	}
	err = container.Provide(servicertodaemon.NewNotificationService)
	if err != nil {
		panic(err)
	}
	err = container.Provide(servicertodaemon.NewUpdateLeadTenantUserService)
	if err != nil {
		panic(err)
	}

	err = container.Provide(usecaserto.NewCreateUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewCreateByGuestUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewUpdateLandlordUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewCreateTenancyUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewUpdateStateUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewUpdateRentPlanUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewUpdateLeadTenantUserUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewCreateInvoiceUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewUpdateEndDateUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewRenewalUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewUpdateStateToWaitingForStandingOrderSetupUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewUpdateStateToWaitingForFirstRentPaymentUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewUpdateStateToActiveUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewUpdateStateToCancelledUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewUpdateStateToEndingUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewUpdateStateToFinishedUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewGetByPropertyIDUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewGetByPropertyIDsUseCase)
	if err != nil {
		panic(err)
	}
	err = container.Provide(usecaserto.NewAnswerFromTenantUseCase)
	if err != nil {
		panic(err)
	}

	err = container.Provide(transportgrpc.NewRtoServer)
	if err != nil {
		panic(err)
	}

	// todo: not used, in renewal
	err = container.Provide(usecaserto.NewUpdateStateToActiveRollingTenancyUseCase)
	if err != nil {
		panic(err)
	}
}

func main() {
	var transportGrpcRtoServer transportgrpc.RtoServer
	var rtoNotificationService servicertodaemon.NotificationService
	var updateStateToWaitingForFirstRentPaymentUseCase usecaserto.UpdateStateToWaitingForFirstRentPaymentUseCase
	var updateStateToActiveUseCase usecaserto.UpdateStateToActiveUseCase
	var updateStateToEndingUseCase usecaserto.UpdateStateToEndingUseCase
	var finishService servicertodaemon.FinishService
	var updateLeadTenantService servicertodaemon.UpdateLeadTenantService
	err = container.Invoke(func(rtoServer transportgrpc.RtoServer,
		rtoNotificationSrv servicertodaemon.NotificationService,
		UpdateStateToWaitingForFirstRentPaymentUseCase usecaserto.UpdateStateToWaitingForFirstRentPaymentUseCase,
		UpdateStateToActiveUseCase usecaserto.UpdateStateToActiveUseCase,
		UpdateStateToEndingUseCase usecaserto.UpdateStateToEndingUseCase,
		finishSrv servicertodaemon.FinishService,
		updateLeadTenantSrv servicertodaemon.UpdateLeadTenantService) {
		transportGrpcRtoServer = rtoServer
		rtoNotificationService = rtoNotificationSrv
		updateStateToWaitingForFirstRentPaymentUseCase = UpdateStateToWaitingForFirstRentPaymentUseCase
		updateStateToActiveUseCase = UpdateStateToActiveUseCase
		updateStateToEndingUseCase = UpdateStateToEndingUseCase
		finishService = finishSrv
		updateLeadTenantService = updateLeadTenantSrv
	})
	if err != nil {
		panic(err)
	}

	defer func() {
		_ = log.Sync()

		tracer.Close()
	}()

	baseGrpcServer, err := grpc.NewServer(serviceName)
	if err != nil {
		panic(err)
	}

	var g group.Group
	{
		// This function just sits and waits for ctrl-C.
		cancelInterrupt := make(chan struct{})
		g.Add(func() error {
			c := make(chan os.Signal, 1)
			signal.Notify(c, syscall.SIGINT, syscall.SIGTERM)
			select {
			case sig := <-c:
				return fmt.Errorf("received signal %s", sig)
			case <-cancelInterrupt:
				return nil
			}
		}, func(error) {
			close(cancelInterrupt)
			cancel()
		})
	}
	{
		server := impl.NewServer(ctx,
			updateStateToWaitingForFirstRentPaymentUseCase,
			updateStateToActiveUseCase,
			updateStateToEndingUseCase)

		apirto.RegisterRtoServiceServer(baseGrpcServer, transportGrpcRtoServer)
		offer.RegisterOfferServiceServer(baseGrpcServer, server)
		process.RegisterProcessMonitorServiceServer(baseGrpcServer, server)
		invoice.RegisterInvoiceNotifyServiceServer(baseGrpcServer, server)
		schedule.RegisterReminderConsumerServiceServer(baseGrpcServer, server)
		invitation.RegisterInvitationNotifyServiceServer(baseGrpcServer, server)

		err = server.RegisterSchedules(ctx)
		if err != nil {
			panic(err)
		}
		grpcListener, err := net.Listen("tcp", addr.Grpc)
		if err != nil {
			panic(fmt.Errorf("fatal error while init gRPC listener: %s", err))
		}

		health.RegisterHealthServiceServer(baseGrpcServer, version.NewHealthServer())

		_ = log.Log("transport", "grpc", "address", addr.Grpc, "msg", "listening "+serviceName+" api")

		err = baseGrpcServer.Serve(grpcListener)
		if err != nil {
			log.Error(err)
			_ = grpcListener.Close()
			return
		}
	}

	g.Add(func() error {
		err = metrics.ConfigureListenAndServe(baseGrpcServer)
		if err == nil {
			for ctx.Err() == nil {
				time.Sleep(time.Second)
			}

			return nil
		}

		return err
	}, func(err error) {
		cancel()
	})
	g.Add(func() error {
		return rtoNotificationService.RemindAll(ctx)
	}, func(err error) {
		cancel()
	})
	g.Add(func() error {
		return finishService.FinishAll(ctx)
	}, func(err error) {
		cancel()
	})
	g.Add(func() error {
		return updateLeadTenantService.UpdateAllLeadTenantUsers(ctx)
	}, func(err error) {
		cancel()
	})

	_ = log.Log("exit", g.Run())
}
