package impl

import (
	"context"
	"fmt"
	"strings"
	"time"

	"github.com/camunda-cloud/zeebe/clients/go/pkg/entities"
	account "gitlab.mashroom.com/mashroom/backend/account/api"
	closecom "gitlab.mashroom.com/mashroom/backend/closecom/api"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	"gitlab.mashroom.com/mashroom/backend/common/generic/bpms"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/common/generic/util"
	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/connections"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/schedule/oneoff"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/status"
	"gitlab.mashroom.com/mashroom/backend/offer/model"
	property "gitlab.mashroom.com/mashroom/backend/property/api"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

// Document rendering option for document service.
const (
	SignMargin           = 0
	SignLineHeight       = 54
	SignWidth            = 206
	SignHeight           = 30
	SignX                = 306
	SignPageAST          = 16
	SignPageASTNoDeposit = 17
	SignPageLandlord     = 29
	SignPageTenant       = 32
)

const completeStatus = "Complete"

func inferRisk(v string) offer.TenantRisk {
	if v == "" {
		return offer.TenantRisk_PENDING_RISK
	}

	low := strings.ToLower(v)
	switch {
	case strings.Contains(low, "low"):
		return offer.TenantRisk_LOW_RISK
	case strings.Contains(low, "high"):
		return offer.TenantRisk_HIGH_RISK
	default:
		return offer.TenantRisk_MEDIUM_RISK
	}
}

//nolint:gocognit,gocyclo // try to reduce cognitive and cyclomatic complexity
func (server *serverImpl) ServeBpms() {
	zbClient, err := bpms.ServeBpms("offer-service", bpms.MapHandler(map[string]bpms.Registrator{
		"set-status": func() (request interface{}, processor bpms.Processor) {
			req := &struct {
				ID     string           `json:"id"`
				Status offer.OfferState `json:"status"`
			}{}
			return req, func(
				ctx context.Context,
				job entities.Job,
			) (reply interface{}, enhance map[string]interface{}, err error) {
				id, err := primitive.ObjectIDFromHex(req.ID)
				if err != nil {
					return nil, nil, err
				}

				m, err := model.GetOfferByID(ctx, id)
				if err != nil {
					return nil, nil, err
				}

				err = status.SetOfferState(ctx, m, req.Status)

				return nil, nil, err
			}
		},
		"notify-applicants": func() (request interface{}, processor bpms.Processor) {
			req := &struct {
				ID       string `json:"id"`
				Body     string `json:"body"`
				Contract string `json:"contract"`
				Tenant   string `json:"tenant"`
			}{}
			return req, func(
				ctx context.Context,
				job entities.Job,
			) (reply interface{}, enhance map[string]interface{}, err error) {
				log.Error("Deprecated call to notify-applicants")

				return
			}
		},
		"create-contract": func() (request interface{}, processor bpms.Processor) {
			req := &struct {
				ID string `json:"id"`
			}{}
			return req, func(
				ctx context.Context,
				job entities.Job,
			) (reply interface{}, enhance map[string]interface{}, err error) {
				id, err := primitive.ObjectIDFromHex(req.ID)
				if err != nil {
					return nil, nil, err
				}
				m, err := model.GetOfferByID(ctx, id)
				if err != nil {
					return nil, nil, err
				}
				_, err = connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
					Id: m.Property,
				})
				if err != nil {
					return nil, nil, err
				}
				_, err = connections.AccountService.Get(ctx, &account.GetRequest{
					Id: m.Landlord.User,
				})
				if err != nil {
					return nil, nil, err
				}

				return map[string]interface{}{
					"contract": m.Contract,
				}, nil, nil
			}
		},
		"update-contract": func() (request interface{}, processor bpms.Processor) {
			req := &struct {
				Event struct {
					EventType string `json:"event_type"`
				} `json:"event"`
				SignatureRequest struct {
					Metadata struct {
						PropertyID string `json:"property_id"`
						ContractID string `json:"contract_id"`
						OfferID    string `json:"offer_id"`
						ViewingID  string `json:"viewing_id"`
					} `json:"metadata"`
					Signatures []struct {
						SignerEmailAddress string `json:"signer_email_address"`
						StatusCode         string `json:"status_code"`
						Error              string `json:"error"`
					} `json:"signatures"`
					IsComplete bool `json:"is_complete"`
					IsDeclined bool `json:"is_declined"`
					HasError   bool `json:"has_error"`
				} `json:"signature_request"`
			}{}
			return req, func(
				ctx context.Context,
				job entities.Job,
			) (reply interface{}, enhance map[string]interface{}, err error) {
				fmt.Printf("%+v\n", req)

				id, err := primitive.ObjectIDFromHex(req.SignatureRequest.Metadata.OfferID)
				if err != nil {
					return nil, nil, err
				}
				m, err := model.GetOfferByID(ctx, id)
				if err != nil {
					return nil, nil, err
				}

				if m.Test {
					ctx = context.WithValue(ctx, grpcctx.CtxIsTestUser, true)
				}

				bgctx := grpcctx.ValuesBackgroundContext(ctx)

				waserr := req.SignatureRequest.IsDeclined || req.SignatureRequest.HasError
				for _, s := range req.SignatureRequest.Signatures {
					trans, err := TranslateSignatureStatus(s.StatusCode)
					if err != nil {
						log.WithError(err).Errorf("invalid signature %+v", s)
						waserr = true
						break
					}
					if s.SignerEmailAddress == util.UnPtrString(m.Landlord.EmailPrimary) {
						if len(m.Landlord.Signature) == 0 ||
							m.Landlord.Signature[len(m.Landlord.Signature)-1].State != trans.State {
							m.Landlord.Signature = append(m.Landlord.Signature, trans)
						}
						if trans.State == offer.SignStatusState_SIGNATURE_PROVIDED {
							go func() {
								notifySignatureProvided(
									bgctx,
									notification.Role_ROLE_LANDLORD,
									m,
								)

								_ = oneoff.ScheduleInvocation(
									bgctx,
									time.Now().UTC().Add(2*time.Hour),
									oneoff.InvocationTypeSignatureProvided,
									&oneoff.InvocationSignatureProvided{
										OfferID: m.ID.Hex(),
										Actor:   notification.Role_ROLE_LANDLORD,
									},
								)
							}()
						}
					} else {
						for _, t := range m.Tenants {
							if s.SignerEmailAddress == util.UnPtrString(t.EmailPrimary) {
								if len(t.Signature) == 0 ||
									t.Signature[len(t.Signature)-1].State != trans.State {
									t.Signature = append(t.Signature, trans)
								}
								if trans.State == offer.SignStatusState_SIGNATURE_PROVIDED {
									go func() {
										notifySignatureProvided(
											bgctx,
											notification.Role_ROLE_TENANT,
											m,
										)

										_ = oneoff.ScheduleInvocation(
											bgctx,
											time.Now().UTC().Add(2*time.Hour),
											oneoff.InvocationTypeSignatureProvided,
											&oneoff.InvocationSignatureProvided{
												OfferID: m.ID.Hex(),
												Actor:   notification.Role_ROLE_TENANT,
											},
										)
									}()
								}
								break
							}
						}

						for _, g := range m.Guarantors {
							if s.SignerEmailAddress == util.UnPtrString(g.EmailPrimary) {
								if len(g.Signature) == 0 ||
									g.Signature[len(g.Signature)-1].State != trans.State {
									g.Signature = append(g.Signature, trans)
								}

								break
							}
						}
					}
				}

				if waserr {
					err := status.SetOfferState(ctx, m, offer.OfferState_ERROR)
					return nil, nil, err
				}

				// set SIGNATURE_PROVIDED status if it's not set
				if req.SignatureRequest.IsComplete {
					if len(m.Landlord.Signature) == 0 ||
						m.Landlord.Signature[len(m.Landlord.Signature)-1].State !=
							offer.SignStatusState_SIGNATURE_PROVIDED {
						m.Landlord.Signature = append(m.Landlord.Signature, &model.SignatureModel{
							State: offer.SignStatusState_SIGNATURE_PROVIDED,
							Time:  time.Now(),
						})
					}

					for _, t := range m.Tenants {
						if len(t.Signature) == 0 ||
							t.Signature[len(t.Signature)-1].State != offer.SignStatusState_SIGNATURE_PROVIDED {
							t.Signature = append(t.Signature, &model.SignatureModel{
								State: offer.SignStatusState_SIGNATURE_PROVIDED,
								Time:  time.Now(),
							})
						}
					}

					for _, g := range m.Guarantors {
						if len(g.Signature) == 0 ||
							g.Signature[len(g.Signature)-1].State != offer.SignStatusState_SIGNATURE_PROVIDED {
							g.Signature = append(g.Signature, &model.SignatureModel{
								State: offer.SignStatusState_SIGNATURE_PROVIDED,
								Time:  time.Now(),
							})
						}
					}
				}

				err = model.UpdateOffer(ctx, m)
				if err != nil {
					return nil, nil, err
				}

				if req.SignatureRequest.IsComplete {
					err = status.SetOfferState(ctx, m, offer.OfferState_SIGNED)
					if err != nil {
						return nil, nil, err
					}
				}

				return nil, nil, nil
			}
		},
		"execute-checks": func() (request interface{}, processor bpms.Processor) {
			req := &struct {
				ID      string   `json:"id"`
				Tenants []string `json:"tenants"`
			}{}
			return req, func(
				ctx context.Context,
				job entities.Job,
			) (reply interface{}, enhance map[string]interface{}, err error) {
				defer func() {
					// TODO: consider returning proper status
					err = nil
				}()

				fmt.Println("checks", req.ID, req.Tenants)
				id, err := primitive.ObjectIDFromHex(req.ID)
				fmt.Println(id, err)
				if err != nil {
					return nil, nil, err
				}
				m, err := model.GetOfferByID(ctx, id)
				fmt.Println(m, err)
				if err != nil {
					return nil, nil, err
				}

				if m.Test {
					ctx = context.WithValue(ctx, grpcctx.CtxIsTestUser, true)
				}

				defer func() {
					// TODO: consider adding OfferState_CHECKS_FAILED
					err = status.SetOfferState(ctx, m, offer.OfferState_AWAITING_CHECKS)

					fmt.Println(err)
				}()

				err = server.executeChecks(ctx, m, req.Tenants)
				if err == ErrNoUser {
					err = nil
				}

				return nil, nil, err
			}
		},
		"tenant-check-update": func() (request interface{}, processor bpms.Processor) {
			req := &struct {
				ID    string `json:"id"`
				Check string `json:"check"`
				Data  struct {
					RiskScore string `json:"risk_score"`
					PDFFileID string `json:"pdf_file_id"`
				} `json:"data"`
			}{}
			return req, func(
				ctx context.Context,
				job entities.Job,
			) (reply interface{}, enhance map[string]interface{}, err error) {
				fmt.Println(req.Data)
				id, err := primitive.ObjectIDFromHex(req.ID)
				if err != nil {
					return nil, nil, err
				}
				m, err := model.GetOfferByID(ctx, id)
				if err != nil {
					return nil, nil, err
				}

				for _, t := range m.Tenants {
					if t.Check == req.Check {
						t.Risk = inferRisk(req.Data.RiskScore)
						if req.Data.PDFFileID != "" {
							t.RiskDocument = util.PtrString(req.Data.PDFFileID)

							go createRiskDocument(
								context.Background(),
								m.ID.Hex(),
								m.LandlordUser(),
								t.User,
								util.JoinNonblankStrings(
									" ",
									util.UnPtrString(t.FirstName),
									util.UnPtrString(t.LastName),
								),
								util.UnPtrString(t.EmailPrimary),
								m.Property,
								req.Data.PDFFileID,
							)
						}
						t.CheckNeeded = false
					}
				}

				for _, g := range m.Guarantors {
					if g.Check == req.Check {
						g.Risk = inferRisk(req.Data.RiskScore)
						if req.Data.PDFFileID != "" {
							g.RiskDocument = util.PtrString(req.Data.PDFFileID)
						}
						g.CheckNeeded = false
					}
				}

				err = model.UpdateOffer(ctx, m)
				fmt.Println(err)
				if err != nil {
					return nil, nil, err
				}

				if !m.ActiveChecks() {
					status.ExportClosecomOpportunity(ctx, m, closecom.OfferStatus_OFFER_STATUS_CONFIRMING_REFERENCE)
				}

				return nil, nil, nil
			}
		},
	}))
	if err != nil {
		panic(err)
	}
	server.bpms = zbClient
	status.BPMS = zbClient
}

func notifySignatureProvided(ctx context.Context, actor notification.Role, off *model.OfferModel) {
	if off == nil {
		log.Errorf("unable to send signature invitation notification: nil offer")
		return
	}

	var recipientID, actorID string

	var target notification.Role

	switch actor {
	case notification.Role_ROLE_LANDLORD:
		actorID = off.LandlordUser()
		recipientID = off.TenantUser()
		target = notification.Role_ROLE_TENANT
	case notification.Role_ROLE_TENANT:
		actorID = off.TenantUser()
		recipientID = off.LandlordUser()
		target = notification.Role_ROLE_LANDLORD
	}

	_, err := connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Actor: &notification.Actor{
			User: actorID,
			Role: actor,
		},
		Entity: notification.Entity_ENTITY_OFFER_SIGNATURE_INVITATION,
		Action: notification.Action_ACTION_PENDING,
		Targets: []*notification.Target{
			{
				User: recipientID,
				Role: target,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: off.Property,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: off.ID.Hex(),
			},
		},
	})
	if err != nil {
		log.WithError(err).Errorf("failed to send signature invitation notification, offer %s", off.ID.Hex())
	}
}

func getSignPageTenant(isGuarantor bool) int32 {
	if isGuarantor {
		return SignPageTenant + 1
	}
	return SignPageTenant
}

func getSignPageLandlord(isGuarantor bool) int32 {
	if isGuarantor {
		return SignPageLandlord + 1
	}
	return SignPageLandlord
}

func getPageAST(isDeposit bool, isGuarantor bool) int32 {
	var page int32

	if isDeposit {
		page = SignPageAST
	} else {
		page = SignPageASTNoDeposit
	}

	if isGuarantor {
		page++
	}

	return page
}

// TranslateSignatureStatus creates new model based on signature status code value.
func TranslateSignatureStatus(code string) (*model.SignatureModel, error) {
	switch code {
	case "success", "signed":
		return &model.SignatureModel{
			State: offer.SignStatusState_SIGNATURE_PROVIDED,
			Time:  time.Now(),
		}, nil
	case "on_hold", "on_hold_by_requester", "awaiting_signature":
		return &model.SignatureModel{
			State: offer.SignStatusState_AWAITING_SIGNATURE,
			Time:  time.Now(),
		}, nil
	case "declined":
		return &model.SignatureModel{
			State: offer.SignStatusState_SIGNATURE_REFUSED,
			Time:  time.Now(),
		}, nil
	case "error_unknown", "error_file", "error_component_position", "error_text_tag":
		return nil, ErrSign
	default:
		return nil, ErrSign
	}
}
