package impl

import (
	"context"

	account "gitlab.mashroom.com/mashroom/backend/account/api"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	"gitlab.mashroom.com/mashroom/backend/common/generic/util"
	invitation "gitlab.mashroom.com/mashroom/backend/invitation/api"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/connections"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/status"
	"gitlab.mashroom.com/mashroom/backend/offer/model"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

func (server *serverImpl) Invite(ctx context.Context, request *offer.InviteRequest) (*offer.InviteResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer())
	if err != nil {
		return nil, err
	}
	old, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	found := false
	for _, t := range old.Tenants {
		if t.User == request.GetUser() {
			found = true
			break
		}
	}

	if !found {
		return nil, status.ErrNoAccess
	}

	if old.State != offer.OfferState_PENDING && old.State != offer.OfferState_ACTIVE {
		return nil, ErrNoAccessStatus
	}

	invitee, _ := connections.AccountService.GetByEmail(ctx, &account.GetByEmailRequest{
		Email: request.GetEmail(),
	})

	prop, err := connections.ListingService.VersionGet(grpcctx.WithServiceUser(ctx), &listing.VersionGetRequest{
		Spec: listing.VersionSpecListingID("", old.PropertyListing),
	})
	if err != nil {
		return nil, err
	}

	_, err = connections.InvitationService.InvitationCreate(ctx, &invitation.InvitationCreateRequest{
		HostId: request.GetUser(),
		Invitee: &invitation.User{
			UserId: invitee.GetAccount().GetIdValue(),
			Email:  request.GetEmail(),
			Phone:  invitee.GetAccount().GetContactValue().GetPhonePrimaryValue(),
			Name: util.JoinNonemptyStrings(
				" ",
				invitee.GetAccount().GetDetailsValue().GetFirstNameValue(),
				invitee.GetAccount().GetDetailsValue().GetLastNameValue(),
			),
		},
		Property: &invitation.Property{
			PropertyId: old.Property,
			VersionId:  prop.GetVersion().GetId(),
		},
		Type:           invitation.InvitationType_INVITATION_TYPE_RELATED_TO_TENANT,
		Offer:          old.ID.Hex(),
		NotifyCallback: connections.SelfRemote,
	})
	if err != nil {
		return nil, err
	}

	return &offer.InviteResponse{
		Status: offer.OfferStatus_SUCCESS,
	}, nil
}

func (server *serverImpl) InvitationChange(
	ctx context.Context,
	req *invitation.InvitationChangeRequest,
) (*invitation.InvitationChangeResponse, error) {
	switch req.GetInvitation().GetStatus() {
	case invitation.InvitationStatus_INVITATION_STATUS_ACCEPTED:
		if req.GetInvitation().GetType() == invitation.InvitationType_INVITATION_TYPE_RELATED_TO_TENANT {
			err := server.tenantInviteAccepted(ctx, req)
			if err != nil {
				return nil, err
			}
		}
	case invitation.InvitationStatus_INVITATION_STATUS_DECLINED,
		invitation.InvitationStatus_INVITATION_STATUS_CANCELED,
		invitation.InvitationStatus_INVITATION_STATUS_TIMEOUT:
	}

	return &invitation.InvitationChangeResponse{}, nil
}

func (server *serverImpl) tenantInviteAccepted(
	ctx context.Context,
	request *invitation.InvitationChangeRequest,
) error {
	id, err := primitive.ObjectIDFromHex(request.GetInvitation().GetOffer())
	if err != nil {
		return err
	}

	old, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return err
	}

	offerLog := model.NewOfferLog(ctx, old, request.GetInvitation().GetInvitee().GetUserId())

	defer func() {
		if err != nil {
			return
		}

		offerLog.LogDiff()
	}()

	acc, err := connections.AccountService.Get(ctx, &account.GetRequest{
		Id: request.GetInvitation().GetInvitee().GetUserId(),
	})
	if err != nil {
		return err
	}

	var ot *model.TenantOfferModel

	for _, t := range old.Tenants {
		if t.User == acc.GetAccount().GetIdValue() ||
			util.UnPtrString(t.EmailPrimary) == acc.GetAccount().GetContactValue().GetEmailPrimaryValue() {
			ot = t

			break
		}
	}

	if ot == nil {
		ot = &model.TenantOfferModel{}

		old.Tenants = append(old.Tenants, ot)
	}

	ot.User = acc.GetAccount().GetIdValue()
	ot.FirstName = util.PtrString(acc.GetAccount().GetDetailsValue().GetFirstNameValue())
	ot.LastName = util.PtrString(acc.GetAccount().GetDetailsValue().GetLastNameValue())
	ot.EmailPrimary = util.PtrString(acc.GetAccount().GetContactValue().GetEmailPrimaryValue())
	ot.PhonePrimary = util.PtrString(acc.GetAccount().GetContactValue().GetPhonePrimaryValue())

	err = model.UpdateOffer(ctx, old)
	if err != nil {
		return err
	}

	err = status.SetOfferState(ctx, old, offer.OfferState_PENDING)
	if err != nil {
		return err
	}

	return nil
}
