package impl

import (
	"context"
	"encoding/json"
	"fmt"
	"strconv"
	"time"

	account "gitlab.mashroom.com/mashroom/backend/account/api"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	"gitlab.mashroom.com/mashroom/backend/common/generic/bind"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/common/generic/util"
	extras "gitlab.mashroom.com/mashroom/backend/extras/api"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/connections"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/invoices"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/status"
	"gitlab.mashroom.com/mashroom/backend/offer/model"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

func (server *serverImpl) InvoiceChange(
	ctx context.Context,
	req *invoice.InvoiceChangeRequest,
) (resp *invoice.InvoiceChangeResponse, err error) {
	if req.GetInvoice().GetCorrelationType() != invoice.CorrelationTypeOffer {
		return &invoice.InvoiceChangeResponse{}, nil
	}

	inv := req.GetInvoice()
	pmt := req.GetPayment()

	offerID, err := primitive.ObjectIDFromHex(inv.GetCorrelationId())
	if err != nil {
		return nil, err
	}

	// todo: for rto
	off, err := model.GetOfferByID(ctx, offerID)
	if err != nil {
		return nil, err
	}

	if off.Test {
		ctx = context.WithValue(ctx, grpcctx.CtxIsTestUser, true)
	}

	err = server.paymentPeriodicSetup(ctx, pmt, inv, off)
	if err != nil {
		return nil, err
	}

	if invoices.HoldingDepositPayment(pmt) {
		notifyHoldingDepositPaid(ctx, pmt, inv, off)
	}

	switch pmt.Status {
	case invoice.PaymentStatus_PAYMENT_CONFIRM:
		err = server.paymentConfirm(ctx, pmt, inv, off)
	case invoice.PaymentStatus_PAYMENT_SUCCESS:
		err = server.paymentSuccess(ctx, pmt, inv, off)
	}

	if err != nil {
		return nil, err
	}

	return &invoice.InvoiceChangeResponse{}, nil
}

func (server *serverImpl) paymentPeriodicSetup(
	ctx context.Context,
	pmt *invoice.Payment,
	inv *invoice.Invoice,
	off *model.OfferModel,
) (err error) {
	switch pmt.Status {
	case invoice.PaymentStatus_PAYMENT_PENDING, invoice.PaymentStatus_PAYMENT_CONFIRM,
		invoice.PaymentStatus_PAYMENT_SUCCESS:
	default:
		return nil
	}

	var update bool

	if inv.GetSetup() && !off.PaidStatuses.StandingOrderSetup &&
		inv.GetCorrelationTarget() == invoice.CorrelationTargetStandingOrder {
		off.PaidStatuses.StandingOrderSetup = true
		update = true

		go func() {
			_, _ = server.updateStateToWaitingForFirstRentPaymentUseCase.UpdateStateToWaitingForFirstRentPayment(ctx,
				&request.UpdateStateToWaitingForFirstRentPaymentRequest{ID: inv.CorrelationId})
		}()
	}

	if int(pmt.Serial+1) == int(*off.Conditions.TermMonths) &&
		pmt.Status == invoice.PaymentStatus_PAYMENT_SUCCESS &&
		inv.GetCorrelationTarget() == invoice.CorrelationTargetStandingOrder {
		go func() {
			_, _ = server.updateStateToEndingUseCase.UpdateStateToEnding(ctx,
				&request.UpdateStateToEndingRequest{ID: inv.CorrelationId})
		}()
	}

	if int(pmt.Serial+1) == int(*off.Conditions.TermMonths) &&
		pmt.Status == invoice.PaymentStatus_PAYMENT_SUCCESS &&
		inv.GetCorrelationTarget() == invoice.CorrelationTargetStandingOrder {
		go func() {
			_, _ = server.updateStateToEndingUseCase.UpdateStateToEnding(ctx,
				&request.UpdateStateToEndingRequest{ID: inv.CorrelationId})
		}()
	}

	if !update {
		return nil
	}

	err = model.UpdateOffer(ctx, off)
	if err != nil {
		log.Error(err)
		return err
	}

	if off.State != offer.OfferState_RENTPAID {
		return status.SetPropertyRented(ctx, off)
	}

	return status.GenerateSchedule(ctx, off)
}

func (server *serverImpl) paymentSuccess(
	ctx context.Context,
	pmt *invoice.Payment,
	inv *invoice.Invoice,
	off *model.OfferModel,
) (err error) {
	offerLog := model.NewOfferLog(ctx, off, inv.GetUser())

	defer func() {
		if err != nil {
			return
		}

		offerLog.LogDiff()
	}()

	var tenants []string

	// handle referencing check payment
	for _, item := range inv.GetServices() {
		if item.GetService().GetType() != extras.ServiceType_REFERENCING_CHECK ||
			item.GetService().GetTargetType() != "user-reference" {
			continue
		}

		for _, t := range off.Tenants {
			if util.UnPtrString(t.EmailPrimary) == item.GetService().GetTargetId() {
				t.CheckNeeded = true

				break
			}
		}

		for _, g := range off.Guarantors {
			if util.UnPtrString(g.EmailPrimary) == item.GetService().GetTargetId() {
				g.CheckNeeded = true

				break
			}
		}

		tenants = append(tenants, item.GetService().GetTargetId())
	}

	if len(tenants) > 0 {
		if err = model.UpdateOffer(ctx, off); err != nil {
			log.Error(err)
			return err
		}

		if err = server.applicantCheckPaymentPaid(ctx, off, tenants); err != nil {
			return err
		}
	}

	switch off.State {
	case offer.OfferState_HOLDING_DEPOSIT, offer.OfferState_AWAITING_CONFIRM:
		err = server.paymentSuccessHoldingConfirm(ctx, inv, off)
	case offer.OfferState_SIGNED, offer.OfferState_AWAITING_CONFIRM_RENT,
		offer.OfferState_AWAITING_TENANCY, offer.OfferState_RENTPAID:
		err = server.paymentSuccessSignedConfirm(ctx, inv, off)
	case offer.OfferState_AWAITING_CONFIRM_EXTRAS:
		err = server.paymentSuccessConfirmExtras(ctx, inv, off)
	}

	if err != nil {
		log.Error(err)
	}

	if _, err = connections.ExpensesInvoiceNotificationService.InvoiceChange(ctx, &invoice.InvoiceChangeRequest{
		Invoice: inv,
		Payment: pmt,
	}); err != nil {
		log.Error(err)
	}

	return
}

func notifyHoldingDepositPaid(
	ctx context.Context,
	pmt *invoice.Payment,
	inv *invoice.Invoice,
	off *model.OfferModel,
) {
	if !invoices.HoldingDepositPayment(pmt) {
		return
	}

	logger := log.WithFields(log.Fields{
		"action":      "notifyHoldingDepositPayed",
		"payment-ref": pmt.Reference,
		"invoice":     inv.InvoiceId,
		"offer":       off.ID.Hex(),
	})

	notify := func(action notification.Action) bool {
		_, err := connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_OFFER_HOLDING_DEPOSIT_PAYMENT,
			Action: action,
			Targets: []*notification.Target{
				{
					User: off.LandlordUser(),
					Role: notification.Role_ROLE_LANDLORD,
				},
				{
					User: off.TenantUser(),
					Role: notification.Role_ROLE_TENANT,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: off.Property,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
					EmbedValue: off.ID.Hex(),
				},
			},
		})
		if err != nil {
			logger.WithError(err).Errorf("notification delivery failed for actio: %s", action.String())
			return false
		}
		return true
	}

	if off.Notifications == nil {
		off.Notifications = make(model.OfferNotifications)
	}

	upd := false
	switch pmt.GetStatus() {
	case invoice.PaymentStatus_PAYMENT_CONFIRM:
		if pmt.GetSystem().GetBank() != nil && !off.Notifications.Sent(model.NotificationDepositBankPayed) {
			if notify(notification.Action_ACTION_CREATE) {
				off.Notifications[model.NotificationDepositBankPayed] = time.Now()
				upd = true
			}
		}
	case invoice.PaymentStatus_PAYMENT_SUCCESS:
		switch {
		case pmt.GetSystem().GetTokenio() != nil:
			notify(notification.Action_ACTION_COMPLETE)
		case pmt.GetSystem().GetBank() != nil && !off.Notifications.Sent(model.NotificationDepositBankVerified):
			if notify(notification.Action_ACTION_VERIFY) {
				off.Notifications[model.NotificationDepositBankVerified] = time.Now()
				upd = true
			}
		}
	}

	if upd {
		err := model.UpdateOffer(ctx, off)
		if err != nil {
			log.Error(err)
		}
	}
}

func (server *serverImpl) paymentSuccessConfirmExtras(
	ctx context.Context,
	inv *invoice.Invoice,
	old *model.OfferModel,
) error {
	user, err := connections.AccountService.Get(ctx, &account.GetRequest{
		Id:    inv.User,
		Admin: true,
	})
	if err != nil {
		log.WithError(err).Error("get user account failed")

		return nil
	}

	_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_EXTRA_PAYMENT,
		Action: notification.Action_ACTION_COMPLETE,
		Targets: []*notification.Target{
			{
				User: user.GetAccount().GetIdValue(),
				Role: notification.Role_ROLE_LANDLORD,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: old.ID.Hex(),
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: old.Property,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_FILE,
				EmbedValue: inv.GetRender(),
			},
		},
	})
	if err != nil {
		log.WithError(err).Errorf("sending failed: %s", notification.EventType_EVENT_PAYMENT_PAID_EXTRA_LANDLORD)

		return nil
	}

	return nil
}

func (server *serverImpl) paymentSuccessSignedConfirm(
	ctx context.Context,
	inv *invoice.Invoice,
	old *model.OfferModel,
) error {
	if len(old.Tenants) == 0 && old.Tenants[0].User != inv.User {
		return nil
	}

	if old.PaidStatuses == nil {
		old.PaidStatuses = &model.PaidStatusesModel{}
	}

	if inv.Status == invoice.InvoiceStatus_INVOICE_FULFILLED {
		switch inv.CorrelationTarget {
		case invoice.CorrelationTargetTenantServices:
			old.PaidStatuses.TenantServices = true

			err := model.UpdateOffer(ctx, old)
			if err != nil {
				return err
			}

			go func() {
				_, _ = server.updateStateToActiveUseCase.UpdateStateToActive(ctx,
					&request.UpdateStateToActiveRequest{ID: old.ID.Hex()})
			}()

			go func(ctx context.Context, m *model.OfferModel) {
				_ = status.TrySetPropertyRentedCreateInsurances(grpcctx.ValuesBackgroundContext(ctx), m, false)
			}(ctx, old)

			status.TryExportClosecomMoveinStatus(ctx, old)
		case invoice.CorrelationTargetDepositReplacement:
			old.PaidStatuses.DepositReplacementSetup = true

			err := model.UpdateOffer(ctx, old)
			if err != nil {
				return err
			}
		}
	}

	if old.State != offer.OfferState_RENTPAID {
		return status.SetPropertyRented(ctx, old)
	}

	return status.GenerateSchedule(ctx, old)
}

func (server *serverImpl) paymentSuccessHoldingConfirm(
	ctx context.Context,
	inv *invoice.Invoice,
	old *model.OfferModel,
) (err error) {
	fmt.Println("processReceipt offer-deposited start")

	logger := log.WithFields(log.Fields{
		"invoice": inv.InvoiceId,
		"offer":   old.ID.Hex(),
	})

	err = old.SetInvoiceAndUpdate(ctx)
	if err != nil {
		logger.WithError(err).Error("updating offer with invoice")

		return err
	}

	bgctx := grpcctx.ValuesBackgroundContext(ctx)

	_, err = server.bpms.NewPublishMessageCommand().
		MessageName("deposit_paid").
		CorrelationKey(old.ID.Hex()).
		TimeToLive(status.BPMTimeToLive).
		Send(bgctx)
	if err != nil {
		logger.WithError(err).Error("NewPublishMessageCommand error deposit_paid")

		return err
	}

	_, err = server.bpms.NewPublishMessageCommand().
		MessageName("offer-deposited").
		CorrelationKey(old.ID.Hex()).
		TimeToLive(status.BPMTimeToLive).
		Send(bgctx)
	if err != nil {
		logger.WithError(err).Error("NewPublishMessageCommand error offer-deposited")

		return err
	}

	err = status.SetOfferState(ctx, old, offer.OfferState_DEPOSIT_CONFIRMED)
	if err != nil {
		logger.WithError(err).Error("set offer state DEPOSIT_CONFIRMED")

		return err
	}

	if old.DepositReplacement() {
		_, err = server.bpms.NewPublishMessageCommand().
			MessageName("offer-run-checks").
			CorrelationKey(old.ID.Hex()).
			TimeToLive(status.BPMTimeToLive).
			Send(bgctx)
		if err != nil {
			logger.WithError(err).Error("NewPublishMessageCommand error offer-run-checks")

			return err
		}

		err = status.SetOfferState(ctx, old, offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT)
		if err != nil {
			logger.WithError(err).Error("set offer state AWAITING_APPLICANT_CHECK_PAYMENT")

			return err
		}
	}

	err = model.UpdateOffer(ctx, old)
	if err != nil {
		logger.WithError(err).Errorf("updating offer")

		return err
	}

	return nil
}

func (server *serverImpl) paymentConfirm(
	ctx context.Context,
	pmt *invoice.Payment,
	inv *invoice.Invoice,
	off *model.OfferModel,
) error {
	accResp, err := connections.AccountService.Get(ctx, &account.GetRequest{
		Id: inv.User,
	})
	if err != nil {
		return err
	}

	ord, _ := connections.InvoiceService.OrderLoad(ctx, &invoice.OrderLoadRequest{
		Id: inv.GetOrderId(),
	})

	for _, s := range inv.Services {
		var proddata string

		for _, p := range ord.GetOrder().GetProducts() {
			if p.GetServiceId() == s.GetService().GetServiceId() {
				bs, _ := bind.ToJSON(p.GetDetails(), nil)
				proddata = string(bs)

				break
			}
		}

		bs, _ := json.Marshal(map[string]interface{}{
			"product": map[string]interface{}{
				"details_raw": proddata,
			},
		})

		proddata = string(bs)

		_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_OFFER_EXTRA_PAYMENT,
			Action: notification.Action_ACTION_COMPLETE,
			Targets: []*notification.Target{
				{
					Role: notification.Role_ROLE_ADMIN,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: off.Property,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_JSON,
					EmbedValue: proddata,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
					EmbedValue: accResp.GetAccount().GetIdValue(),
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PAYMENT_COST,
					EmbedValue: strconv.FormatFloat(s.GetService().GetPrice(), 'f', 2, 64),
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PAYMENT_TIME,
					EmbedValue: strconv.FormatInt(pmt.GetCreatedAt().GetSeconds(), 10),
				},
			},
		})
		if err != nil {
			log.Error(err)

			return err
		}
	}

	return nil
}

func (server *serverImpl) applicantCheckPaymentPaid(
	ctx context.Context,
	old *model.OfferModel,
	tenants []string,
) error {
	if old.State == offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT {
		err := status.SetOfferState(ctx, old, offer.OfferState_CHECKING)
		if err != nil {
			return err
		}

		bgctx := grpcctx.ValuesBackgroundContext(ctx)

		msg, err := server.bpms.NewPublishMessageCommand().
			MessageName("referencing-checks-paid").
			CorrelationKey(old.ID.Hex()).
			TimeToLive(status.BPMTimeToLive).
			VariablesFromMap(map[string]interface{}{
				"tenants": tenants,
			})
		if err != nil {
			log.Error(err)

			return err
		}

		_, err = msg.Send(bgctx)
		if err != nil {
			log.Error(err)

			return err
		}
	}

	return nil
}
