package impl

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"strconv"
	"strings"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/api/generic/query"

	expenses "gitlab.mashroom.com/mashroom/backend/expenses/api"

	"google.golang.org/protobuf/proto"

	account "gitlab.mashroom.com/mashroom/backend/account/api"
	cardpayment "gitlab.mashroom.com/mashroom/backend/cardpayment/api"
	closecom "gitlab.mashroom.com/mashroom/backend/closecom/api"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/bind"
	"gitlab.mashroom.com/mashroom/backend/common/generic/env"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/common/generic/unwrap"
	"gitlab.mashroom.com/mashroom/backend/common/generic/util"
	document "gitlab.mashroom.com/mashroom/backend/document/api"
	extras "gitlab.mashroom.com/mashroom/backend/extras/api"
	identity "gitlab.mashroom.com/mashroom/backend/identity/api"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/connections"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/invoices"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/mixpanel"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/status"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/synchronization"
	"gitlab.mashroom.com/mashroom/backend/offer/model"
	property "gitlab.mashroom.com/mashroom/backend/property/api"
	signature "gitlab.mashroom.com/mashroom/backend/signature/api"
	tenancy "gitlab.mashroom.com/mashroom/backend/tenancy/api"
	tokenio "gitlab.mashroom.com/mashroom/backend/tokenio/api"
	viewing "gitlab.mashroom.com/mashroom/backend/viewing/api"
	zeebe "gitlab.mashroom.com/mashroom/backend/zeebe/api"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
	"google.golang.org/protobuf/types/known/structpb"
	"google.golang.org/protobuf/types/known/timestamppb"
	"google.golang.org/protobuf/types/known/wrapperspb"
)

// SanctionsCheckFailed - is ...
const SanctionsCheckFailed = "Ineligible: Background check failed"

// IsProduction bool flag describes the service is on production side or not
var IsProduction = env.ResolveBool("IS_PRODUCTION", false)

func (server *serverImpl) Get(ctx context.Context, request *offer.GetRequest) (resp *offer.GetResponse, err error) {
	id, err := primitive.ObjectIDFromHex(request.GetId())
	if err != nil {
		return nil, err
	}
	result, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	var verif *offer.OfferVerification
	if request.Verify { // get verification
		verif = model.MakeVerification()
		valid := false
		model.VerifyAll(&valid, result, verif)
	}

	return &offer.GetResponse{
		Offer:        result.RenderOfferModel(ctx, request.GetUser()),
		Verification: verif,
	}, nil
}

func (server *serverImpl) ServicePropertyList(
	ctx context.Context,
	request *offer.ListRequest,
) (*offer.ListResponse, error) {
	models, err := model.ListOpenOffersByProperty(ctx, request.GetPropertyValue(), "")
	if err != nil {
		return nil, err
	}
	var offers []*offer.Offer
	for _, m := range models {
		if len(m.Tenants) == 0 {
			continue
		}
		render := m.RenderOfferModel(ctx, request.GetUser())

		offers = append(offers, render)
	}

	return &offer.ListResponse{Offers: offers}, nil
}

func (server *serverImpl) ServiceGet(ctx context.Context, request *offer.GetRequest) (*offer.GetResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetId())
	if err != nil {
		return nil, err
	}
	result, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return nil, err
	}

	return &offer.GetResponse{Offer: result.RenderOfferModel(ctx, request.GetUser())}, nil
}

func (server *serverImpl) ListByProperty(
	ctx context.Context,
	request *offer.ListRequest,
) (*offer.ListbyPropertyResponse, error) {
	if request.GetProperty() == nil {
		return nil, ErrNoProperty
	}

	models, err := model.ListByProperty(ctx, request.GetPropertyValue(), request.GetListing())
	if err != nil {
		return nil, err
	}
	var offers []*offer.Offer
	for _, m := range models {
		offers = append(offers, m.RenderOfferModel(ctx, request.GetUser()))
	}
	return &offer.ListbyPropertyResponse{Offers: offers}, nil
}

func (server *serverImpl) ListByPropertyTenant(
	ctx context.Context,
	request *offer.ListRequest,
) (*offer.ListbyPropertyResponse, error) {
	if request.GetProperty() == nil || request.GetUser() == "" {
		return nil, ErrNoProperty
	}
	models, err := model.ListByPropertyTenant(ctx, request.GetPropertyValue(), request.GetUser(), request.GetListing())
	if err != nil {
		return nil, err
	}
	var offers []*offer.Offer
	for _, m := range models {
		offers = append(offers, m.RenderOfferModel(ctx, request.GetUser()))
	}
	return &offer.ListbyPropertyResponse{Offers: offers}, nil
}

func (server *serverImpl) ListByPropertyLandlord(
	ctx context.Context,
	request *offer.ListRequest,
) (*offer.ListbyPropertyResponse, error) {
	if request.GetProperty() == nil || request.GetUser() == "" {
		return nil, ErrNoProperty
	}

	states := make([]offer.OfferState, 0)

	for _, s := range request.GetStatesValue().GetValue() {
		states = append(states, s)
	}

	models, err := model.ListByPropertyLandlord(ctx, request.GetPropertyValue(), request.GetListing(), states)
	if err != nil {
		return nil, err
	}
	var offers []*offer.Offer
	for _, m := range models {
		offers = append(offers, m.RenderOfferModel(ctx, request.GetUser()))
	}
	return &offer.ListbyPropertyResponse{Offers: offers}, nil
}

func (server *serverImpl) ListOutgoing(ctx context.Context, request *offer.ListRequest) (*offer.ListResponse, error) {
	states := []offer.OfferState{
		offer.OfferState_PENDING,
		offer.OfferState_ACTIVE,
		offer.OfferState_SIGNING,
	}
	if request.GetStates() != nil {
		states = request.GetStatesValue().GetValue()
	}
	models, err := model.ListOutgoing(ctx, request.GetUser(), request.GetPropertyValue(), states)
	if err != nil {
		return nil, err
	}
	var offers []*offer.Offer
	for _, m := range models {
		offers = append(offers, m.RenderOfferModel(ctx, request.GetUser()))
	}
	return &offer.ListResponse{Offers: offers}, nil
}

func (server *serverImpl) ListIncoming(ctx context.Context, request *offer.ListRequest) (*offer.ListResponse, error) {
	states := []offer.OfferState{
		offer.OfferState_ACTIVE,
		offer.OfferState_SIGNING,
	}
	if request.GetStates() != nil {
		states = request.GetStatesValue().GetValue()
	}
	models, err := model.ListIncoming(ctx, request.GetUser(), request.GetPropertyValue(), request.GetListing(), states)
	if err != nil {
		return nil, err
	}
	var offers []*offer.Offer
	for _, m := range models {
		offers = append(offers, m.RenderOfferModel(ctx, request.GetUser()))
	}
	return &offer.ListResponse{Offers: offers}, nil
}

//nolint:gocognit // try to reduce high cognitive complexity
func (server *serverImpl) Create(ctx context.Context, request *offer.CreateRequest) (*offer.CreateResponse, error) {
	prop, err := connections.ListingService.VersionGet(
		grpcctx.WithServiceUser(ctx),
		&listing.VersionGetRequest{
			Spec: listing.VersionSpecPropertyID("", request.GetOffer().GetProperty()),
		},
	)
	if err != nil {
		return nil, ErrNoProperty
	}

	acc, err := connections.AccountService.Get(ctx, &account.GetRequest{Id: request.GetUser()})
	if err != nil {
		log.WithError(err).WithField("user", request.GetUser()).Error("on getting account")
		return nil, ErrUnknownUser
	}

	if !acc.GetAccount().GetContactValue().GetContactStatusValue().GetPhoneVerifiedValue() {
		return nil, ErrPhoneNotVerified
	}

	if prop.GetVersion().GetTenancyId() != "" || prop.GetVersion().GetStatus() == listing.VersionStatus_VERSION_LET {
		return nil, ErrPropRentedOrOffered
	}

	// check an active offer of the tenant on this property
	openOffers, err := model.ListOpenOffersByProperty(ctx, prop.GetVersion().GetPropertyId(), request.GetUser())
	if err != nil && !errors.Is(err, mongo.ErrNoDocuments) {
		return nil, err
	}
	if len(openOffers) > 0 {
		return nil, ErrHaveActiveOffer
	}
	log.Printf("no active offers on property: %s by user: %s", prop.GetVersion().GetPropertyId(), request.User)

	models, err := model.ListOpenOffersByProperty(ctx, request.GetOffer().GetProperty(), "")
	if err == nil {
		for _, m := range models {
			if len(m.Tenants) == 0 {
				continue
			}
			invs, err := connections.InvoiceService.InvoiceList(ctx, &invoice.InvoiceListRequest{
				CorrelationType: invoice.CorrelationTypeOffer,
				CorrelationId:   m.ID.Hex(),
				Statuses: []invoice.InvoiceStatus{
					invoice.InvoiceStatus_INVOICE_OPEN,
					invoice.InvoiceStatus_INVOICE_FULFILLED,
				},
			})
			if err == nil && len(invs.GetInvoices()) > 0 {
				// if offer has holding deposit status and higher
				return nil, grpcerr.New(grpcerr.InvalidState, "property is under offer already")
			}
		}
	}

	landlord, err := connections.AccountService.Get(ctx, &account.GetRequest{
		Id: prop.GetVersion().GetPeople().GetLandlord().GetValue(),
	})
	if err != nil {
		return nil, err
	}

	details := &offer.OfferDetails{}
	if request.GetOffer().GetDetails() != nil {
		details = request.GetOffer().GetDetails()
	}

	m := &model.OfferModel{
		ID:              primitive.NewObjectID(),
		Property:        request.GetOffer().GetProperty(),
		PropertyIvr:     prop.GetVersion().GetIvr(),
		PropertyListing: prop.GetVersion().GetListingId(),
		PropertyAddress: prop.GetVersion().GetLocation().FormatAddress(),
		Viewing:         request.GetOffer().GetViewing(),
		CreatedAt:       time.Now(),
		UpdatedAt:       time.Now(),
		State:           offer.OfferState_PENDING,
		Landlord: &model.LandlordModel{
			User:         landlord.GetAccount().GetIdValue(),
			FirstName:    util.PtrString(landlord.GetAccount().GetDetailsValue().GetFirstNameValue()),
			LastName:     util.PtrString(landlord.GetAccount().GetDetailsValue().GetLastNameValue()),
			EmailPrimary: util.PtrString(landlord.GetAccount().GetContactValue().GetEmailPrimaryValue()),
			PhonePrimary: util.PtrString(landlord.GetAccount().GetContactValue().GetPhonePrimaryValue()),
			BillingAddress: util.PtrString(model.GetBillingAddressFromLocation(
				landlord.GetAccount().GetLocationValue(),
			)),
		},
		Manager: prop.GetVersion().GetPeople().GetManager().GetValue(),
		Details: &model.OfferDetailsModel{
			MovinAlone:     unwrap.BoolValue(details.GetMoveinAlone()),
			AdultFlatmates: unwrap.BoolValue(details.GetAdultFlatmates()),
		},
	}

	if acc.GetAccount().GetTest() != nil {
		m.Test = acc.GetAccount().GetTestValue()
	}

	err = m.UpdateOfferModel(ctx, request.GetOffer())
	if err != nil {
		return nil, err
	}

	found := false
	for _, v := range m.Tenants {
		if v.User == request.GetUser() {
			found = true
			break
		}
	}
	if !found {
		acc, err := connections.AccountService.Get(ctx, &account.GetRequest{
			Id: request.GetUser(),
		})
		if err != nil {
			return nil, err
		}

		ownTenant := &model.TenantOfferModel{
			User:           acc.GetAccount().GetIdValue(),
			FirstName:      util.PtrString(acc.GetAccount().GetDetailsValue().GetFirstNameValue()),
			LastName:       util.PtrString(acc.GetAccount().GetDetailsValue().GetLastNameValue()),
			EmailPrimary:   util.PtrString(acc.GetAccount().GetContactValue().GetEmailPrimaryValue()),
			PhonePrimary:   util.PtrString(acc.GetAccount().GetContactValue().GetPhonePrimaryValue()),
			BillingAddress: util.PtrString(model.GetBillingAddressFromLocation(acc.GetAccount().GetLocationValue())),
		}

		m.Tenants = append([]*model.TenantOfferModel{ownTenant}, m.Tenants...)
	}

	err = model.CreateOffer(ctx, m)
	if err != nil {
		return nil, err
	}

	synchronization.ServiceOfferSync(ctx, m)

	defer func() {
		if err != nil {
			fail := model.RemoveOffer(ctx, m.ID)
			if fail != nil {
				err = fmt.Errorf("%w: failed to remove offer after %s: %s", status.ErrInvalidState, err, fail)
			}
		}
	}()

	if request.GetOffer().GetViewing() != "" {
		_, err = connections.ViewingService.ServiceViewingOffer(
			grpcctx.WithServiceUser(ctx),
			&viewing.ServiceViewingOfferRequest{
				Spec: &viewing.ViewingSpec{
					Select: &viewing.ViewingSpecSelect{
						ViewingId: request.GetOffer().GetViewing(),
					},
				},
				Offer: m.ID.Hex(),
			},
		)
		if err != nil {
			log.WithError(err).
				Errorf("offering property (%s) viewing failed", request.GetOffer().GetProperty())

			return nil, err
		}
	}

	var verif *offer.OfferVerification
	verif = model.MakeVerification()
	valid := false
	model.VerifyAll(&valid, m, verif)

	return &offer.CreateResponse{
		Id:           m.ID.Hex(),
		Status:       offer.OfferStatus_SUCCESS,
		Verification: verif,
		Offer:        m.RenderOfferModel(ctx, request.GetUser()),
	}, nil
}

//nolint:gocyclo,gocognit
func (server *serverImpl) CreateForRenew(
	ctx context.Context,
	request *offer.CreateForRenewRequest,
) (*offer.CreateResponse, error) {
	// check the renewal offer for the tenancy
	renewalOffer, err := model.GetOffersByPrevTenancy(ctx, request.GetTenancy())
	if err != nil {
		return nil, err
	}

	if len(renewalOffer) > 0 {
		return &offer.CreateResponse{
			Id:     renewalOffer[0].ID.Hex(),
			Status: offer.OfferStatus_SUCCESS,
			Offer:  renewalOffer[0].RenderOfferModel(ctx, request.GetUser()),
		}, nil
	}

	svcctx := grpcctx.WithServiceUser(ctx)

	prop, err := connections.ListingService.VersionGet(
		svcctx,
		&listing.VersionGetRequest{
			Spec: listing.VersionSpecPropertyID("", request.GetOffer().GetProperty()),
		},
	)
	if err != nil {
		return nil, err
	}

	oldOfferID, err := primitive.ObjectIDFromHex(request.GetOffer().GetId())
	if err != nil {
		return nil, err
	}

	reOffer, err := model.GetOfferByID(ctx, oldOfferID)
	if err != nil {
		return nil, err
	}

	// landlords account
	landlord, err := connections.AccountService.Get(ctx, &account.GetRequest{Id: request.GetUser()})
	if err != nil {
		return nil, ErrUnknownUser
	}

	m := &model.OfferModel{
		ID:              primitive.NewObjectID(),
		Property:        request.GetOffer().GetProperty(),
		PropertyIvr:     reOffer.PropertyIvr,
		PropertyListing: prop.GetVersion().GetListingId(),
		PropertyAddress: prop.GetVersion().GetLocation().FormatAddress(),
		CreatedAt:       time.Now(),
		UpdatedAt:       time.Now(),
		State:           offer.OfferState_PENDING,
		Landlord: &model.LandlordModel{
			User:         landlord.GetAccount().GetIdValue(),
			FirstName:    util.PtrString(landlord.GetAccount().GetDetailsValue().GetFirstNameValue()),
			LastName:     util.PtrString(landlord.GetAccount().GetDetailsValue().GetLastNameValue()),
			EmailPrimary: util.PtrString(landlord.GetAccount().GetContactValue().GetEmailPrimaryValue()),
			PhonePrimary: util.PtrString(landlord.GetAccount().GetContactValue().GetPhonePrimaryValue()),
			BillingAddress: util.PtrString(model.GetBillingAddressFromLocation(
				landlord.GetAccount().GetLocationValue(),
			)),
		},
		Manager:       prop.GetVersion().GetPeople().GetManager().GetValue(),
		Tenants:       reOffer.Tenants,
		Guarantors:    reOffer.Guarantors,
		Clause:        reOffer.Clause,
		Details:       reOffer.Details,
		Test:          reOffer.Test,
		OriginalOffer: request.GetOffer().GetId(),
		PrevTenancy:   request.GetTenancy(),
		Renewal:       true,
	}

	now := time.Now()
	movein := request.GetOffer().GetConditions().GetMoveinDate().AsTime()
	start := util.UnPtrTime(reOffer.Conditions.MoveinDate)

	if movein.Before(now) {
		request.GetOffer().GetConditions().MoveinDate = timestamppb.New(
			start.AddDate(0, int(util.UnPtrInt32(reOffer.Conditions.TermMonths)), 0),
		)
	}

	err = m.UpdateOfferModel(ctx, &offer.Offer{
		Conditions: request.Offer.Conditions,
	})
	if err != nil {
		return nil, err
	}

	_, err = connections.TenancyService.ServiceTenancySetNewOffer(svcctx, &tenancy.ServiceTenancySetNewOfferRequest{
		TenancyId:  request.GetTenancy(),
		NewOfferId: m.ID.Hex(),
	})
	if err != nil {
		return nil, err
	}

	depositReplacement := m.DepositReplacement()

	for _, t := range m.Tenants {
		t.Signature = model.SignatureModelArray{}
		if t.User != "" {
			acc, err := connections.AccountService.Get(ctx, &account.GetRequest{Id: t.User})
			if err != nil {
				log.WithError(err).WithFields(
					log.Fields{
						"original_offer": m.OriginalOffer,
						"offer":          m.ID.Hex(),
						"user":           t.User,
					}).Error("tenant's account not found for renewal offer")
				continue
			}
			check := false
			if v := acc.GetAccount().GetDetailsValue().GetFirstNameValue(); util.UnPtrString(t.FirstName) != v {
				t.FirstName = &v
				check = true
			}
			if v := acc.GetAccount().GetDetailsValue().GetLastNameValue(); util.UnPtrString(t.LastName) != v {
				t.LastName = &v
				check = true
			}
			if v := acc.GetAccount().GetContactValue().GetEmailPrimaryValue(); util.UnPtrString(t.EmailPrimary) != v {
				t.EmailPrimary = &v
				check = true
			}
			if v := acc.GetAccount().GetContactValue().GetPhonePrimaryValue(); util.UnPtrString(t.PhonePrimary) != v {
				t.PhonePrimary = &v
				check = true
			}
			if v := model.GetBillingAddressFromLocation(
				acc.GetAccount().GetLocationValue(),
			); util.UnPtrString(t.BillingAddress) != v {
				t.BillingAddress = &v
				check = true
			}

			if check {
				t.RentprofileCreatedAt = nil
				t.RentprofileStatus = nil
				t.RiskURL = nil
				t.Risk = offer.TenantRisk_UNKNOWN_TENANT_RISK
				t.RiskDocument = nil
				t.CheckNeeded = depositReplacement
			}
		}
	}

	for _, g := range m.Guarantors {
		g.Signature = model.SignatureModelArray{}
		if g.User != "" {
			acc, err := connections.AccountService.Get(ctx, &account.GetRequest{Id: g.User})
			if err != nil {
				log.WithError(err).WithFields(
					log.Fields{
						"original_offer": m.OriginalOffer,
						"offer":          m.ID.Hex(),
						"user":           g.User,
					}).Error("guarantor's account not found for renewal offer")
				continue
			}
			check := false
			if v := acc.GetAccount().GetDetailsValue().GetFirstNameValue(); util.UnPtrString(g.FirstName) != v {
				g.FirstName = &v
				check = true
			}
			if v := acc.GetAccount().GetDetailsValue().GetLastNameValue(); util.UnPtrString(g.LastName) != v {
				g.LastName = &v
				check = true
			}
			if v := acc.GetAccount().GetContactValue().GetEmailPrimaryValue(); util.UnPtrString(g.EmailPrimary) != v {
				g.EmailPrimary = &v
				check = true
			}
			if v := acc.GetAccount().GetContactValue().GetPhonePrimaryValue(); util.UnPtrString(g.PhonePrimary) != v {
				g.PhonePrimary = &v
				check = true
			}
			if v := model.GetBillingAddressFromLocation(
				acc.GetAccount().GetLocationValue(),
			); util.UnPtrString(g.BillingAddress) != v {
				g.BillingAddress = &v
				check = true
			}

			if check {
				g.RentprofileCreatedAt = nil
				g.RentprofileStatus = nil
				g.RiskURL = nil
				g.Risk = offer.TenantRisk_UNKNOWN_TENANT_RISK
				g.RiskDocument = nil
				g.CheckNeeded = depositReplacement
			}
		}
	}

	for i := range m.Clause {
		m.Clause[i].State = offer.ClauseState_CLAUSE_PENDING
	}

	err = model.CreateOffer(ctx, m)
	if err != nil {
		return nil, err
	}

	reOffer.State = offer.OfferState_AWAITING_RENEW

	reOffer.Sync.Reset()

	err = model.UpdateOffer(ctx, reOffer)
	if err != nil {
		log.WithError(err).Error("changing renewed offer state")
	}

	synchronization.ServiceOfferSync(ctx, m)

	_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_RENEWAL,
		Action: notification.Action_ACTION_CREATE,
		Targets: []*notification.Target{
			{
				User: m.TenantUser(),
				Role: notification.Role_ROLE_TENANT,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: m.Property,
			},
		},
	})
	if err != nil {
		log.WithError(err).Error("send EventType_EVENT_OFFER_TENANCY_OFFERED_RENEWAL_NOTIFY_TENANT" +
			" notification error")
	}

	return &offer.CreateResponse{
		Id:     m.ID.Hex(),
		Status: offer.OfferStatus_SUCCESS,
		Offer:  m.RenderOfferModel(ctx, request.GetUser()),
	}, nil
}

func (server *serverImpl) ServiceUpdate(
	ctx context.Context,
	request *offer.ServiceUpdateRequest,
) (response *offer.ServiceUpdateResponse, err error) {
	id, _ := primitive.ObjectIDFromHex(request.GetOffer().GetId())
	req := &offer.UpdateRequest{
		User:  request.User,
		Offer: request.Offer,
	}
	resp, err := server.doUpdate(ctx, req, true)
	if err != nil {
		return nil, err
	}

	bgctx := grpcctx.ValuesBackgroundContext(ctx)

	if request.GetRenderContract() {
		_, err = server.bpms.NewPublishMessageCommand().
			MessageName("contract-prepared").
			CorrelationKey(id.Hex()).
			TimeToLive(status.BPMTimeToLive).
			Send(bgctx)
		if err != nil {
			return nil, err
		}
	}
	result, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return nil, err
	}

	offerLog := model.NewOfferLog(ctx, result, request.GetUser())

	if result.State == offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES {
		err = status.GenerateDraftContract(ctx, result)
		if err != nil {
			return nil, err
		}
		err = model.UpdateOffer(ctx, result)
		if err != nil {
			return nil, err
		}
	}

	offerLog.LogDiff()

	return &offer.ServiceUpdateResponse{
		Status: resp.Status,
		Errors: resp.Errors,
		Offer:  result.RenderOfferModel(ctx, request.GetUser()),
	}, nil
}

func (server *serverImpl) Update(
	ctx context.Context,
	request *offer.UpdateRequest,
) (*offer.UpdateResponse, error) {
	acc, err := connections.AccountService.Get(ctx, &account.GetRequest{Id: request.GetUser()})
	if err != nil {
		log.WithError(err).WithField("user", request.GetUser()).Error("on getting account")
		return nil, ErrUnknownUser
	}

	if !acc.GetAccount().GetContactValue().GetContactStatusValue().GetPhoneVerifiedValue() {
		return nil, ErrPhoneNotVerified
	}

	return server.doUpdate(ctx, request, false)
}

func (server *serverImpl) RemoveSecondLandlord(
	ctx context.Context,
	request *offer.RemoveSecondLandlordRequest,
) (response *offer.RemoveSecondLandlordResponse, err error) {
	id, err := primitive.ObjectIDFromHex(request.GetOfferId())
	if err != nil {
		return nil, err
	}

	old, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	offerLog := model.NewOfferLog(ctx, old, request.GetUser())

	old.Landlords = nil

	err = model.UpdateOffer(ctx, old)
	if err != nil {
		return nil, err
	}

	offerLog.LogDiff()

	return &offer.RemoveSecondLandlordResponse{
		Success: true,
		Offer:   old.RenderOfferModel(ctx, request.GetUser()),
	}, nil
}

func (server *serverImpl) ensureInvoiceState(ctx context.Context, old *model.OfferModel, off *offer.Offer) error {
	if old.State == offer.OfferState_HOLDING_DEPOSIT &&
		off.GetConditions().GetPrice() != nil &&
		old.Conditions != nil &&
		old.Conditions.Price != nil &&
		*old.Conditions.Price != off.GetConditions().GetPrice().GetValue() {
		*old.Conditions.Price = off.GetConditions().GetPrice().GetValue()

		err := status.OfferCreateHoldingDepositInvoice(ctx, old)
		if err != nil {
			return err
		}
	}

	return nil
}

func extractMatchingEmail(user string, old *model.OfferModel) string {
	switch {
	case user == old.LandlordUser():
		return util.UnPtrString(old.Landlord.EmailPrimary)
	case user == old.TenantUser():
		return util.UnPtrString(old.Tenants[0].EmailPrimary)
	}

	return ""
}

//nolint:gocognit
func (server *serverImpl) doUpdate(
	ctx context.Context,
	request *offer.UpdateRequest,
	adminSite bool,
) (response *offer.UpdateResponse, err error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer().GetId())
	if err != nil {
		return nil, err
	}

	var old *model.OfferModel

	old, err = model.GetOfferByID(ctx, id)
	if err != nil {
		return nil, err
	}

	isTenant := !adminSite && request.GetUser() != old.Landlord.User

	if adminSite {
		response, err = server.updateRentPlan(ctx, old, request, adminSite)
		if err != nil || response != nil {
			return response, err
		}
	} else {
		switch {
		case err != nil:
			return nil, err
		case len(old.Tenants) == 0:
			return nil, status.ErrNoAccess
		case old.TenantUser() != request.GetUser() && old.LandlordUser() != request.GetUser():
			return nil, status.ErrNoAccess
		}

		switch old.State {
		case offer.OfferState_ACTIVE,
			offer.OfferState_ACCEPTED,
			offer.OfferState_HOLDING_DEPOSIT,
			offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES,
			offer.OfferState_DEPOSIT_CONFIRMED,
			offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT,
			offer.OfferState_CHECKING,
			offer.OfferState_AWAITING_CHECKS:
			if isTenant && time.Since(old.UpdatedAt) > time.Second*5 {
				_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
					Entity: notification.Entity_ENTITY_OFFER,
					Action: notification.Action_ACTION_UPDATE,
					Targets: []*notification.Target{
						{
							User: old.LandlordUser(),
							Role: notification.Role_ROLE_LANDLORD,
						},
					},
					Embeds: []*notification.EventEmbed{
						{
							EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
							EmbedValue: old.Property,
						},
						{
							EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
							EmbedValue: old.ID.Hex(),
						},
						{
							EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
							EmbedValue: old.TenantUser(),
						},
					},
				})

				if err != nil {
					log.WithError(err).Error("sending offer update to manager notification")
				}
			}
		case offer.OfferState_PENDING:
		case offer.OfferState_SIGNED,
			offer.OfferState_RENTPAID,
			offer.OfferState_AWAITING_TENANCY:
			return server.updateRentPlan(ctx, old, request, adminSite)
		default:
			return nil, ErrNoAccessStatus
		}
	}

	offerLog := model.NewOfferLog(ctx, old, request.GetUser())

	err = server.ensureInvoiceState(ctx, old, request.GetOffer())
	if err != nil {
		return nil, err
	}

	err = old.UpdateOfferModel(ctx, request.GetOffer())
	if err != nil {
		return nil, err
	}

	switch old.State {
	case offer.OfferState_DEPOSIT_CONFIRMED,
		offer.OfferState_CHECKING,
		offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT,
		offer.OfferState_AWAITING_GUARANTOR,
		offer.OfferState_AWAITING_CHECKS,
		offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES:
		err = status.GenerateDraftContract(ctx, old)
		if err != nil {
			return nil, err
		}
	case offer.OfferState_HOLDING_DEPOSIT:
		bgctx := grpcctx.ValuesBackgroundContext(ctx)

		_, _ = server.bpms.NewPublishMessageCommand().
			MessageName("offer-updated").
			CorrelationKey(id.Hex()).
			TimeToLive(status.BPMTimeToLive).
			Send(bgctx)

		old.IsLocked = true
	}

	if isTenant {
		old.UpdatedByTenantAt = time.Now()
	}

	err = model.UpdateOffer(ctx, old)
	if err != nil {
		return nil, err
	}

	var verif *offer.OfferVerification
	verif = model.MakeVerification()
	valid := false
	model.VerifyAll(&valid, old, verif)

	offerLog.LogDiff()

	email := extractMatchingEmail(request.GetUser(), old)

	bgctx := grpcctx.ValuesBackgroundContext(ctx)

	if old.State > offer.OfferState_PENDING {
		go func() {
			if err := mixpanel.TrackOfferUpdated(bgctx, request.GetUser(), email, old); err != nil {
				log.WithError(err).Error("failed to track OfferUpdated mixpanel event")
			}
		}()
	}

	return &offer.UpdateResponse{
		Status:       offer.OfferStatus_SUCCESS,
		Verification: verif,
		Offer:        old.RenderOfferModel(ctx, request.GetUser()),
	}, nil
}

func updateRentPlanInitialSelection(old *model.OfferModel, request *offer.UpdateRequest) bool {
	return old.RentPlan == 0 && request.GetUser() == old.LandlordUser() &&
		(old.State == offer.OfferState_SIGNED || old.State == offer.OfferState_AWAITING_TENANCY)
}

func updateRentPlanLegacySelection(old *model.OfferModel, request *offer.UpdateRequest) bool {
	rentPlan := request.GetOffer().GetRentPlan()

	return old.RentPlan == offer.RentPlan_RENT_PLAN_LEGACY && request.GetUser() == old.LandlordUser() &&
		(rentPlan == offer.RentPlan_RENT_PLAN_ESSENTIAL || rentPlan == offer.RentPlan_RENT_PLAN_SECURED ||
			rentPlan == offer.RentPlan_RENT_PLAN_MANUAL)
}

func (server *serverImpl) updateRentPlan(
	ctx context.Context,
	old *model.OfferModel,
	request *offer.UpdateRequest,
	adminSite bool,
) (resp *offer.UpdateResponse, err error) {
	rentPlan := request.GetOffer().GetRentPlan()

	if rentPlan == 0 {
		return
	}

	if request.GetUser() != old.LandlordUser() && !adminSite {
		return
	}

	initialSelection := updateRentPlanInitialSelection(old, request)

	legacySelection := updateRentPlanLegacySelection(old, request)

	if !(initialSelection || legacySelection || adminSite) {
		return nil, ErrNoAccessStatus
	}

	init := old.RentPlan

	old.RentPlan = rentPlan

	err = status.TryCreateSignedInvoices(ctx, old, false)
	if err != nil {
		return
	}

	if old.RentPlan == offer.RentPlan_RENT_PLAN_MANUAL ||
		(old.RentPlan != offer.RentPlan_RENT_PLAN_MANUAL &&
			(old.State == offer.OfferState_SIGNED || old.State == offer.OfferState_AWAITING_TENANCY)) {
		err = status.SetPropertyRented(ctx, old)
		if err != nil {
			return
		}
	}

	err = model.UpdateOffer(ctx, old)
	if err != nil {
		return
	}

	go func(old *model.OfferModel, init offer.RentPlan, adminSite bool) {
		server.rentPlanUpgradedNotify(context.Background(), old, init, adminSite)
	}(old, init, adminSite)

	if init != 0 && init != old.RentPlan && old.ScheduleGenerated {
		_, err = connections.ExpensesService.ServiceRentPlanUpdate(ctx, &expenses.ServiceRentPlanUpdateRequest{
			CorrelationType:          expenses.CorrelationType_CORRELATION_TYPE_OFFER,
			CorrelationId:            old.ID.Hex(),
			RentCollectionFeePercent: old.RentPlan.Percent(),
		})
		if err != nil {
			return
		}
	}

	if legacySelection && old.ScheduleGenerated {
		go func(ctx context.Context, m *model.OfferModel) {
			_ = status.TrySetPropertyRentedCreateInsurances(grpcctx.ValuesBackgroundContext(ctx), m, false)
		}(ctx, old)
	}

	return &offer.UpdateResponse{
		Status:       offer.OfferStatus_SUCCESS,
		Verification: model.MakeVerification(),
		Offer:        old.RenderOfferModel(ctx, request.GetUser()),
	}, nil
}

func (server *serverImpl) rentPlanUpgradedNotify(
	ctx context.Context,
	old *model.OfferModel,
	oldRentPlan offer.RentPlan,
	isAdminTriggered bool,
) {
	bs, _ := json.Marshal(map[string]interface{}{
		"collection": map[string]interface{}{
			"old_rent_plan": oldRentPlan.String(),
			"new_rent_plan": old.RentPlan.String(),
			"updated_at":    time.Now().Format("2006-01-02T15:04:05Z07:00"),
		},
		"offer": map[string]interface{}{
			"id": old.ID.Hex(),
		},
	})

	_, err := connections.NotificationService.SendNotification(
		context.Background(),
		&notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_OFFER_RENT_COLLECTION,
			Action: notification.Action_ACTION_UPDATE,
			Targets: []*notification.Target{
				{
					Role: notification.Role_ROLE_ADMIN,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: old.Property,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_JSON,
					EmbedValue: string(bs),
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
					EmbedValue: old.LandlordUser(),
				},
			},
		})
	if err != nil {
		log.WithError(err).Error("Rent plan has been upgraded admin notification")
	}

	notificationEntity := notification.Entity_ENTITY_OFFER_RENT_COLLECTION_NOT_MANUAL

	if old.RentPlan == offer.RentPlan_RENT_PLAN_MANUAL {
		notificationEntity = notification.Entity_ENTITY_OFFER_RENT_COLLECTION_MANUAL
	}

	bs, _ = json.Marshal(map[string]interface{}{
		"is_admin_changed_rent_plan": isAdminTriggered,
	})

	if notificationEntity > 0 {
		_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
			Entity: notificationEntity,
			Action: notification.Action_ACTION_COMPLETE,
			Targets: []*notification.Target{
				{
					User: old.LandlordUser(),
					Role: notification.Role_ROLE_LANDLORD,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: old.Property,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
					EmbedValue: old.ID.Hex(),
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_JSON,
					EmbedValue: string(bs),
				},
			},
		})
		if err != nil {
			log.WithError(err).Error("Rent plan has been upgraded landlord notification")
		}

		_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
			Entity: notificationEntity,
			Action: notification.Action_ACTION_COMPLETE,
			Targets: []*notification.Target{
				{
					User: old.TenantUser(),
					Role: notification.Role_ROLE_TENANT,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: old.Property,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
					EmbedValue: old.ID.Hex(),
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_JSON,
					EmbedValue: string(bs),
				},
			},
		})
		if err != nil {
			log.WithError(err).Error("Rent plan has been upgraded tenant notification")
		}
	}
}

func (server *serverImpl) executeChecksSet(
	ctx context.Context,
	m *model.OfferModel,
	tenants []string,
	sReq *identity.TenantsInviteRequest,
) (ts []*model.TenantOfferModel, gs []*model.GuarantorOfferModel, err error) {
	tenantsmap := make(map[string]struct{})

	for _, t := range tenants {
		if t == "" {
			continue
		}

		tenantsmap[t] = struct{}{}
	}

	var tenantNames, checkApplicants []string

	ts = make([]*model.TenantOfferModel, 0)
	gs = make([]*model.GuarantorOfferModel, 0)

	for _, t := range m.Tenants {
		_, ok := tenantsmap[util.UnPtrString(t.EmailPrimary)]
		if !ok {
			continue
		}

		ts = append(ts, t)
		sReq.Tenants = append(sReq.Tenants, &identity.TenantDescription{
			ApplicantName:   util.UnPtrString(t.FirstName) + " " + util.UnPtrString(t.LastName),
			ApplicantEmail:  util.UnPtrString(t.EmailPrimary),
			ApplicantMobile: util.UnPtrString(t.PhonePrimary),
		})

		tenantNames = append(tenantNames, util.UnPtrString(t.FirstName)+" "+util.UnPtrString(t.LastName))

		tenantEmailPrimary := util.UnPtrString(t.EmailPrimary)
		if tenantEmailPrimary == "" {
			continue
		}

		checkApplicants = append(checkApplicants, tenantEmailPrimary)
	}

	for _, g := range m.Guarantors {
		_, ok := tenantsmap[util.UnPtrString(g.EmailPrimary)]
		if !ok {
			continue
		}

		gs = append(gs, g)
		sReq.Guarantors = append(sReq.Guarantors, &identity.GuarantorDescription{
			ApplicantName: util.UnPtrString(g.FirstName) +
				" " + util.UnPtrString(g.LastName),
			ApplicantEmail:  util.UnPtrString(g.EmailPrimary),
			ApplicantMobile: util.UnPtrString(g.PhonePrimary),
			TenantName:      strings.Join(tenantNames, " and "),
		})

		guarantorEmailPrimary := util.UnPtrString(g.EmailPrimary)
		if guarantorEmailPrimary == "" {
			continue
		}

		checkApplicants = append(checkApplicants, guarantorEmailPrimary)
	}

	if len(checkApplicants) > 0 {
		err = server.applicantCheckPaymentPaid(ctx, m, checkApplicants)
		if err != nil {
			return
		}
	}

	return
}

func (server *serverImpl) executeChecks(ctx context.Context, m *model.OfferModel, tenants []string) error {
	if !m.InProgress() {
		return nil
	}

	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id: m.Property,
	})
	if err != nil {
		return err
	}

	loc := prop.GetProperty().GetLocation()

	if m.Conditions == nil {
		m.Conditions = &model.OfferConditionsModel{}
	}

	sReq := &identity.TenantsInviteRequest{
		Id: m.ID.Hex(),
		Address: fmt.Sprintf(
			"%s %s %s",
			loc.GetBuilding().GetValue(),
			loc.GetStreet().GetValue(),
			loc.GetCity().GetValue(),
		),
		Postcode:  loc.GetPostcode().GetValue(),
		TotalRent: int32(util.UnPtrFloat64(m.Conditions.Price)),
	}

	ts, gs, err := server.executeChecksSet(ctx, m, tenants, sReq)
	if err != nil {
		return err
	}

	if len(sReq.Tenants) == 0 && len(sReq.Guarantors) == 0 {
		return ErrNoUser
	}

	log.Printf("check tenants: %#v", sReq.GetTenants())
	log.Printf("check guarantors: %#v", sReq.GetGuarantors())
	resp, err := connections.IdentityService.TenantsInvite(ctx, sReq)
	if err != nil {
		return err
	}
	for i, n := range resp.RequestId {
		if i >= len(ts) {
			gs[i-len(ts)].Check = n
			gs[i-len(ts)].RiskDocument = nil
			gs[i-len(ts)].Risk = offer.TenantRisk_PENDING_RISK
			gs[i-len(ts)].RiskURL = nil
		} else {
			ts[i].Check = n
			ts[i].RiskDocument = nil
			ts[i].Risk = offer.TenantRisk_PENDING_RISK
			ts[i].RiskURL = nil
		}
	}

	status.ExportClosecomOpportunity(ctx, m, closecom.OfferStatus_OFFER_STATUS_AWAITING_REFERENCE)

	err = server.handlePurchaseOfReferenceCheck(ctx, m, sReq)

	if err != nil {
		log.WithError(err).Errorf("Failed to handle purchase of ref check, offer %s", m.ID.Hex())
	}

	log.Printf("check ids: %#v", m)

	return nil
}

func (Server *serverImpl) handlePurchaseOfReferenceCheck(
	ctx context.Context,
	offer *model.OfferModel,
	identityRequest *identity.TenantsInviteRequest,
) error {
	checkItems := []*notification.SlackCheckItem{}

	mainTenant := offer.TenantModel()

	if mainTenant == nil {
		mainTenant = &model.TenantOfferModel{}
	}

	for index, tenant := range identityRequest.GetTenants() {
		var userId string

		if tenant.ApplicantEmail == util.UnPtrString(mainTenant.EmailPrimary) {
			userId = mainTenant.User
		}

		checkItems = append(checkItems, &notification.SlackCheckItem{
			Id:       userId,
			Index:    int64(index) + 1,
			Name:     tenant.ApplicantName,
			Email:    tenant.ApplicantEmail,
			ItemType: notification.ItemType_TYPE_TENANT,
		})
	}

	for index, guarantor := range identityRequest.GetGuarantors() {
		checkItems = append(checkItems, &notification.SlackCheckItem{
			Index:    int64(index) + 1,
			Name:     guarantor.ApplicantName,
			Email:    guarantor.ApplicantEmail,
			ItemType: notification.ItemType_TYPE_GUARANTOR,
		})
	}

	landlordLeadResponse, err := connections.ClosecomService.LeadContactGet(
		grpcctx.WithServiceUser(ctx),
		&closecom.LeadContactGetRequest{UserId: offer.Landlord.User},
	)

	if err != nil {
		return err
	}

	mainTenantLeadResponse, err := connections.ClosecomService.LeadContactGet(
		grpcctx.WithServiceUser(ctx),
		&closecom.LeadContactGetRequest{UserId: mainTenant.User},
	)

	if err != nil {
		return err
	}

	slackRequest := &notification.SendSlackCheckStartRequest{
		Landlord: &notification.SlackObject{
			Name:   fmt.Sprintf("%s %s", util.UnPtrString(offer.Landlord.FirstName), util.UnPtrString(offer.Landlord.LastName)),
			Id:     offer.Landlord.User,
			LeadId: landlordLeadResponse.GetLeadContact().GetLeadId(),
		},
		MainTenant: &notification.SlackObject{
			Name:   fmt.Sprintf("%s %s", util.UnPtrString(mainTenant.FirstName), util.UnPtrString(mainTenant.LastName)),
			Id:     offer.TenantUser(),
			LeadId: mainTenantLeadResponse.GetLeadContact().GetLeadId(),
		},
		Offer: &notification.SlackObject{
			Id: offer.ID.Hex(),
		},
		Property: &notification.SlackObject{
			Address: offer.PropertyAddress,
			Id:      offer.Property,
		},
		CheckItems: checkItems,
	}

	_, err = connections.NotificationService.SendSlackCheckStart(grpcctx.WithServiceUser(ctx), slackRequest)

	if err != nil {
		return err
	}

	return nil
}

func (server *serverImpl) ServiceApprove(
	ctx context.Context,
	request *offer.ApproveRequest,
) (response *offer.ApproveResponse, err error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer())
	if err != nil {
		return nil, err
	}
	var old *model.OfferModel
	old, err = model.GetOfferByID(ctx, id)
	if err != nil {
		return nil, err
	}

	offerLog := model.NewOfferLog(ctx, old, request.GetUser())

	old.UpdateApproveOfferModel(true, request.GetUser())
	err = model.UpdateOffer(ctx, old)
	if err != nil {
		return nil, err
	}

	offerLog.LogDiff()

	return &offer.ApproveResponse{
		Ok:       true,
		Approved: true,
		Offer:    old.RenderOfferModel(ctx, request.GetUser()),
	}, nil
}

func (server *serverImpl) Submit(
	ctx context.Context,
	request *offer.SubmitRequest,
) (response *offer.SubmitResponse, err error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer())
	if err != nil {
		return nil, err
	}
	old, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	offerLog := model.NewOfferLog(ctx, old, request.GetUser())

	defer func() {
		if err != nil {
			return
		}

		offerLog.LogDiff()
	}()

	if len(old.Tenants) == 0 || old.Tenants[0].User != request.GetUser() {
		return nil, status.ErrNoAccess
	}

	switch old.State {
	case offer.OfferState_PENDING,
		offer.OfferState_AWAITING_RENEW:
		// Ok, go further
	case offer.OfferState_ERROR,
		offer.OfferState_EXPIRED,
		offer.OfferState_RENTPAID:
		return nil, ErrNoAccessStatus
	default:
		return &offer.SubmitResponse{
			Status: offer.OfferStatus_SUCCESS,
		}, nil
	}

	err = status.SetOfferState(ctx, old, offer.OfferState_ACTIVE) // also sends platform events
	if err != nil {
		return nil, err
	}

	if old.ScheduleNotifications == nil {
		old.ScheduleNotifications = make(map[model.ScheduleNotificationType]*model.ScheduleNotification)
	}

	now := time.Now()

	old.ScheduleNotifications[model.ScheduleNotificationNotResponded3hrs] = &model.ScheduleNotification{
		StartedAt:  now,
		IsNotified: false,
	}

	movein := model.UnPtrTime(old.Conditions.MoveinDate)
	if time.Until(movein) > time.Hour*24*7 {
		old.ScheduleNotifications[model.ScheduleNotificationWeekLeftToMovein] = &model.ScheduleNotification{
			StartedAt:  movein,
			IsNotified: false,
		}
	}

	err = model.UpdateOffer(ctx, old)
	if err != nil {
		return nil, err
	}

	go func() {
		bgctx := grpcctx.ValuesBackgroundContext(ctx)
		if err := mixpanel.TrackTenantOfferSubmitted(bgctx, old); err != nil {
			log.WithError(err).Error("failed to track OfferSubmitted mixpanel event")
		}

		if err := mixpanel.TrackLandlordOfferReceived(bgctx, old); err != nil {
			log.WithError(err).Error("failed to track OfferReceived mixpanel event")
		}
	}()

	return &offer.SubmitResponse{
		Status: offer.OfferStatus_SUCCESS,
	}, nil
}

// MarkAsNotified marks offer schedule notification
func (server *serverImpl) MarkAsNotified(
	ctx context.Context,
	request *offer.MarkAsNotifiedRequest,
) (*offer.MarkAsNotifiedResponse, error) {
	return nil, grpcerr.NotFound
}

// ListNotNotified gets list didn't notify offers by notification event types
func (server *serverImpl) ListNotNotified(
	ctx context.Context,
	request *offer.ListNotNotifiedRequest,
) (*offer.ListResponse, error) {
	return nil, grpcerr.NotFound
}

func (server *serverImpl) UpdateTenant(
	ctx context.Context,
	request *offer.UpdateTenantRequest,
) (*offer.UpdateTenantResponse, error) {
	return nil, status.ErrNoAccess
}

func (server *serverImpl) RemoveTenant(
	ctx context.Context,
	request *offer.RemoveTenantRequest,
) (*offer.RemoveTenantResponse, error) {
	return nil, status.ErrNoAccess
}

func (server *serverImpl) Accept(
	ctx context.Context,
	request *offer.AcceptRequest,
) (response *offer.AcceptResponse, err error) {
	offId, err := primitive.ObjectIDFromHex(request.GetId())
	if err != nil {
		return nil, err
	}

	old, err := model.GetOffer(ctx, offId, request.GetUser())
	if err != nil {
		return nil, err
	}

	offerLog := model.NewOfferLog(ctx, old, request.GetUser())

	defer func() {
		if err != nil {
			return
		}

		offerLog.LogDiff()
	}()

	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id: old.Property,
	})

	if err != nil {
		return nil, err
	}

	people := prop.GetProperty().GetPeople()
	if !util.EqualStringValue(&request.User, people.GetLandlord()) || old.State != offer.OfferState_ACTIVE {
		return nil, status.ErrNoAccess
	}

	oldoffers, err := model.GetOffersByProperty(ctx, old.Property)
	if err != nil {
		return nil, err
	}

	var awaitTermination string

	for _, oldoffer := range oldoffers {
		v, s := offerProcessingState(ctx, oldoffer)
		if v {
			return nil, ErrPropRentedOrOffered
		}

		if s != "" {
			awaitTermination = s
		}
	}

	old.AwaitingTermination = awaitTermination

	err = status.SetOfferState(ctx, old, offer.OfferState_ACCEPTED)
	if err != nil {
		return nil, err
	}

	nextstate := offer.OfferState_HOLDING_DEPOSIT

	if old.OriginalOffer != "" {
		nextstate = offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES
	}

	err = status.SetOfferState(ctx, old, nextstate)
	if err != nil {
		return nil, err
	}

	bgctx := grpcctx.ValuesBackgroundContext(ctx)
	_, err = connections.ZeebeExporter.Start(ctx, &zeebe.StartRequest{
		Remote:  connections.SelfRemote,
		Process: listing.PropertyOfferProcessID,
		Payload: &structpb.Struct{
			Fields: map[string]*structpb.Value{
				"offer": {Kind: &structpb.Value_StringValue{
					StringValue: offId.Hex(),
				}},
				"property": {Kind: &structpb.Value_StringValue{
					StringValue: prop.GetProperty().GetId(),
				}},
				"landlord": {Kind: &structpb.Value_StringValue{
					StringValue: people.GetLandlord().GetValue(),
				}},
				"manager": {Kind: &structpb.Value_StringValue{
					StringValue: people.GetManager().GetValue(),
				}},
				"renewal_process": {Kind: &structpb.Value_BoolValue{
					BoolValue: old.Renewal,
				}},
			},
		},
	})
	if err != nil {
		return nil, err
	}

	//_, err = server.bpms.NewPublishMessageCommand().
	//	MessageName("offer-accepted").
	//	CorrelationKey(offId.Hex()).
	//	TimeToLive(status.BPMTimeToLive).
	//	Send(bgctx)
	//if err != nil {
	//	return nil, err
	//}

	go notifyOfferAccepted(
		bgctx,
		old.ID.Hex(),
		old.Landlord.User,
		old.Tenants[0].User,
		old.Property,
	)

	email := extractMatchingEmail(request.GetUser(), old)

	go func() {
		if err := mixpanel.TrackOfferAccepted(bgctx, request.GetUser(), email, old); err != nil {
			log.WithError(err).Error("failed to track OfferAccepted mixpanel event")
		}
	}()

	return &offer.AcceptResponse{Status: offer.OfferStatus_SUCCESS}, nil
}

func offerProcessingState(ctx context.Context, m *model.OfferModel) (bool, string) {
	switch m.State {
	case offer.OfferState_PENDING, offer.OfferState_ACTIVE, offer.OfferState_UNKNOWN_OFFER_STATE:
		return false, ""
	default:
		v, s := offerTerminalState(ctx, m)

		return !v, s
	}
}

func offerTerminalState(ctx context.Context, m *model.OfferModel) (bool, string) {
	switch m.State {
	case offer.OfferState_DENIED,
		offer.OfferState_CANCELED,
		offer.OfferState_TERMINATED,
		offer.OfferState_FINISHED,
		offer.OfferState_EXPIRED,
		offer.OfferState_ERROR,
		offer.OfferState_EARLY_TERMINATION:
		return true, ""
	case offer.OfferState_RENTPAID, offer.OfferState_SIGNED, offer.OfferState_AWAITING_RENEW:
		if m.Conditions != nil {
			months := util.UnPtrInt32(m.Conditions.TermMonths)

			if util.UnPtrTime(m.Conditions.MoveinDate).AddDate(0, int(months), 0).Before(time.Now()) {
				return true, ""
			}
		}

		tenancyResp, err := connections.TenancyService.TenancyGet(ctx, &tenancy.TenancyGetRequest{
			Spec: &tenancy.TenancySpec{
				UserId: m.LandlordUser(),
				Details: &tenancy.TenancySpec_Select{
					Select: &tenancy.TenancySpecSelect{
						States: []tenancy.TenancyState{
							tenancy.TenancyState_TENANCY_ACTIVE,
							tenancy.TenancyState_TENANCY_TERMINATING,
						},
						By: &tenancy.TenancySpecSelect_OfferId{
							OfferId: m.ID.Hex(),
						},
					},
				},
			},
		})
		if errors.Is(err, grpcerr.NotFound) {
			return true, ""
		}

		if err != nil {
			log.WithError(err).Errorf("getting tenancy for %v", m.ID.Hex())
			return false, ""
		}

		landInt := tenancyResp.GetTenancy().GetLandlordIntention()
		relistAllowedT := landInt.TerminationDate.AsTime().AddDate(0, -2, 0)

		if time.Now().After(relistAllowedT.Truncate(24 * time.Hour)) {
			return true, m.ID.Hex()
		}
	}

	return false, ""
}

func notifyOfferAccepted(ctx context.Context, offerID, landlordID, tenantID, propertyID string) {
	_, err := connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER,
		Action: notification.Action_ACTION_ACCEPT,
		Targets: []*notification.Target{
			{
				User: landlordID,
				Role: notification.Role_ROLE_LANDLORD,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: propertyID,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: offerID,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
				EmbedValue: tenantID,
			},
		},
	})
	if err != nil {
		log.WithError(err).Error("sending offer accept notification")
	}

	_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER,
		Action: notification.Action_ACTION_ACCEPT,
		Targets: []*notification.Target{
			{
				User: tenantID,
				Role: notification.Role_ROLE_TENANT,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: propertyID,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: offerID,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
				EmbedValue: landlordID,
			},
		},
	})
	if err != nil {
		log.WithError(err).Error("sending offer accept notification")
	}
}

// Deny is an landlord/manager action to decline offer.
func (server *serverImpl) Deny(
	ctx context.Context,
	request *offer.DenyRequest,
) (response *offer.DenyResponse, err error) {
	if request.GetUser() == "" {
		return nil, status.ErrNoAccess
	}

	id, err := primitive.ObjectIDFromHex(request.GetId())
	if err != nil {
		return nil, err
	}

	old, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id: old.Property,
	})
	if err != nil {
		return nil, err
	}

	offerLog := model.NewOfferLog(ctx, old, request.GetUser())

	defer func() {
		if err != nil {
			return
		}

		offerLog.LogDiff()
	}()

	curInv := &invoice.Invoice{}
	isNoAccess := true
	oldState := old.State

	switch oldState {
	case offer.OfferState_ACTIVE,
		offer.OfferState_ACCEPTED,
		offer.OfferState_HOLDING_DEPOSIT,
		offer.OfferState_AWAITING_CONFIRM,
		offer.OfferState_CONTRACT_PREPARATION,
		offer.OfferState_AWAITING_CHECKS,
		offer.OfferState_SIGNING,
		offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT,
		offer.OfferState_DEPOSIT_CONFIRMED,
		offer.OfferState_AWAITING_GUARANTOR,
		offer.OfferState_CHECKING,
		offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES:
		isNoAccess = false
	}

	landlordID := prop.GetProperty().GetPeople().GetLandlord().GetValue()

	email := extractMatchingEmail(request.GetUser(), old)

	if request.GetUser() != landlordID || isNoAccess {
		if !(request.GetUser() == old.TenantUser() &&
			oldState == offer.OfferState_AWAITING_GUARANTOR) {
			return nil, status.ErrNoAccess
		}
	}

	err = status.SetOfferState(ctx, old, offer.OfferState_DENIED)
	if err != nil {
		return nil, err
	}

	priceDeposit := 0.0
	for _, it := range curInv.GetServices() {
		priceDeposit += it.GetService().GetPrice() + it.GetService().GetTaxAmount()
	}

	if oldState == offer.OfferState_ACTIVE || oldState == offer.OfferState_HOLDING_DEPOSIT {
		_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_OFFER,
			Action: notification.Action_ACTION_REJECT,
			Targets: []*notification.Target{
				{
					User: old.TenantUser(),
					Role: notification.Role_ROLE_TENANT,
				},
				{
					User: old.LandlordUser(),
					Role: notification.Role_ROLE_LANDLORD,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: old.Property,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
					EmbedValue: old.ID.Hex(),
				},
			},
		})
		if err != nil {
			log.WithError(err).Errorf("sending undepsoited offer deny notification failed")
		}
	}

	go func() {
		bgctx := grpcctx.ValuesBackgroundContext(ctx)
		if err := mixpanel.TrackOfferDeclined(bgctx, request.GetUser(), email, old); err != nil {
			log.WithError(err).Error("failed to track OfferDeclined mixpanel event")
		}
	}()

	server.returnPropertyToListed(ctx, old.PropertyListing)

	paid, err := invoices.HoldingDepositHasPaid(ctx, old.ID.Hex())
	if err != nil {
		log.WithError(err).Error("offer deny: HoldingDepositHasPaid() failed")
	}

	if !paid {
		return &offer.DenyResponse{Status: offer.OfferStatus_SUCCESS}, nil
	}

	_, _ = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_DEPOSITED,
		Action: notification.Action_ACTION_REJECT,
		Targets: []*notification.Target{
			{
				User: old.TenantUser(),
				Role: notification.Role_ROLE_TENANT,
			},
			{
				User: old.LandlordUser(),
				Role: notification.Role_ROLE_LANDLORD,
			},
			{
				Role: notification.Role_ROLE_ADMIN,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: old.Property,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: old.ID.Hex(),
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
				EmbedValue: old.Tenants[0].User,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PAYMENT_COST,
				EmbedValue: strconv.FormatFloat(priceDeposit, 'f', 2, 64),
			},
		},
	})
	if err != nil {
		log.WithError(err).Errorf("sending deposited offer notification failed")
	}

	return &offer.DenyResponse{Status: offer.OfferStatus_SUCCESS}, nil
}

func (server *serverImpl) returnPropertyToListed(ctx context.Context, listingID string) {
	ver, err := connections.ListingService.VersionGet(grpcctx.WithServiceUser(ctx), &listing.VersionGetRequest{
		Spec: listing.VersionSpecListingID("", listingID),
	})
	if err != nil {
		return
	}

	_, err = connections.ListingService.VersionPublish(grpcctx.WithServiceUser(ctx), &listing.VersionPublishRequest{
		VersionId: ver.GetVersion().GetId(),
		LetAgreed: wrapperspb.Bool(false),
	})
	if err != nil {
		log.WithError(err).Errorf("Error republish listing: %v", listingID)
	}
}

// Cancel allows to cancel offer from tenant side.
func (server *serverImpl) Cancel(
	ctx context.Context,
	request *offer.CancelRequest,
) (response *offer.CancelResponse, err error) {
	id, err := primitive.ObjectIDFromHex(request.GetId())
	if err != nil {
		return nil, err
	}

	old, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	if request.GetUser() == "" || request.GetUser() != old.TenantUser() {
		return nil, status.ErrNoAccess
	}

	offerLog := model.NewOfferLog(ctx, old, request.GetUser())

	defer func() {
		if err != nil {
			return
		}

		offerLog.LogDiff()
	}()

	switch old.State {
	case offer.OfferState_ACTIVE,
		offer.OfferState_ACCEPTED,
		offer.OfferState_PENDING,
		offer.OfferState_HOLDING_DEPOSIT,
		offer.OfferState_AWAITING_CONFIRM,
		offer.OfferState_CONTRACT_PREPARATION,
		offer.OfferState_AWAITING_CHECKS,
		offer.OfferState_SIGNING,
		offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT,
		offer.OfferState_DEPOSIT_CONFIRMED,
		offer.OfferState_AWAITING_GUARANTOR,
		offer.OfferState_CHECKING,
		offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES:
	default:
		return nil, ErrNoAccessStatus
	}

	oldState := old.State

	err = status.SetOfferState(ctx, old, offer.OfferState_CANCELED)
	if err != nil {
		return nil, err
	}

	notify := func(entity notification.Entity) error {
		_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
			Entity: entity,
			Action: notification.Action_ACTION_CANCEL,
			Targets: []*notification.Target{
				{
					User: old.TenantUser(),
					Role: notification.Role_ROLE_TENANT,
				},
				{
					User: old.LandlordUser(),
					Role: notification.Role_ROLE_LANDLORD,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: old.Property,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
					EmbedValue: old.ID.Hex(),
				},
			},
		})
		return err
	}

	if oldState != offer.OfferState_PENDING && oldState != offer.OfferState_ACTIVE {
		err = notify(notification.Entity_ENTITY_OFFER)
		if err != nil {
			log.WithError(err).Error("sending non-deposited offer cancel notification failed")
		}
	}

	paid, err := invoices.HoldingDepositHasPaid(ctx, old.ID.Hex())
	if err != nil {
		log.WithError(err).Error("offer cancel: HoldingDepositHasPaid() failed")
	}

	if paid {
		err = notify(notification.Entity_ENTITY_OFFER_DEPOSITED)
		if err != nil {
			log.WithError(err).Error("sending deposited offer cancel notification failed")
		}
	}

	server.returnPropertyToListed(ctx, old.PropertyListing)

	return &offer.CancelResponse{Status: offer.OfferStatus_SUCCESS}, nil
}

func (server *serverImpl) RequestChange(
	ctx context.Context,
	request *offer.RequestChangeRequest,
) (response *offer.RequestChangeResponse, err error) {
	id, err := primitive.ObjectIDFromHex(request.GetId())
	if err != nil {
		return nil, err
	}
	old, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	offerLog := model.NewOfferLog(ctx, old, request.GetUser())

	defer func() {
		if err != nil {
			return
		}

		offerLog.LogDiff()
	}()

	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id: old.Property,
	})

	if err != nil {
		return nil, err
	}

	people := prop.GetProperty().GetPeople()
	if !util.EqualStringValue(&request.User, people.GetLandlord()) || old.State != offer.OfferState_ACTIVE {
		return nil, status.ErrNoAccess
	}

	err = status.SetOfferState(ctx, old, offer.OfferState_PENDING)
	if err != nil {
		return nil, err
	}

	return &offer.RequestChangeResponse{Status: offer.OfferStatus_SUCCESS}, nil
}

func (server *serverImpl) ExecuteCheck(
	ctx context.Context,
	request *offer.ExecuteCheckRequest,
) (response *offer.CheckResponse, err error) {
	id, err := primitive.ObjectIDFromHex(request.GetId())
	if err != nil {
		return nil, err
	}

	m, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return nil, err
	}

	if m.Test {
		ctx = context.WithValue(ctx, grpcctx.CtxIsTestUser, true)
	}

	offerLog := model.NewOfferLog(ctx, m, "")

	defer func() {
		if err != nil {
			return
		}

		offerLog.LogDiff()
	}()

	if !m.InProgress() {
		return &offer.CheckResponse{
			Errors: []offer.OfferError{offer.OfferError_INVALID_OFFER_STATUS},
		}, nil
	}

	if m.State == offer.OfferState_AWAITING_CHECKS || m.State == offer.OfferState_CHECKING {
		err = server.retryCheck(ctx, m, request.GetTenants())
		if err != nil {
			return nil, err
		}
		return &offer.CheckResponse{Status: offer.OfferStatus_SUCCESS}, nil
	}

	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id: m.Property,
	})
	fmt.Println(prop, err)
	if err != nil {
		return nil, err
	}

	loc := prop.GetProperty().GetLocation()
	fmt.Println(loc)
	if m.Conditions == nil {
		m.Conditions = &model.OfferConditionsModel{}
	}

	err = server.executeChecks(ctx, m, request.GetTenants())
	if err == ErrNoUser {
		return &offer.CheckResponse{
			Errors: []offer.OfferError{offer.OfferError_USER_NOT_EXISTS},
		}, nil
	}

	if err != nil {
		return nil, err
	}

	err = model.UpdateOffer(ctx, m)
	if err != nil {
		return nil, err
	}

	return &offer.CheckResponse{Status: offer.OfferStatus_SUCCESS}, nil
}

func (server *serverImpl) retryCheck(
	ctx context.Context,
	old *model.OfferModel,
	tenants []string,
) error {
	if old.State == offer.OfferState_AWAITING_CHECKS || old.State == offer.OfferState_CHECKING {
		bgctx := grpcctx.ValuesBackgroundContext(ctx)

		msg, err := server.bpms.NewPublishMessageCommand().
			MessageName("offer-retry-checks").
			CorrelationKey(old.ID.Hex()).
			TimeToLive(status.BPMTimeToLive).
			VariablesFromMap(map[string]interface{}{
				"tenants": tenants,
			})
		if err != nil {
			log.Error(err)

			return err
		}

		_, err = msg.Send(bgctx)
		if err != nil {
			log.Error(err)

			return err
		}
	}

	return nil
}

func checkRiskAcceptable(t *model.TenantOfferModel, guarantors []*model.GuarantorOfferModel) bool {
	switch t.Risk {
	case offer.TenantRisk_UNKNOWN_TENANT_RISK, offer.TenantRisk_PENDING_RISK:
		return false
	case offer.TenantRisk_HIGH_RISK:
		if len(guarantors) == 0 {
			return false
		}

		for _, g := range guarantors {
			switch g.Risk {
			case offer.TenantRisk_UNKNOWN_TENANT_RISK, offer.TenantRisk_PENDING_RISK, offer.TenantRisk_HIGH_RISK:
				return false
			}
		}
	}

	return true
}

//nolint:gocognit,gocyclo // try to reduce high cognitive and cyclomatic complexity
func (server *serverImpl) Check(ctx context.Context, request *offer.CheckRequest) (*offer.CheckResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetId())
	if err != nil {
		return nil, err
	}
	old, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	checkInit := old.State == offer.OfferState_DEPOSIT_CONFIRMED ||
		old.State == offer.OfferState_CHECKING ||
		old.State == offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT ||
		old.State == offer.OfferState_AWAITING_GUARANTOR
	checkProcess := checkInit || old.State == offer.OfferState_AWAITING_CHECKS

	depositReplacement := old.DepositReplacement()

	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id: old.Property,
	})
	if err != nil {
		return nil, err
	}

	people := prop.GetProperty().GetPeople()
	if !util.EqualStringValue(&request.User, people.GetLandlord()) || !checkProcess {
		return nil, status.ErrNoAccess
	}

	if depositReplacement && request.GetForceskip() && old.State != offer.OfferState_AWAITING_CHECKS {
		for _, t := range old.Tenants {
			if !checkRiskAcceptable(t, old.Guarantors) {
				return nil, ErrCantSkipChecks
			}
		}
	}

	bgctx := grpcctx.ValuesBackgroundContext(ctx)

	if request.GetForceskip() &&
		(old.State == offer.OfferState_DEPOSIT_CONFIRMED ||
			old.State == offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT) {
		if old.State == offer.OfferState_AWAITING_GUARANTOR {
			_, err = server.bpms.NewPublishMessageCommand().
				MessageName("guarantor-added").
				CorrelationKey(id.Hex()).
				TimeToLive(status.BPMTimeToLive).
				Send(bgctx)
			if err != nil {
				log.Error(err)
				return nil, err
			}
		}

		err = status.SetOfferState(ctx, old, offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES)
		if err != nil {
			return nil, err
		}

		if old.State == offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT {
			_, err := connections.InvoiceService.InvoiceCancel(ctx, &invoice.InvoiceCancelRequest{
				CorrelationType:   invoice.CorrelationTypeOffer,
				CorrelationId:     old.ID.Hex(),
				CorrelationTarget: invoice.CorrelationTargetReferenceCheck,
				User:              request.GetUser(),
			})

			if err != nil && !errors.Is(err, grpcerr.NotFound) {
				return nil, err
			}

			_, err = server.bpms.NewPublishMessageCommand().
				MessageName("proceed-without-checks").
				CorrelationKey(id.Hex()).
				TimeToLive(status.BPMTimeToLive).
				Send(bgctx)
			if err != nil {
				log.Error(err)
				return nil, err
			}
		}

		_, err = server.bpms.NewPublishMessageCommand().
			MessageName("offer-skip-checks").
			CorrelationKey(id.Hex()).
			TimeToLive(status.BPMTimeToLive).
			Send(bgctx)
		if err != nil {
			log.Error(err)
			return nil, err
		}
	}

	if (old.State == offer.OfferState_AWAITING_CHECKS || old.State == offer.OfferState_CHECKING) &&
		request.GetForceskip() {
		for _, t := range old.Tenants {
			if !checkRiskAcceptable(t, old.Guarantors) {
				return nil, ErrCantSkipChecks
			}
		}

		err = status.SetOfferState(ctx, old, offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES)
		if err != nil {
			return nil, err
		}

		_, err = server.bpms.NewPublishMessageCommand().
			MessageName("checks-complete").
			CorrelationKey(id.Hex()).
			TimeToLive(status.BPMTimeToLive).
			Send(bgctx)
		if err != nil {
			log.Error(err)
			return nil, err
		}
	}

	if (old.State == offer.OfferState_DEPOSIT_CONFIRMED ||
		old.State == offer.OfferState_AWAITING_CHECKS) &&
		!request.GetForceskip() {
		err = status.SetOfferState(ctx, old, offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT)
		if err != nil {
			return nil, err
		}

		_, err = server.bpms.NewPublishMessageCommand().
			MessageName("offer-run-checks").
			CorrelationKey(id.Hex()).
			TimeToLive(status.BPMTimeToLive).
			Send(bgctx)
		if err != nil {
			log.Error(err)
			return nil, err
		}
	}

	return &offer.CheckResponse{Status: offer.OfferStatus_SUCCESS}, nil
}

// RequestRequiredGuarantor to add a new guarantor for the current tenants,
// if any of tenants has HIGH risk from rentprofile checking, and the landlord
// decided to check a new guarantor of the failed tenant.
// Also can be called if the tenant and his guarantor still have HIGH risks.
func (server *serverImpl) RequestRequiredGuarantor(
	ctx context.Context,
	request *offer.RequestRequiredGuarantorRequest,
) (response *offer.RequestRequiredGuarantorResponse, err error) {
	id, err := primitive.ObjectIDFromHex(request.GetId())
	if err != nil {
		return nil, err
	}
	old, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	offerLog := model.NewOfferLog(ctx, old, request.GetUser())

	defer func() {
		if err != nil {
			return
		}

		offerLog.LogDiff()
	}()

	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id: old.Property,
	})
	if err != nil {
		return nil, err
	}

	people := prop.GetProperty().GetPeople()

	if !util.EqualStringValue(&request.User, people.GetLandlord()) ||
		(old.State != offer.OfferState_AWAITING_CHECKS && old.State != offer.OfferState_CHECKING) {
		return nil, status.ErrNoAccess
	}

	old.GuarantorRequired = true

	err = model.UpdateOffer(ctx, old)
	if err != nil {
		return nil, err
	}

	err = status.SetOfferState(ctx, old, offer.OfferState_AWAITING_GUARANTOR)
	if err != nil {
		return nil, err
	}

	bgctx := grpcctx.ValuesBackgroundContext(ctx)

	_, err = server.bpms.NewPublishMessageCommand().
		MessageName("guarantor-required").
		CorrelationKey(id.Hex()).
		TimeToLive(status.BPMTimeToLive).
		Send(bgctx)
	if err != nil {
		log.Error(err)
		return nil, err
	}

	return &offer.RequestRequiredGuarantorResponse{Success: true}, nil
}

// AddGuarantors adds new guarantors by tenant, if the offer status is "AWAITING_GUARANTOR"
func (server *serverImpl) AddGuarantors(
	ctx context.Context,
	request *offer.AddGuarantorsRequest,
) (response *offer.AddGuarantorsResponse, err error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer())
	if err != nil {
		return nil, err
	}
	old, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	offerLog := model.NewOfferLog(ctx, old, request.GetUser())

	if request.User != old.Tenants[0].User || old.State != offer.OfferState_AWAITING_GUARANTOR {
		return nil, status.ErrNoAccess
	}
	if request.GetTenantValue() == nil {
		return nil, ErrAddGuarantorTenantsRequired
	}

	for _, tnew := range request.GetTenantValue() {
		_, err = old.UpdateOfferGuarantorModel(ctx, &offer.Guarantor{
			FirstName:      wrapperspb.String(tnew.GetGuarantorFirstName()),
			LastName:       wrapperspb.String(tnew.GetGuarantorLastName()),
			EmailPrimary:   wrapperspb.String(tnew.GetGuarantorEmail()),
			PhonePrimary:   wrapperspb.String(tnew.GetGuarantorPhone()),
			BillingAddress: wrapperspb.String(tnew.GetGuarantorBillingAddress()),
		}, true, true)
		if err != nil {
			return nil, err
		}
	}

	err = model.UpdateOffer(ctx, old)
	if err != nil {
		return nil, err
	}

	err = status.SetOfferState(ctx, old, offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT)
	if err != nil {
		return nil, err
	}

	bgctx := grpcctx.ValuesBackgroundContext(ctx)

	_, err = server.bpms.NewPublishMessageCommand().
		MessageName("guarantor-added").
		CorrelationKey(id.Hex()).
		TimeToLive(status.BPMTimeToLive).
		Send(bgctx)
	if err != nil {
		log.Error(err)
		return nil, err
	}

	offerLog.LogDiff()

	return &offer.AddGuarantorsResponse{
		Success: true,
		Offer:   old.RenderOfferModel(ctx, ""),
	}, nil
}

func (server *serverImpl) ServicesList(
	ctx context.Context,
	request *offer.ListServicesRequest,
) (*offer.ListServicesResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer())
	if err != nil {
		return nil, err
	}
	userID := request.GetUser()
	old, err := model.GetOffer(ctx, id, userID)
	if err != nil {
		return nil, err
	}

	switch old.State {
	case offer.OfferState_SIGNED, offer.OfferState_AWAITING_TENANCY:
		// ok
	default:
		return nil, ErrNoAccessStatus
	}

	var services []*extras.ServiceSnapshot

	switch {
	case len(old.Tenants) > 0 && old.Tenants[0].User == userID:
		svcs, err := connections.ExtrasService.List(ctx, &extras.ListRequest{
			Role: extras.ServiceRole_TENANT_SERVICES,
		})
		if err != nil {
			return nil, err
		}

		services = extras.SnapshotServices(svcs.GetExtraServices())
	case old.Landlord.User == userID:
		svcs, err := connections.ExtrasService.List(ctx, &extras.ListRequest{
			Role: extras.ServiceRole_LANDLORD_SERVICES,
		})
		if err != nil {
			return nil, err
		}

		services = extras.SnapshotServices(svcs.GetExtraServices())
	}

	if services == nil {
		return nil, status.ErrNoAccess
	}

	var inServices []*extras.ServiceSnapshot

	for _, service := range services {
		switch service.GetType() {
		case extras.ServiceType_DEPOSIT_INSURANCE,
			extras.ServiceType_RENT_GUARANTEE_INSURANCE,
			extras.ServiceType_HOME_EMERGENCY_INSURANCE:
			inServices = append(inServices, service)
		}
	}

	if len(inServices) > 0 {
		accResp, err := connections.AccountService.Get(ctx, &account.GetRequest{
			Id: userID,
		})
		if err != nil {
			return nil, err
		}
		name := strings.TrimSpace(strings.Join([]string{
			accResp.GetAccount().GetDetailsValue().GetLastNameValue(),
			accResp.GetAccount().GetDetailsValue().GetFirstNameValue(),
		}, " "))
		checkResp, err := connections.IdentityService.SanctionsCheck(ctx, &identity.SanctionsCheckRequest{
			Name: name,
		})
		if err != nil {
			return nil, err
		}

		if checkResp.GetFound() {
			for _, service := range inServices {
				service.Title = fmt.Sprintf("%s - %s", service.GetTitle(), SanctionsCheckFailed)
			}
		}
	}

	return &offer.ListServicesResponse{
		Services: services,
	}, nil
}

// ServicesSelect selects services for signed offer
func (server *serverImpl) ServicesSelect(
	ctx context.Context,
	request *offer.SelectServicesRequest,
) (*offer.SelectServicesResponse, error) {
	return nil, errors.New("not implemented")
}

// DepositSelect - create first month deposit replacement invoice, create deposit schedule
func (server *serverImpl) DepositSelect(
	ctx context.Context,
	request *offer.DepositSelectRequest,
) (*offer.DepositSelectResponse, error) {
	return &offer.DepositSelectResponse{
		Success: true,
		Errors:  nil,
	}, nil
}

func (server *serverImpl) GetPropertyIvrNum(ctx context.Context, propertyID string) (string, error) {
	ivrNum := ""
	if propertyID == "" {
		return ivrNum, grpcerr.New(grpcerr.InvalidData, "property ID is empty")
	}

	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id: propertyID,
	})
	if err != nil {
		log.WithError(err).Info("CreatePayment get property error")
		return ivrNum, err
	}

	return prop.GetProperty().GetIvr(), nil
}

//nolint:gocognit // try to reduce high cognitive and cyclomatic complexity
func (server *serverImpl) Payment(
	ctx context.Context,
	request *offer.PaymentRequest,
) (response *offer.PaymentResponse, err error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer())
	if err != nil {
		return nil, err
	}

	old, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	offerLog := model.NewOfferLog(ctx, old, request.GetUser())

	defer func() {
		if err != nil {
			return
		}

		offerLog.LogDiff()
	}()

	switch old.State {
	case offer.OfferState_HOLDING_DEPOSIT:
		if len(old.Tenants) == 0 || old.Tenants[0].User != request.GetUser() {
			return nil, status.ErrNoAccess
		}
	case offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT:
		if old.Landlord.User != request.GetUser() {
			return nil, status.ErrNoAccess
		}
	case offer.OfferState_SIGNED, offer.OfferState_AWAITING_TENANCY:
		// ok
	default:
		return nil, ErrNoAccessStatus
	}

	paymentType := request.GetType()
	if paymentType == offer.PaymentType_UNKNOWN_PAYMENT_TYPE {
		paymentType = offer.PaymentType_CARD
	}

	switch {
	case len(old.Tenants) > 0 && old.Tenants[0].User == request.GetUser():
		correlationTarget := invoice.CorrelationTargetTenantServices

		if old.State == offer.OfferState_HOLDING_DEPOSIT {
			correlationTarget = invoice.CorrelationTargetHoldingDeposit
		}

		var inv *invoice.InvoiceGetResponse

		inv, err = connections.InvoiceService.InvoiceGet(ctx, &invoice.InvoiceGetRequest{
			CorrelationType:   invoice.CorrelationTypeOffer,
			CorrelationId:     old.ID.Hex(),
			CorrelationTarget: correlationTarget,
			Type:              invoice.InvoiceType_INVOICE_TYPE_PAYMENT,
			User:              request.GetUser(),
			PaymentStatuses: []invoice.PaymentStatus{
				invoice.PaymentStatus_PAYMENT_CONFIRM,
				invoice.PaymentStatus_PAYMENT_SUCCESS,
			},
			Statuses: []invoice.InvoiceStatus{invoice.InvoiceStatus_INVOICE_OPEN},
		})
		if inv.GetInvoice() != nil {
			return nil, ErrPaymentExists
		}

		inv, err = connections.InvoiceService.InvoiceGet(ctx, &invoice.InvoiceGetRequest{
			CorrelationType:   invoice.CorrelationTypeOffer,
			CorrelationId:     old.ID.Hex(),
			CorrelationTarget: correlationTarget,
			User:              request.GetUser(),
			Statuses:          []invoice.InvoiceStatus{invoice.InvoiceStatus_INVOICE_OPEN},
		})
		if err != nil {
			return nil, err
		}

		var desc string

		purpose := tokenio.PaymentPurpose_UPFRONT_PAYMENT

		switch old.State {
		case offer.OfferState_HOLDING_DEPOSIT:
			desc = "Deposit payment for property " + old.Property + " at mashroom.com"
		case offer.OfferState_SIGNED:
			desc = "Payment for rent and services for property " + old.Property + " at mashroom.com"
			purpose = tokenio.PaymentPurpose_RENT_PAYMENT
		}

		var res *offer.PaymentResponse

		switch paymentType {
		case offer.PaymentType_CARD:
			res = &offer.PaymentResponse{
				Tokenio: &offer.PaymentTokenioResponse{},
			}

			var resp *invoice.PaymentCreateResponse

			resp, err = connections.InvoiceService.PaymentCreate(ctx, &invoice.PaymentCreateRequest{
				InvoiceId: inv.GetInvoice().GetInvoiceId(),
				User:      request.GetUser(),
				System: &invoice.PaymentSystemDataRequest{
					Data: &invoice.PaymentSystemDataRequest_Tokenio{
						Tokenio: &invoice.PaymentSystemTokenioRequest{
							Description:     desc,
							Ivr:             old.PropertyIvr,
							Purpose:         purpose,
							RedirectSuccess: request.GetRedirectSuccess(),
							RedirectError:   request.GetRedirectError(),
						},
					},
				},
			})
			if err != nil {
				return nil, err
			}

			res.Tokenio.Url = resp.GetData().GetTokenio().GetUrl()
			res.Tokenio.Csrf = resp.GetData().GetTokenio().GetCsrf()
		case offer.PaymentType_BANK_ACCOUNT:
			var pid string

			pid, err = server.holdingDepositPaymentBankAccount(
				ctx,
				old,
				inv,
				request.GetPaymentProcedure(),
				request.GetUser(),
			)

			res = &offer.PaymentResponse{
				Bank: &offer.PaymentBankResponse{
					Id:      pid,
					Success: true,
				},
			}
		default:
			return nil, grpcerr.New(grpcerr.InvalidData, "unknown payment type")
		}

		return res, nil
	case old.Landlord.User == request.GetUser():
		if request.GetApplicantCheck() == nil {
			return server.createServicesPayment(ctx, old, request, paymentType)
		}

		if request.GetApplicantCheck().GetSkip() {
			return server.skipApplicationCheckPayment(ctx, old)
		}

		return server.createReferenceCheckPayment(ctx, old, request)
	}

	return nil, status.ErrNoAccess
}

func (server *serverImpl) holdingDepositPaymentBankAccount(
	ctx context.Context,
	old *model.OfferModel,
	inv *invoice.InvoiceGetResponse,
	proc offer.PaymentProcedure,
	user string,
) (pid string, err error) {
	if proc == offer.PaymentProcedure_PAYMENT_PROCEDURE_CONFIRMATION {
		for _, p := range inv.GetInvoice().GetPayments() {
			if p.GetStatus() == invoice.PaymentStatus_PAYMENT_PENDING {
				pid = p.GetPaymentId()

				break
			}
		}

		_, err = connections.InvoiceService.PaymentConfirm(ctx, &invoice.PaymentConfirmRequest{
			PaymentId: pid,
			User:      user,
		})
		if err != nil {
			return
		}

		err = server.awaitPaymentConfirm(ctx, old)
		if err != nil {
			return
		}

		return
	}

	resp, err := connections.InvoiceService.PaymentCreate(ctx, &invoice.PaymentCreateRequest{
		InvoiceId: inv.GetInvoice().GetInvoiceId(),
		User:      user,
		System: &invoice.PaymentSystemDataRequest{
			Data: &invoice.PaymentSystemDataRequest_Bank{
				Bank: &invoice.PaymentSystemBankRequest{},
			},
		},
	})
	if err != nil {
		return
	}

	pid = resp.GetPayment().GetPaymentId()

	return
}

// DirectDebit - ...
func (server *serverImpl) DirectDebit(
	ctx context.Context,
	request *offer.DirectDebitRequest,
) (*offer.DirectDebitResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer())
	if err != nil {
		return nil, err
	}

	old, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	switch old.State {
	case offer.OfferState_SIGNED,
		offer.OfferState_AWAITING_CONFIRM_RENT,
		offer.OfferState_RENTPAID,
		offer.OfferState_AWAITING_TENANCY:
		if len(old.Tenants) == 0 || old.Tenants[0].User != request.GetUser() {
			return nil, status.ErrNoAccess
		}
	default:
		return nil, ErrNoAccessStatus
	}

	inv, err := connections.InvoiceService.InvoiceGet(ctx, &invoice.InvoiceGetRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     request.GetOffer(),
		CorrelationTarget: invoice.CorrelationTargetDepositReplacement,
		Statuses: []invoice.InvoiceStatus{
			invoice.InvoiceStatus_INVOICE_OPEN,
		},
	})
	if err != nil {
		return nil, err
	}

	for _, p := range inv.GetInvoice().GetPayments() {
		if p.Status == invoice.PaymentStatus_PAYMENT_PENDING {
			return nil, ErrDRPAlreadyBeenSetup
		}
	}

	resp, err := connections.InvoiceService.PaymentCreate(ctx, &invoice.PaymentCreateRequest{
		InvoiceId:         inv.GetInvoice().GetInvoiceId(),
		User:              request.User,
		CorrelationTarget: invoice.CorrelationTargetDepositReplacement,
		System: &invoice.PaymentSystemDataRequest{
			Data: &invoice.PaymentSystemDataRequest_Stripe{
				Stripe: &invoice.PaymentSystemStripeRequest{
					Type:            cardpayment.PaymentType_DIRECT_DEBIT,
					RedirectSuccess: request.GetRedirectSuccess(),
					RedirectCancel:  request.GetRedirectError(),
				},
			},
		},
	})
	if err != nil {
		return nil, err
	}

	if old.PaidStatuses == nil {
		old.PaidStatuses = &model.PaidStatusesModel{}
	}

	old.PaidStatuses.DepositReplacement = true

	err = model.UpdateOffer(ctx, old)
	if err != nil {
		return nil, err
	}

	err = status.SetPropertyRented(ctx, old)
	if err != nil {
		return nil, err
	}

	status.TryExportClosecomMoveinStatus(ctx, old)

	return &offer.DirectDebitResponse{
		Stripe: &offer.PaymentStripeResponse{
			ClientSecret: resp.GetData().GetStripe().GetClientSecret(),
		},
	}, nil
}

var dataPeriodicPaymentFrequency = map[int32]invoice.PeriodicPaymentFrequency{
	1:  invoice.PeriodicPaymentFrequency_FREQUENCY_MONTHLY,
	3:  invoice.PeriodicPaymentFrequency_FREQUENCY_QUARTERLY,
	6:  invoice.PeriodicPaymentFrequency_FREQUENCY_TWICE_A_YEAR,
	12: invoice.PeriodicPaymentFrequency_FREQUENCY_YEARLY,
}

func paymentDayClamp(d *int32) (day int) {
	if d != nil {
		day = int(*d)
	}

	if day < int(model.FirstRentPaymentDay) {
		day = int(model.FirstRentPaymentDay)
	}

	if day > int(model.LastRentPaymentDay) {
		day = int(model.LastRentPaymentDay)
	}

	return
}

//nolint:gocognit
func (server *serverImpl) StandingOrder( // todo: call for rto from frontend
	ctx context.Context,
	request *offer.StandingOrderRequest,
) (*offer.StandingOrderResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer())
	if err != nil {
		return nil, err
	}

	old, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	switch old.State {
	case offer.OfferState_SIGNED,
		offer.OfferState_AWAITING_CONFIRM_RENT,
		offer.OfferState_RENTPAID,
		offer.OfferState_AWAITING_TENANCY,
		offer.OfferState_RTO_STATE_WAITING_FOR_FIRST_RENT_PAYMENT,
		offer.OfferState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP:
		if len(old.Tenants) == 0 || old.Tenants[0].User != request.GetUser() {
			return nil, status.ErrNoAccess
		}
	default:
		return nil, ErrNoAccessStatus
	}

	if old.Conditions == nil ||
		old.Conditions.PaymentFrequency == nil ||
		old.Conditions.Price == nil ||
		old.Conditions.TermMonths == nil ||
		old.Conditions.MoveinDate == nil {
		return nil, fmt.Errorf(
			"%w: empty condition payment frequency/price/term months/movein date",
			grpcerr.InvalidState,
		)
	}

	inv, err := connections.InvoiceService.InvoiceGet(ctx, &invoice.InvoiceGetRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     old.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetStandingOrder,
		Statuses:          []invoice.InvoiceStatus{invoice.InvoiceStatus_INVOICE_OPEN},
	})
	if err != nil {
		return nil, err
	}

	freq := *old.Conditions.PaymentFrequency
	frequency := dataPeriodicPaymentFrequency[freq]
	if frequency == 0 {
		return nil, fmt.Errorf("unsupported standing order payment frequency: %v, offer: %v", freq, old.ID.Hex())
	}

	res := &offer.StandingOrderResponse{
		Tokenio: &offer.PaymentTokenioResponse{},
	}

	var lastPayMonths int32

	termMonths := *old.Conditions.TermMonths

	if termMonths%freq == 0 {
		lastPayMonths = termMonths - freq
	} else {
		lastPayMonths = termMonths - (termMonths % freq) - freq
	}

	movein := *old.Conditions.MoveinDate
	day := paymentDayClamp(old.Conditions.RentPaymentDay)
	startAt := timestamppb.New(movein.AddDate(0, 1, -movein.Day()+day))
	endAt := timestamppb.New(movein.AddDate(0, int(lastPayMonths)+1, 0))
	daily := *old.Conditions.Price * status.MonthlyToDailyFactor

	resp, err := connections.InvoiceService.PaymentCreate(ctx, &invoice.PaymentCreateRequest{
		InvoiceId:         inv.GetInvoice().GetInvoiceId(),
		User:              request.GetUser(),
		CorrelationTarget: invoice.CorrelationTargetStandingOrder,
		System: &invoice.PaymentSystemDataRequest{
			Data: &invoice.PaymentSystemDataRequest_Tokenio{
				Tokenio: &invoice.PaymentSystemTokenioRequest{
					Description: fmt.Sprintf(
						"Standing order payment for rent and services for property %v at mashroom.com",
						old.Property,
					),
					Ivr:             old.PropertyIvr,
					Purpose:         tokenio.PaymentPurpose_STANDING_ORDER_PAYMENT,
					Frequency:       frequency,
					StartAt:         startAt,
					EndAt:           endAt,
					DailyAmount:     daily,
					RedirectSuccess: request.GetRedirectSuccess(),
					RedirectError:   request.GetRedirectError(),
				},
			},
		},
	})
	if err != nil {
		return nil, err
	}

	res.Tokenio.Url = resp.GetPayment().GetSystem().GetTokenio().GetResponse().GetUrl()
	res.Tokenio.Csrf = resp.GetPayment().GetSystem().GetTokenio().GetResponse().GetCsrf()

	if old.PaidStatuses == nil {
		old.PaidStatuses = &model.PaidStatusesModel{}
	}

	old.PaidStatuses.StandingOrder = true

	if old.State == offer.OfferState_RTO_STATE_WAITING_FOR_FIRST_RENT_PAYMENT ||
		old.State == offer.OfferState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP {
		// todo: usecase
	} else {
		err = model.UpdateOffer(ctx, old)
		if err != nil {
			return nil, err
		}

		err = status.SetPropertyRented(ctx, old)
		if err != nil {
			return nil, err
		}
	}

	return res, nil
}

func (server *serverImpl) createReferenceCheckPayment(
	ctx context.Context,
	old *model.OfferModel,
	request *offer.PaymentRequest,
) (*offer.PaymentResponse, error) {
	if len(request.GetApplicantCheck().GetTenants()) == 0 {
		return nil, ErrTenantsRequired
	}

	service, err := extras.GetServiceSnapshot(ctx, connections.ExtrasService, extras.ProductLandlordReferencingCheck)
	if err != nil {
		return nil, err
	}

	var checkServices []*extras.ServiceSnapshot

	for _, t := range request.GetApplicantCheck().GetTenants() {
		serviceInstance := &extras.ServiceSnapshot{}

		proto.Merge(serviceInstance, service)

		serviceInstance.TargetType = "user-reference"
		serviceInstance.TargetId = t

		checkServices = append(checkServices, serviceInstance)
	}

	refinv, err := connections.InvoiceService.InvoiceCreate(ctx, &invoice.InvoiceCreateRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     old.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetReferenceCheck,
		Property:          old.Property,
		Ivr:               old.PropertyIvr,
		Listing:           old.PropertyListing,
		User:              old.Landlord.User,
		Services:          checkServices,
		NotifyCallback:    connections.SelfRemote,
	})
	if err != nil {
		return nil, err
	}

	resp, err := connections.InvoiceService.PaymentCreate(ctx, &invoice.PaymentCreateRequest{
		InvoiceId: refinv.GetInvoice().GetInvoiceId(),
		User:      request.GetUser(),
		System: &invoice.PaymentSystemDataRequest{
			Data: &invoice.PaymentSystemDataRequest_Stripe{
				Stripe: &invoice.PaymentSystemStripeRequest{
					Type:            cardpayment.PaymentType_CARD,
					RedirectSuccess: request.GetRedirectSuccess(),
					RedirectCancel:  request.GetRedirectError(),
				},
			},
		},
	})
	if err != nil {
		log.Error(err)
		return nil, err
	}

	return &offer.PaymentResponse{
		Stripe: &offer.PaymentStripeResponse{
			Id:           resp.GetData().GetStripe().GetStripeId(),
			ClientSecret: resp.GetData().GetStripe().GetClientSecret(),
		},
		Success: true,
	}, nil
}

func (server *serverImpl) skipApplicationCheckPayment(
	ctx context.Context,
	old *model.OfferModel,
) (*offer.PaymentResponse, error) {
	if old.DepositReplacement() {
		return nil, ErrCantSkipChecks
	}

	err := status.SetOfferState(ctx, old, offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES)
	if err != nil {
		return nil, err
	}

	bgctx := grpcctx.ValuesBackgroundContext(ctx)

	_, err = server.bpms.NewPublishMessageCommand().
		MessageName("proceed-without-checks").
		CorrelationKey(old.ID.Hex()).
		TimeToLive(status.BPMTimeToLive).
		Send(bgctx)
	if err != nil {
		return nil, err
	}

	return &offer.PaymentResponse{
		Success: true,
	}, nil
}

func (server *serverImpl) createServicesPayment(
	ctx context.Context,
	old *model.OfferModel,
	request *offer.PaymentRequest,
	paymentType offer.PaymentType,
) (*offer.PaymentResponse, error) {
	inv, err := connections.InvoiceService.InvoiceGet(ctx, &invoice.InvoiceGetRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     old.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetLandlordServices,
		User:              request.GetUser(),
		Statuses:          []invoice.InvoiceStatus{invoice.InvoiceStatus_INVOICE_OPEN},
	})
	if err != nil {
		return nil, err
	}

	resp, err := connections.InvoiceService.PaymentCreate(ctx, &invoice.PaymentCreateRequest{
		InvoiceId: inv.GetInvoice().GetInvoiceId(),
		User:      request.GetUser(),
		System: &invoice.PaymentSystemDataRequest{
			Data: &invoice.PaymentSystemDataRequest_Stripe{
				Stripe: &invoice.PaymentSystemStripeRequest{
					Type:            cardpayment.PaymentType(paymentType),
					RedirectSuccess: request.GetRedirectSuccess(),
					RedirectCancel:  request.GetRedirectError(),
				},
			},
		},
	})
	if err != nil {
		return nil, err
	}

	return &offer.PaymentResponse{
		Stripe: &offer.PaymentStripeResponse{
			Id:           resp.GetData().GetStripe().GetStripeId(),
			ClientSecret: resp.GetData().GetStripe().GetClientSecret(),
		},
	}, nil
}

func (server *serverImpl) awaitPaymentConfirm(ctx context.Context, off *model.OfferModel) error {
	var newstate offer.OfferState

	switch off.State {
	case offer.OfferState_HOLDING_DEPOSIT:
		newstate = offer.OfferState_AWAITING_CONFIRM
	case offer.OfferState_SIGNED:
		newstate = offer.OfferState_AWAITING_CONFIRM_RENT
	case offer.OfferState_RENTPAID:
		newstate = offer.OfferState_AWAITING_CONFIRM_EXTRAS
	default:
		return nil
	}

	err := off.SetInvoiceAndUpdate(ctx)
	if err != nil {
		return err
	}

	bgctx := grpcctx.ValuesBackgroundContext(ctx)

	_, err = server.bpms.NewPublishMessageCommand().
		MessageName("deposit_paid").
		CorrelationKey(off.ID.Hex()).
		TimeToLive(status.BPMTimeToLive).
		Send(bgctx)
	if err != nil {
		return err
	}

	return status.SetOfferState(ctx, off, newstate)
}

func (server *serverImpl) ListFilter(
	ctx context.Context,
	request *offer.ListFilterRequest,
) (*offer.ListFilterResponse, error) {
	params := request.GetParamsValue()
	offers, total, err := model.FilterOffers(ctx, params)
	if err != nil {
		return nil, err
	}

	var ofs []*offer.Offer
	for _, o := range offers {
		ofs = append(ofs, o.RenderOfferModel(ctx, ""))
	}

	return &offer.ListFilterResponse{
		Offers: ofs,
		Total:  int32(total),
	}, nil
}

func (server *serverImpl) DeleteDocument(
	ctx context.Context,
	req *offer.DeleteDocumentRequest,
) (*offer.DeleteDocumentResponse, error) {
	// offs, err := model.GetOffersByUser(ctx, req.GetUser())
	// if ((err != nil) || (len(offs) == 0)) && (req.GetAdminSite()) {
	//	offs, err = model.GetOffersByDocument(ctx, req.GetId())
	// }
	// if err != nil {
	//	return &offer.DeleteDocumentResponse{
	//		Deleted: false,
	//		Error:   false,
	//	}, err
	// }
	//
	// deleted := false
	// for _, off := range offs {
	//	found := false
	//	for _, tenant := range off.Tenants {
	//		var iDocs []string
	//		for _, iDoc := range tenant.IdentityDocuments {
	//			if iDoc == req.GetId() {
	//				found = true
	//				continue
	//			}
	//			iDocs = append(iDocs, iDoc)
	//		}
	//		tenant.IdentityDocuments = iDocs
	//
	//		var billDocs []string
	//		for _, billDoc := range tenant.BillDocuments {
	//			if billDoc == req.GetId() {
	//				found = true
	//				continue
	//			}
	//			billDocs = append(billDocs, billDoc)
	//		}
	//		tenant.BillDocuments = billDocs
	//	}
	//	if found {
	//		err = model.UpdateOffer(ctx, off)
	//		if err != nil {
	//			return &offer.DeleteDocumentResponse{
	//				Deleted: false,
	//				Error:   true,
	//			}, err
	//		}
	//		deleted = true
	//	}
	// }

	return nil, nil
}

func (server *serverImpl) DeleteOffersByProperty(
	ctx context.Context,
	request *offer.DeleteOffersByPropertyRequest,
) (*offer.DeleteOffersByPropertyResponse, error) {
	_, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id:   request.GetProperty(),
		User: request.GetUser(),
	})
	if err == nil {
		return &offer.DeleteOffersByPropertyResponse{
			Success:     false,
			Description: "You cannot delete an offer with an existing property",
		}, err
	}

	err = model.RemoveOffersByProperty(ctx, request.Property, request.User)
	if err != nil {
		return &offer.DeleteOffersByPropertyResponse{Success: false}, err
	}

	return &offer.DeleteOffersByPropertyResponse{Success: true}, err
}

func (server *serverImpl) FinishOffersByProperty(
	ctx context.Context,
	request *offer.FinishOffersByPropertyRequest,
) (*offer.FinishOffersByPropertyResponse, error) {
	offers, err := model.DeleteFinishOffersByProperty(ctx, request.GetProperty(), request.GetUser())
	if err != nil {
		return &offer.FinishOffersByPropertyResponse{Success: false}, err
	}

	for _, m := range offers {
		synchronization.ServiceOfferSync(ctx, m)
	}

	return &offer.FinishOffersByPropertyResponse{Success: true}, err
}

func (server *serverImpl) RenderCsv(
	ctx context.Context,
	request *offer.RenderCsvRequest,
) (*offer.RenderCsvResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer())
	if err != nil {
		return nil, err
	}
	old, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return nil, err
	}

	if old.Conditions == nil {
		old.Conditions = &model.OfferConditionsModel{}
	}

	if !old.DepositReplacement() || !IsProduction {
		return &offer.RenderCsvResponse{
			Success: true,
		}, nil
	}

	fid, err := status.RenderCSV(ctx, old, request.GetNeedToSend())
	if err != nil {
		return nil, err
	}

	return &offer.RenderCsvResponse{
		Success: true,
		File:    fid,
	}, nil
}

//nolint:revive // should be renamed to DeleteOfferByID but defined by gRPC API interface
func (server *serverImpl) DeleteOfferById(
	ctx context.Context,
	request *offer.DeleteOfferByIdRequest,
) (*offer.DeleteOfferByIdResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetId())
	if err != nil {
		return nil, err
	}
	result, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	if result.Deleted {
		return &offer.DeleteOfferByIdResponse{
			Success: false,
			Errors:  []offer.OfferError{offer.OfferError_OFFER_ALREADY_DELETED},
		}, nil
	}

	err = model.DeleteOffer(ctx, result, request.GetUser())
	if err != nil {
		return nil, err
	}

	return &offer.DeleteOfferByIdResponse{
		Success: true,
	}, nil
}

//nolint:revive // should be renamed to RestoreOfferByID but defined by gRPC API interface
func (server *serverImpl) RestoreOfferById(
	ctx context.Context,
	request *offer.RestoreOfferByIdRequest,
) (*offer.RestoreOfferByIdResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetId())
	if err != nil {
		return nil, err
	}
	result, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	if !result.Deleted {
		return &offer.RestoreOfferByIdResponse{
			Success: false,
			Errors:  []offer.OfferError{offer.OfferError_OFFER_NOT_DELETED},
		}, nil
	}

	err = model.RestoreOffer(ctx, result, request.GetUser())
	if err != nil {
		return nil, err
	}

	return &offer.RestoreOfferByIdResponse{
		Success: true,
	}, nil
}

// Updating the results of tenant checks by the identity service.
func (server *serverImpl) TenantCheckUpdate(
	ctx context.Context,
	request *offer.TenantCheckUpdateRequest,
) (response *offer.TenantCheckUpdateResponse, err error) {
	id, err := primitive.ObjectIDFromHex(request.Offer)
	if err != nil {
		return nil, err
	}
	m, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return nil, err
	}

	offerLog := model.NewOfferLog(ctx, m, "")

	defer func() {
		if err == nil {
			offerLog.LogDiff()
		}
	}()

	for _, t := range m.Tenants {
		if t.Check == request.Check {
			t.Risk = inferRisk(request.RiskScore)
			if request.PdfId != "" {
				t.RiskDocument = util.PtrString(request.GetPdfId())

				go createRiskDocument(
					context.Background(),
					m.ID.Hex(),
					m.LandlordUser(),
					t.User,
					util.JoinNonblankStrings(
						" ",
						util.UnPtrString(t.FirstName),
						util.UnPtrString(t.LastName),
					),
					util.UnPtrString(t.EmailPrimary),
					m.Property,
					request.GetPdfId(),
				)
			}
			t.RentprofileStatus = util.PtrString(request.Status)
			t.RiskURL = util.PtrString(request.ProfileLink)
			t.RentprofileCreatedAt = util.PtrTime(time.Unix(request.GetInviteDate(), 0))
			if util.UnPtrString(t.RiskDocument) != "" && util.UnPtrString(t.RentprofileStatus) == completeStatus {
				t.CheckNeeded = false
			}
		}
	}

	for _, g := range m.Guarantors {
		if g.Check == request.Check {
			g.Risk = inferRisk(request.RiskScore)
			if request.PdfId != "" {
				g.RiskDocument = util.PtrString(request.PdfId)
			}
			g.RentprofileStatus = util.PtrString(request.Status)
			g.RiskURL = util.PtrString(request.ProfileLink)
			g.RentprofileCreatedAt = util.PtrTime(time.Unix(request.GetInviteDate(), 0))
			if util.UnPtrString(g.RiskDocument) != "" && util.UnPtrString(g.RentprofileStatus) == completeStatus {
				g.CheckNeeded = false
			}
		}
	}

	if !m.ActiveChecks() {
		status.ExportClosecomOpportunity(ctx, m, closecom.OfferStatus_OFFER_STATUS_CONFIRMING_REFERENCE)
	}

	err = model.UpdateOffer(ctx, m, model.WithForceNotificationUpdate())
	if err != nil {
		return nil, err
	}

	return &offer.TenantCheckUpdateResponse{Success: true}, nil
}

// SetRiskReport forces to set risk document value (Rentprofile report file ID) for tenant or guarantor.
func (server *serverImpl) SetRiskReport(
	ctx context.Context,
	request *offer.SetRiskReportRequest,
) (*offer.SetRiskReportResponse, error) {
	switch {
	case request.GetSenderId() == "":
		return nil, fmt.Errorf("%w: empty sender ID", grpcerr.InvalidData)
	case request.GetUserEmail() == "":
		return nil, fmt.Errorf("%w: empty user primary email", grpcerr.InvalidData)
	case request.GetFileId() == "":
		return nil, fmt.Errorf("%w: empty report file ID", grpcerr.InvalidData)
	}

	logger := log.WithFields(log.Fields{
		"action":     "SetRiskReport",
		"sender":     request.GetSenderId(),
		"user_email": request.GetUserEmail(),
		"file":       request.GetFileId(),
		"offer":      request.GetOfferId(),
	})
	logger.Info("started")

	// TODO Check sender is admin

	var modified bool
	off, err := model.GetOfferByHexID(ctx, request.GetOfferId())
	if err != nil {
		logger.WithError(err).Error("failed to get offer")

		return nil, err
	}

	switch {
	case off.InProgress():
		logger.Error("offer is in progress")

		return nil, fmt.Errorf("%w: disallowed by offer state", grpcerr.InvalidState)
	case off.Deleted:
		logger.Error("offer deleted")

		return nil, fmt.Errorf("%w: offer deleted", grpcerr.InvalidState)
	}

	info, err := connections.DocumentService.FileInfo(ctx, &document.FileInfoRequest{
		Id: request.GetFileId(),
	})
	if err != nil {
		logger.WithError(err).Error("failed to get report info")

		return nil, err
	}

	if ct := info.GetFile().GetMime(); ct != "application/pdf" {
		logger.Errorf("uploaded report has unexpected content-type (%s)", ct)

		return nil, fmt.Errorf("%w: report has unexpected content-type (%s)", grpcerr.InvalidState, ct)
	}

	if l := info.GetFile().GetSize(); l == 0 {
		logger.Errorf("uploaded report has zero size")

		return nil, fmt.Errorf("%w: report has zero size", grpcerr.InvalidState)
	}

	offerDiff := model.NewOfferLog(ctx, off, "")

	defer func() {
		if modified {
			offerDiff.LogDiff()
		}
	}()

	found := false
	if t := off.TenantMatch(func(t *model.TenantOfferModel) bool {
		return t.EmailPrimary != nil && *t.EmailPrimary == request.GetUserEmail()
	}); t != nil {
		found = true
		t.RiskDocument = &request.FileId

		go createRiskDocument(
			context.Background(),
			off.ID.Hex(),
			off.LandlordUser(),
			t.User,
			util.JoinNonblankStrings(
				" ",
				util.UnPtrString(t.FirstName),
				util.UnPtrString(t.LastName),
			),
			util.UnPtrString(t.EmailPrimary),
			off.Property,
			request.GetFileId(),
		)

		if request.GetRisk() != nil {
			t.Risk = request.GetRisk().GetValue()
		}

		if request.GetRiskUrl() != nil {
			t.RiskURL = util.PtrString(request.GetRiskUrl().GetValue())
		}

		t.CheckNeeded = false
	}

	if g := off.GuarantorMatch(func(g *model.GuarantorOfferModel) bool {
		return g.EmailPrimary != nil && *g.EmailPrimary == request.GetUserEmail()
	}); g != nil {
		found = true
		g.RiskDocument = &request.FileId

		if request.GetRisk() != nil {
			g.Risk = request.GetRisk().GetValue()
		}

		if request.GetRiskUrl() != nil {
			g.RiskURL = util.PtrString(request.GetRiskUrl().GetValue())
		}

		g.CheckNeeded = false
	}

	if !found {
		logger.Error("not applicable: tenant/guarantor not found")

		return nil, fmt.Errorf("%w: tenant/guarantor not found", grpcerr.InvalidState)
	}

	if !off.ActiveChecks() {
		status.ExportClosecomOpportunity(ctx, off, closecom.OfferStatus_OFFER_STATUS_CONFIRMING_REFERENCE)
	}

	if err = model.UpdateOffer(ctx, off); err != nil {
		logger.WithError(err).Error("failed to update offer")

		return nil, err
	}

	logger.Info("completed")
	modified = true

	return &offer.SetRiskReportResponse{}, nil
}

func createRiskDocument(ctx context.Context, offerID, landlord, tenant, tname, temail, property, id string) {
	if id == "" {
		return
	}

	_, err := connections.DocumentService.Create(ctx, &document.CreateRequest{
		User:        landlord,
		Property:    property,
		File:        id,
		Expiration:  nil,
		Description: "Referencing Report for " + tname,
		Code:        "",
		Amount:      0,
		ProductId:   "",
		SkipNotify:  false,
		IssuedAt:    nil,
		TypeCode:    document.TypeReferencingReport,
		Attributes: map[string]string{
			"referencing_user_id": tenant,
			"referencing_name":    tname,
			"referencing_email":   temail,
			"offer":               offerID,
		},
	})
	if err != nil {
		log.WithError(err).Error("creating referncing check document")
	}
}

// AddContractClause adds new contract clause (For MVP only landlord has the opportunity to add clauses)
func (server *serverImpl) AddContractClause(
	ctx context.Context,
	request *offer.AddContractClauseRequest,
) (response *offer.AddContractClauseResponse, err error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer())
	if err != nil {
		return nil, err
	}
	off, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	offerLog := model.NewOfferLog(ctx, off, request.GetUser())

	if off.Landlord.User != request.GetUser() {
		return &offer.AddContractClauseResponse{
			Errors: []offer.OfferError{offer.OfferError_USER_NO_ACCESS},
		}, nil
	}

	if server.checkOfferStatusForUpdate(off) {
		return &offer.AddContractClauseResponse{
			Errors: []offer.OfferError{offer.OfferError_WRONG_CURRENT_OFFER_STATUS},
		}, nil
	}

	off.Clause = append(off.Clause, &model.ClauseModel{
		ID:        primitive.NewObjectID().Hex(),
		Text:      request.GetClauseText(),
		State:     offer.ClauseState_CLAUSE_PENDING,
		CreatedAt: time.Now(),
	})

	err = model.UpdateOffer(ctx, off)
	if err != nil {
		return nil, err
	}

	offerLog.LogDiff()

	// notify tenant
	notifyClause(
		ctx,
		off,
		notification.Role_ROLE_TENANT,
		off.TenantUser(),
		off.LandlordUser(),
		notification.Action_ACTION_CREATE,
	)

	return &offer.AddContractClauseResponse{
		Success: true,
		Offer:   off.RenderOfferModel(ctx, request.GetUser()),
	}, nil
}

// UpdateContractClauses edits a text of a contract clause / remove clause from clauses list
// (for MVP only landlord has the opportunity to edit a text/remove a clause).
func (server *serverImpl) UpdateContractClauses(
	ctx context.Context,
	request *offer.UpdateContractClausesRequest,
) (response *offer.UpdateContractClausesResponse, err error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer())
	if err != nil {
		return nil, err
	}
	off, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	offerLog := model.NewOfferLog(ctx, off, request.GetUser())

	if off.Landlord.User != request.GetUser() {
		return &offer.UpdateContractClausesResponse{
			Errors: []offer.OfferError{offer.OfferError_USER_NO_ACCESS},
		}, nil
	}

	if len(request.GetClause()) == 0 {
		// offer had a clauses
		needNotifyTenant := len(off.Clause) > 0

		off.Clause = make([]*model.ClauseModel, 0)

		// TODO update draft contract
		err = model.UpdateOffer(ctx, off)
		if err != nil {
			return nil, err
		}

		// notify tenant
		if needNotifyTenant {
			notifyClause(
				ctx,
				off,
				notification.Role_ROLE_TENANT,
				off.TenantUser(),
				off.LandlordUser(),
				notification.Action_ACTION_REMOVE,
			)
		}

		return &offer.UpdateContractClausesResponse{
			Success: true,
			Offer:   off.RenderOfferModel(ctx, request.User),
		}, nil
	}

	if server.checkOfferStatusForUpdate(off) {
		return &offer.UpdateContractClausesResponse{
			Errors: []offer.OfferError{offer.OfferError_WRONG_CURRENT_OFFER_STATUS},
		}, nil
	}

	newClauses := make([]*model.ClauseModel, 0)
	deleted := make([]*model.ClauseModel, 0)
	for _, c := range off.Clause {
		found := false
		for _, nc := range request.GetClause() {
			if c.ID == nc.GetId() {
				found = true
				if c.State == offer.ClauseState_CLAUSE_PENDING {
					c.Text = nc.GetText()
				}
				newClauses = append(newClauses, c)
			}
		}

		if !found {
			deleted = append(deleted, c)
		}
	}

	off.Clause = newClauses

	updateDraftContract := false

	for _, d := range deleted {
		if d.State == offer.ClauseState_CLAUSE_ACCEPTED {
			updateDraftContract = true
			break
		}
	}

	if updateDraftContract {
		err = status.GenerateDraftContract(ctx, off)
		if err != nil {
			return nil, err
		}
	}

	err = model.UpdateOffer(ctx, off)
	if err != nil {
		return nil, err
	}

	offerLog.LogDiff()

	// notify tenant
	notifyClause(
		ctx,
		off,
		notification.Role_ROLE_TENANT,
		off.TenantUser(),
		off.LandlordUser(),
		notification.Action_ACTION_UPDATE,
	)

	return &offer.UpdateContractClausesResponse{
		Success: true,
		Offer:   off.RenderOfferModel(ctx, request.GetUser()),
	}, nil
}

func notifyClause(
	ctx context.Context,
	off *model.OfferModel,
	role notification.Role,
	targetID, actorID string,
	action notification.Action,
) {
	targets := []*notification.Target{
		{
			User: targetID,
			Role: role,
		},
	}

	switch action {
	case notification.Action_ACTION_UPDATE, notification.Action_ACTION_CREATE,
		notification.Action_ACTION_ACCEPT, notification.Action_ACTION_REJECT:
		targets = append(targets, &notification.Target{
			Role: notification.Role_ROLE_ADMIN,
		})
	}

	_, err := connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity:  notification.Entity_ENTITY_OFFER_CLAUSE,
		Action:  action,
		Targets: targets,
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: off.Property,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: off.ID.Hex(),
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
				EmbedValue: actorID,
			},
		},
	})
	if err != nil {
		log.WithError(err)
	}
}

// ChangeStateContractClauses - accept or deny a proposed contract clause
// (for MVP only tenant has the opportunity for accepting/denying).
func (server *serverImpl) ChangeStateContractClauses(
	ctx context.Context,
	request *offer.ChangeStateContractClausesRequest,
) (response *offer.ChangeStateContractClausesResponse, err error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer())
	if err != nil {
		return nil, err
	}
	off, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	offerLog := model.NewOfferLog(ctx, off, request.GetUser())

	if off.Tenants[0].User != request.GetUser() {
		return &offer.ChangeStateContractClausesResponse{
			Errors: []offer.OfferError{offer.OfferError_USER_NO_ACCESS},
		}, nil
	}

	if server.checkOfferStatusForUpdate(off) {
		return &offer.ChangeStateContractClausesResponse{
			Errors: []offer.OfferError{offer.OfferError_WRONG_CURRENT_OFFER_STATUS},
		}, nil
	}

	finded := false
	if len(off.Clause) > 0 {
		for _, c := range off.Clause {
			if c.ID == request.ClauseId {
				c.State = request.State
				finded = true
				break
			}
		}

		if !finded {
			return &offer.ChangeStateContractClausesResponse{
				Errors: []offer.OfferError{offer.OfferError_NO_SUCH_CLAUSE},
			}, nil
		}
	} else {
		return &offer.ChangeStateContractClausesResponse{
			Errors: []offer.OfferError{offer.OfferError_NO_SUCH_CLAUSE},
		}, nil
	}

	if request.State == offer.ClauseState_CLAUSE_ACCEPTED {
		err = status.GenerateDraftContract(ctx, off)
		if err != nil {
			return nil, err
		}
	}

	err = model.UpdateOffer(ctx, off)
	if err != nil {
		return nil, err
	}

	offerLog.LogDiff()

	var action notification.Action

	switch request.GetState() {
	case offer.ClauseState_CLAUSE_ACCEPTED:
		action = notification.Action_ACTION_ACCEPT
	case offer.ClauseState_CLAUSE_DECLINED:
		action = notification.Action_ACTION_REJECT
	}

	// notification to LL -> tenant has accepted or rejected a clause
	notifyClause(
		ctx,
		off,
		notification.Role_ROLE_LANDLORD,
		off.LandlordUser(),
		off.TenantUser(),
		action,
	)

	return &offer.ChangeStateContractClausesResponse{
		Success: true,
		Offer:   off.RenderOfferModel(ctx, request.GetUser()),
	}, nil
}

// ProgressToSigning - progress to signing after reviewing and clauses.
// Works only in CONTRACT_REVIEW_AND_CLAUSES status and if there are not clauses in the PENDING status.
func (server *serverImpl) ProgressToSigning(
	ctx context.Context,
	request *offer.ProgressToSigningRequest,
) (*offer.ProgressToSigningResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer())
	if err != nil {
		return nil, err
	}

	off, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	if off.ReadyForSigning == nil {
		off.ReadyForSigning = &model.OfferReadySigningModel{}
	}

	if off.Conditions == nil {
		return nil, grpcerr.New(grpcerr.InvalidData, "no offer conditions provided")
	}

	switch {
	case off.Landlord.User == request.GetUser():
		off.ReadyForSigning.Landlord = true
		switch {
		case off.Conditions.TermMonths == nil:
			return nil, grpcerr.New(grpcerr.InvalidData, "no tenancy length provided")
		case off.Conditions.FirstRentPaymentDate == nil:
			return nil, grpcerr.New(grpcerr.InvalidData, "no first rent payment date provided")
		case off.Conditions.RentPaymentDay == nil:
			return nil, grpcerr.New(grpcerr.InvalidData, "no rent payment day provided")
		case off.Conditions.MoveinDate == nil:
			return nil, grpcerr.New(grpcerr.InvalidData, "no movein date provided")
		}
	case off.Tenants[0].User == request.GetUser():
		switch {
		case off.Conditions.Price == nil:
			return nil, grpcerr.New(grpcerr.InvalidData, "no offer price provided")
		case off.Conditions.DepositType == 0:
			return nil, grpcerr.New(grpcerr.InvalidData, "no offer deposit type selected")
		case off.Conditions.PaymentFrequency == nil:
			return nil, grpcerr.New(grpcerr.InvalidData, "no rent payment frequency provided")
		}
		off.ReadyForSigning.Tenant = true
	default:
		return &offer.ProgressToSigningResponse{
			Errors: []offer.OfferError{offer.OfferError_USER_NO_ACCESS},
		}, nil
	}

	if server.checkOfferStatusForUpdate(off) {
		return &offer.ProgressToSigningResponse{
			Errors: []offer.OfferError{offer.OfferError_WRONG_CURRENT_OFFER_STATUS},
		}, nil
	}

	ok := true
	for _, c := range off.Clause {
		if c.State == offer.ClauseState_CLAUSE_PENDING {
			ok = false
			break
		}
	}
	if !ok {
		return &offer.ProgressToSigningResponse{
			Errors: []offer.OfferError{offer.OfferError_CLAUSES_IN_PENDING},
		}, nil
	}

	err = model.UpdateOffer(ctx, off)
	if err != nil {
		return nil, err
	}

	if !off.ReadyForSigning.Tenant || !off.ReadyForSigning.Landlord {
		return &offer.ProgressToSigningResponse{
			Errors: []offer.OfferError{offer.OfferError_NO_UPDATES},
		}, nil
	}

	contract, err := server.createContract(ctx, off)
	if err != nil {
		return nil, err
	}

	bgctx := grpcctx.ValuesBackgroundContext(ctx)

	msg, err := server.bpms.NewPublishMessageCommand().
		MessageName("contract-prepared").
		CorrelationKey(id.Hex()).
		TimeToLive(status.BPMTimeToLive).
		VariablesFromMap(map[string]interface{}{
			"contract": contract,
		})
	if err != nil {
		return nil, err
	}

	_, err = msg.Send(bgctx)
	if err != nil {
		return nil, err
	}

	return &offer.ProgressToSigningResponse{
		Success: true,
	}, nil
}

//nolint:gocognit,gocyclo
func (server *serverImpl) createContract(ctx context.Context, off *model.OfferModel) (string, error) {
	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id:   off.Property,
		User: off.LandlordUser(),
	})
	if err != nil {
		return "", err
	}
	landlord, err := connections.AccountService.Get(ctx, &account.GetRequest{
		Id: off.Landlord.User,
	})
	if err != nil {
		return "", err
	}
	cs := struct {
		Landlord *account.Account   `json:"landlord"`
		Offer    *offer.Offer       `json:"offer"`
		Property *property.Property `json:"property"`
		Now      string             `json:"now"`
	}{
		Now:      time.Now().Format(time.RFC1123),
		Landlord: landlord.GetAccount(),
		Offer:    off.RenderOfferModel(ctx, ""),
		Property: prop.GetProperty(),
	}

	bs, err := bind.ToJSON(cs, &bind.Config{
		GetOmitEmpty: false,
		SetOmitEmpty: false,
	})
	if err != nil {
		return "", err
	}

	fileID, err := status.CreateContract(ctx, bs, false)
	if err != nil {
		return "", err
	}

	off.Contract = fileID
	off.ContractCreatedAt = time.Now()
	err = model.UpdateOffer(ctx, off)
	if err != nil {
		return "", err
	}

	months := util.UnPtrInt32(off.Conditions.TermMonths)

	expiration := timestamppb.New(util.UnPtrTime(off.Conditions.MoveinDate).AddDate(0, int(months), 0))

	_, err = connections.DocumentService.Create(ctx, &document.CreateRequest{
		Type:       document.DocumentType_CONTRACT,
		User:       prop.GetProperty().GetPeople().GetLandlord().GetValue(),
		Property:   prop.GetProperty().GetId(),
		File:       fileID,
		Expiration: expiration,
	})
	if err != nil {
		return "", err
	}

	for _, t := range off.Tenants {
		if len(t.User) > 0 {
			_, err = connections.DocumentService.Create(ctx, &document.CreateRequest{
				Type:       document.DocumentType_CONTRACT,
				User:       t.User,
				Property:   prop.GetProperty().GetId(),
				File:       fileID,
				Expiration: expiration,
			})
			if err != nil {
				return "", err
			}
		}
	}

	loc := prop.GetProperty().GetLocation()
	addr := loc.FormatAddress()

	var signers []*signature.Signer

	landlordDet := landlord.GetAccount().GetDetailsValue()
	var landlordDocs []string

	docsLL, err := connections.DocumentService.List(
		ctx,
		&document.ListRequest{
			User: landlord.GetAccount().GetIdValue(),
		},
	)
	if err != nil {
		return "", err
	}
	for _, doc := range docsLL.Documents {
		if doc.GetTypeCode() == document.TypeIdentity && doc.GetState() == document.DocumentState_ACTIVE {
			landlordDocs = append(landlordDocs, doc.Id)
		}
	}

	isDraft := true
	if off.Approve != nil && off.Approve.Approved {
		isDraft = false
	}

	isDeposit := true
	if off.Conditions != nil {
		isDeposit = off.Conditions.DepositType == offer.DepositType_DEPOSIT_TRADITIONAL_5WEEK
	}

	isGuarantor := len(off.Guarantors) > 0

	breakClause := off.Conditions.BreakClause != nil && *off.Conditions.BreakClause > 0

	var formFields []*signature.DocumentFormField
	page := getPageAST(isDeposit, isGuarantor)

	if !breakClause {
		page--
	}

	// landlord
	signersCount := 0
	signers = append(signers, &signature.Signer{
		Name:        landlordDet.GetFirstNameValue() + " " + landlordDet.GetLastNameValue(),
		Email:       landlord.GetAccount().GetContactValue().GetEmailPrimaryValue(),
		IdDocuments: landlordDocs,
	})
	formFields = append(formFields, &signature.DocumentFormField{
		Id:       fmt.Sprintf("signature0to%d", signersCount),
		Type:     "signature",
		X:        SignX,
		Y:        int32(SignLineHeight*(signersCount+1) + SignMargin),
		Width:    SignWidth,
		Height:   SignHeight,
		Required: true,
		Signer:   int32(signersCount),
		Page:     int32(page),
	})

	for _, l := range off.Landlords {
		signers = append(signers, &signature.Signer{
			Name:  util.UnPtrString(l.FirstName) + " " + util.UnPtrString(l.LastName),
			Email: util.UnPtrString(l.EmailPrimary),
		})

		signersCount++

		formFields = append(formFields, &signature.DocumentFormField{
			Id:       fmt.Sprintf("signature0to%d", signersCount),
			Type:     "signature",
			X:        SignX,
			Y:        int32(SignLineHeight*(signersCount+1) + SignMargin),
			Width:    SignWidth,
			Height:   SignHeight,
			Required: true,
			Signer:   int32(signersCount),
			Page:     int32(page),
		})
	}

	// tenants
	for _, t := range off.Tenants {
		signers = append(signers, &signature.Signer{
			Name:  util.UnPtrString(t.FirstName) + " " + util.UnPtrString(t.LastName),
			Email: util.UnPtrString(t.EmailPrimary),
		})

		signersCount++

		formFields = append(formFields, &signature.DocumentFormField{
			Id:       fmt.Sprintf("signature0to%d", signersCount),
			Type:     "signature",
			X:        SignX,
			Y:        int32(SignLineHeight*(signersCount+1) + SignMargin),
			Width:    SignWidth,
			Height:   SignHeight,
			Required: true,
			Signer:   int32(signersCount),
			Page:     int32(page),
		})
	}

	// guarantors
	for _, g := range off.Guarantors {
		signers = append(signers, &signature.Signer{
			Name: util.UnPtrString(g.FirstName) +
				" " +
				util.UnPtrString(g.LastName),
			Email: util.UnPtrString(g.EmailPrimary),
		})
		signersCount++
		formFields = append(formFields, &signature.DocumentFormField{
			Id:       fmt.Sprintf("signature0to%d", signersCount),
			Type:     "signature",
			X:        SignX,
			Y:        int32(SignLineHeight*(signersCount+1) + SignMargin),
			Width:    SignWidth,
			Height:   SignHeight,
			Required: true,
			Signer:   int32(signersCount),
			Page:     int32(page),
		})
	}

	if !isDeposit {
		signersCount = 0
		landlordPage := getSignPageLandlord(isGuarantor)
		if !breakClause {
			landlordPage--
		}
		formFields = append(formFields, &signature.DocumentFormField{
			Id:       fmt.Sprintf("signature1to%d", signersCount),
			Type:     "signature",
			X:        SignX,
			Y:        int32(SignLineHeight*(signersCount+2) + SignMargin),
			Width:    SignWidth,
			Height:   SignHeight,
			Required: true,
			Signer:   int32(signersCount),
			Page:     landlordPage,
		})

		for range off.Landlords {
			signersCount++
			formFields = append(formFields, &signature.DocumentFormField{
				Id:       fmt.Sprintf("signature1to%d", signersCount),
				Type:     "signature",
				X:        SignX,
				Y:        int32(SignLineHeight*(signersCount+2) + SignMargin),
				Width:    SignWidth,
				Height:   SignHeight,
				Required: true,
				Signer:   int32(signersCount),
				Page:     landlordPage,
			})
		}

		tenantPage := getSignPageTenant(isGuarantor)
		if !breakClause {
			tenantPage--
		}
		for range off.Tenants {
			signersCount++
			formFields = append(formFields, &signature.DocumentFormField{
				Id:       fmt.Sprintf("signature2to%d", signersCount),
				Type:     "signature",
				X:        SignX,
				Y:        int32(SignLineHeight*(signersCount+1) + SignMargin),
				Width:    SignWidth,
				Height:   SignHeight,
				Required: true,
				Signer:   int32(signersCount),
				Page:     tenantPage,
			})
		}
	}

	docForm := &signature.DocumentForm{
		FormFields: formFields,
	}
	var formDocuments []*signature.DocumentForm
	if !isDraft {
		formDocuments = append(formDocuments, docForm)
	}

	sig, err := connections.SignatureService.SigningCreate(ctx, &signature.SigningCreateRequest{
		Title:          "Mashroom contract",
		Subject:        "Signing contract #" + off.ID.Hex() + " on property in " + addr,
		Message:        "Please review and sign following contract",
		Signers:        signers,
		DocumentForms:  formDocuments,
		DocumentFileId: fileID,
		Metadata: &signature.SigningMetadata{
			OfferId:    off.ID.Hex(),
			PropertyId: prop.GetProperty().GetId(),
			ViewingId:  off.Viewing,
		},
	})
	if err != nil {
		return "", err
	}

	off.SignatureRequestID = sig.GetSigning().GetSignatureRequestId()
	off.Landlord.Signature = append(off.Landlord.Signature, &model.SignatureModel{
		State: offer.SignStatusState_AWAITING_SIGNATURE,
		Time:  time.Now(),
	})
	for _, t := range off.Tenants {
		t.Signature = append(t.Signature, &model.SignatureModel{
			State: offer.SignStatusState_AWAITING_SIGNATURE,
			Time:  time.Now(),
		})
	}

	for _, g := range off.Guarantors {
		g.Signature = append(g.Signature, &model.SignatureModel{
			State: offer.SignStatusState_AWAITING_SIGNATURE,
			Time:  time.Now(),
		})
	}

	err = model.UpdateOffer(ctx, off)
	if err != nil {
		return "", err
	}

	err = status.SetOfferState(ctx, off, offer.OfferState_SIGNING)
	if err != nil {
		return "", err
	}

	return fileID, nil
}

// RenewTenancy renews tenancy. Creates new offer based on current one.
func (server *serverImpl) RenewTenancy(
	ctx context.Context,
	request *offer.RenewTenancyRequest,
) (*offer.RenewTenancyResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetOldOffer())
	if err != nil {
		return nil, err
	}
	off, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	off.RenewTenancy()
	err = model.CreateOffer(ctx, off)
	if err != nil {
		return nil, err
	}

	synchronization.ServiceOfferSync(ctx, off)

	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{Id: off.Property})
	if err != nil {
		return nil, err
	}

	_, err = connections.ZeebeExporter.Start(ctx, &zeebe.StartRequest{
		Remote:  connections.SelfRemote,
		Process: listing.PropertyOfferProcessID,
		Payload: &structpb.Struct{
			Fields: map[string]*structpb.Value{
				"offer": {Kind: &structpb.Value_StringValue{
					StringValue: off.ID.Hex(),
				}},
				"property": {Kind: &structpb.Value_StringValue{
					StringValue: prop.GetProperty().GetId(),
				}},
				"landlord": {Kind: &structpb.Value_StringValue{
					StringValue: prop.GetProperty().GetPeople().GetLandlord().GetValue(),
				}},
				"manager": {Kind: &structpb.Value_StringValue{
					StringValue: prop.GetProperty().GetPeople().GetManager().GetValue(),
				}},
			},
		},
	})
	if err != nil {
		return nil, err
	}

	return &offer.RenewTenancyResponse{
		Success: true,
		Offer:   off.RenderOfferModel(ctx, ""),
	}, nil
}

// AcceptUpdate accepts updates of already accepted offer
func (server *serverImpl) AcceptUpdate(
	ctx context.Context,
	request *offer.AcceptUpdateRequest,
) (response *offer.AcceptUpdateResponse, err error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer())
	if err != nil {
		return nil, err
	}
	off, err := model.GetOffer(ctx, id, request.GetUser())
	if err != nil {
		return nil, err
	}

	offerLog := model.NewOfferLog(ctx, off, request.GetUser())

	defer func() {
		if err != nil {
			return
		}

		offerLog.LogDiff()
	}()

	if off.Landlord.User != request.GetUser() {
		return &offer.AcceptUpdateResponse{
			Errors: []offer.OfferError{offer.OfferError_USER_NO_ACCESS},
		}, nil
	}

	if !off.IsLocked {
		return &offer.AcceptUpdateResponse{
			Errors: []offer.OfferError{offer.OfferError_NO_UPDATES},
		}, nil
	}

	bgctx := grpcctx.ValuesBackgroundContext(ctx)

	if off.State == offer.OfferState_HOLDING_DEPOSIT {
		_, err = server.bpms.NewPublishMessageCommand().
			MessageName("updates-accepted").
			CorrelationKey(id.Hex()).
			TimeToLive(status.BPMTimeToLive).
			Send(bgctx)
		if err != nil {
			return nil, err
		}

		off.IsLocked = false
		off.AcceptedAt = time.Now()
	}

	err = model.UpdateOffer(ctx, off)
	if err != nil {
		return nil, err
	}

	return &offer.AcceptUpdateResponse{
		Success: true,
	}, nil
}

// UpdateDraftContract - the offer draft contract updating by admin tools.
// Updating only in CONTRACT_REVIEW_AND_CLAUSES status.
func (server *serverImpl) UpdateDraftContract(
	ctx context.Context,
	request *offer.UpdateDraftContractRequest,
) (response *offer.UpdateDraftContractResponse, err error) {
	id, err := primitive.ObjectIDFromHex(request.Offer)
	if err != nil {
		return nil, err
	}
	off, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return nil, err
	}

	offerLog := model.NewOfferLog(ctx, off, "")

	if off.State == offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES {
		err = status.GenerateDraftContract(ctx, off)
		if err != nil {
			return nil, err
		}
		err = model.UpdateOffer(ctx, off)
		if err != nil {
			return nil, err
		}

		offerLog.LogDiff()

		return &offer.UpdateDraftContractResponse{
			Success: true,
			Offer:   off.RenderOfferModel(ctx, ""),
		}, nil
	}

	return &offer.UpdateDraftContractResponse{
		Success: false,
		Errors: []offer.OfferError{
			offer.OfferError_INVALID_OFFER_STATUS,
		},
	}, nil
}

// SkipReferencingStage to skip the offer referencing stage by admin tools
func (server *serverImpl) SkipReferencingStage(
	ctx context.Context,
	request *offer.SkipReferencingStageRequest,
) (*offer.SkipReferencingStageResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.Offer)
	if err != nil {
		return nil, err
	}

	off, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return nil, err
	}

	switch off.State {
	case offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT:
		err = status.SetOfferState(ctx, off, offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES)
		if err != nil {
			return nil, err
		}

		_, err := connections.InvoiceService.InvoiceCancel(ctx, &invoice.InvoiceCancelRequest{
			CorrelationType:   invoice.CorrelationTypeOffer,
			CorrelationId:     off.ID.Hex(),
			CorrelationTarget: invoice.CorrelationTargetReferenceCheck,
			User:              off.Landlord.User,
		})

		if err != nil && !errors.Is(err, grpcerr.NotFound) {
			return nil, err
		}

		bgctx := grpcctx.ValuesBackgroundContext(ctx)

		_, err = server.bpms.NewPublishMessageCommand().
			MessageName("proceed-without-checks").
			CorrelationKey(id.Hex()).
			TimeToLive(status.BPMTimeToLive).
			Send(bgctx)
		if err != nil {
			log.Error(err)
			return nil, err
		}

	case offer.OfferState_AWAITING_CHECKS, offer.OfferState_CHECKING:
		err = status.SetOfferState(ctx, off, offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES)
		if err != nil {
			return nil, err
		}

		bgctx := grpcctx.ValuesBackgroundContext(ctx)

		_, err = server.bpms.NewPublishMessageCommand().
			MessageName("checks-complete").
			CorrelationKey(id.Hex()).
			TimeToLive(status.BPMTimeToLive).
			Send(bgctx)
		if err != nil {
			log.Error(err)
			return nil, err
		}
	default:
		return &offer.SkipReferencingStageResponse{
			Success: false,
			Errors: []offer.OfferError{
				offer.OfferError_INVALID_OFFER_STATUS,
			},
		}, nil
	}

	err = model.UpdateOffer(ctx, off)
	if err != nil {
		return nil, err
	}

	return &offer.SkipReferencingStageResponse{
		Success: true,
		Offer:   off.RenderOfferModel(ctx, ""),
	}, nil
}

// ServiceRelatedProperties returns list of property IDs given user has relation to via offers
func (server *serverImpl) ServiceRelatedProperties(
	ctx context.Context,
	req *offer.ServiceRelatedPropertiesRequest,
) (*offer.ServiceRelatedPropertiesResponse, error) {
	res, err := model.ListRelatedOutgoingProperties(ctx, req.GetUser())
	if err != nil {
		return nil, err
	}

	return &offer.ServiceRelatedPropertiesResponse{
		PropertyIds: res,
	}, nil
}

// Payout - payment for landlord from platform
func (server *serverImpl) Payout(
	ctx context.Context,
	request *offer.PayoutRequest,
) (*offer.PayoutResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer())
	if err != nil {
		return nil, err
	}

	old, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return nil, err
	}

	inv, err := connections.InvoiceService.InvoiceGet(ctx, &invoice.InvoiceGetRequest{
		InvoiceId: request.GetInvoiceId(),
	})
	if err != nil {
		return nil, err
	}

	if inv.GetInvoice().GetStatus() != invoice.InvoiceStatus_INVOICE_OPEN {
		return nil, status.ErrInvalidState
	}

	resp, err := connections.InvoiceService.PaymentCreate(ctx, &invoice.PaymentCreateRequest{
		User:      old.Landlord.User,
		InvoiceId: inv.GetInvoice().GetInvoiceId(),
		System: &invoice.PaymentSystemDataRequest{
			Data: &invoice.PaymentSystemDataRequest_Tokenio{
				Tokenio: &invoice.PaymentSystemTokenioRequest{
					Description: fmt.Sprintf(
						"Payout for rent for property %v at mashroom.com",
						old.Property,
					),
					Purpose:         tokenio.PaymentPurpose_PAYOUT,
					RedirectSuccess: request.GetRedirectSuccess(),
					RedirectError:   request.GetRedirectError(),
				},
			},
		},
	})
	if err != nil {
		return nil, err
	}

	res := &offer.PayoutResponse{
		Tokenio: &offer.PaymentTokenioResponse{},
	}

	res.Tokenio.Url = resp.GetData().GetTokenio().GetUrl()
	res.Tokenio.Csrf = resp.GetData().GetTokenio().GetCsrf()

	return res, nil
}

func (server *serverImpl) ServiceTerminate(
	ctx context.Context,
	request *offer.ServiceTerminateRequest,
) (*offer.ServiceTerminateResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer())
	if err != nil {
		return nil, err
	}

	off, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return nil, err
	}

	if off.Conditions.MoveinDate == nil || off.Conditions.BreakClause == nil {
		return nil, grpcerr.New(grpcerr.InvalidState, "no break clause")
	}

	breakt := off.Conditions.MoveinDate.AddDate(0, int(*off.Conditions.BreakClause), 0)
	if breakt.After(time.Now()) {
		return nil, grpcerr.New(grpcerr.InvalidState, "break clause time in the future")
	}

	state := off.State

	endt := off.Conditions.MoveinDate.AddDate(0, int(*off.Conditions.TermMonths), 0)
	if endt.After(request.GetDate().AsTime()) {
		state = offer.OfferState_EARLY_TERMINATION
	}

	err = model.OfferTerminate(ctx, id, request.GetDate().AsTime(), state)
	if err != nil {
		return nil, err
	}

	return &offer.ServiceTerminateResponse{}, nil
}

func (server *serverImpl) ServiceCompleteCheck(
	ctx context.Context,
	request *offer.ServiceCompleteCheckRequest,
) (*offer.ServiceCompleteCheckResponse, error) {
	return nil, grpcerr.NotFound
}

// CancelSigningStage to return offer to contract review stage by admin tools
func (server *serverImpl) CancelSigningStage(
	ctx context.Context,
	request *offer.CancelSigningStageRequest,
) (*offer.CancelSigningStageResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.Offer)
	if err != nil {
		return nil, err
	}

	off, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return nil, err
	}

	switch off.State {
	case offer.OfferState_SIGNING:
		for _, c := range off.Clause {
			if c.State == offer.ClauseState_CLAUSE_ACCEPTED {
				c.State = offer.ClauseState_CLAUSE_PENDING
			}
		}
		err = status.SetOfferState(ctx, off, offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES)
		if err != nil {
			return nil, err
		}

		bgctx := grpcctx.ValuesBackgroundContext(ctx)

		_, err = server.bpms.NewPublishMessageCommand().
			MessageName("signing-canceled").
			CorrelationKey(id.Hex()).
			TimeToLive(status.BPMTimeToLive).
			Send(bgctx)
		if err != nil {
			log.Error(err)
			return nil, err
		}

		_, err = connections.SignatureService.SigningCancel(ctx, &signature.SigningCancelRequest{
			SignatureRequestId: off.SignatureRequestID,
		})
		if err != nil {
			log.Error(err)
			return nil, err
		}
		off.SignatureRequestID = ""
		off.ReadyForSigning.Landlord = false
		off.ReadyForSigning.Tenant = false
	default:
		return &offer.CancelSigningStageResponse{
			Success: false,
			Errors: []offer.OfferError{
				offer.OfferError_INVALID_OFFER_STATUS,
			},
		}, nil
	}

	err = model.UpdateOffer(ctx, off)
	if err != nil {
		return nil, err
	}

	return &offer.CancelSigningStageResponse{
		Success: true,
		Offer:   off.RenderOfferModel(ctx, ""),
	}, nil
}

// SetAdmin to set admin in offer model + in property and invoices associated with the offer
func (server *serverImpl) SetAdmin(
	ctx context.Context,
	request *offer.SetAdminRequest,
) (*offer.SetAdminResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetOffer())
	if err != nil {
		return nil, err
	}

	off, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return nil, err
	}

	off.Administrator = request.Admin

	err = model.UpdateOffer(ctx, off)
	if err != nil {
		return nil, err
	}

	_, err = connections.InvoiceService.ServiceSetAdmin(ctx, &invoice.ServiceSetAdminRequest{
		Offer: request.Offer,
		Admin: request.Admin,
	})
	if err != nil {
		return nil, err
	}

	return &offer.SetAdminResponse{}, nil
}

// ServiceSetAdmin to set admin in offer model
func (server *serverImpl) ServiceSetAdmin(
	ctx context.Context,
	request *offer.ServiceSetAdminRequest,
) (*offer.ServiceSetAdminResponse, error) {
	var offers []*model.OfferModel
	var err error

	offers, err = model.ListByProperty(ctx, request.Property, "")
	if err != nil && err != mongo.ErrNoDocuments {
		return nil, err
	}

	for _, off := range offers {
		off.Administrator = request.Admin

		err = model.UpdateOffer(ctx, off)
		if err != nil {
			return nil, err
		}
	}

	return &offer.ServiceSetAdminResponse{}, nil
}

// ServiceUpdateOffersLL to set landlord info for active offers
func (server *serverImpl) ServiceUpdateOffersLL(
	ctx context.Context,
	request *offer.ServiceUpdateOffersLLRequest,
) (*offer.ServiceUpdateOffersLLResponse, error) {
	var offers []*model.OfferModel
	var err error

	ll := request.GetLandlord()
	offers, err = model.ListNotSignedByLandlord(ctx, ll.GetUser())
	if err != nil && err != mongo.ErrNoDocuments {
		return nil, err
	}

	for _, off := range offers {
		off.Landlord.UpdateLandlordModel(ll)

		if off.State == offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES {
			err = status.GenerateDraftContract(ctx, off)
			if err != nil {
				return nil, err
			}
		}

		err = model.UpdateOffer(ctx, off)
		if err != nil {
			return nil, err
		}
	}

	return &offer.ServiceUpdateOffersLLResponse{}, nil
}

func (server *serverImpl) ServiceListingMigrate(
	ctx context.Context,
	req *offer.ServiceListingMigrateRequest,
) (*offer.ServiceListingMigrateResponse, error) {
	err := model.ListingOldNewMigrate(ctx, req.GetOldListing(), req.GetNewListing())
	if err != nil {
		return nil, err
	}

	return &offer.ServiceListingMigrateResponse{}, nil
}

func (server *serverImpl) OfferPropertyIDFilter(
	ctx context.Context,
	req *offer.OfferPropertyIDFilterRequest,
) (response *offer.OfferPropertyIDFilterResponse, err error) {
	propertyIDs, err := model.OfferPropertyIDs(ctx, req.GetUserId())
	if err != nil {
		return nil, err
	}

	return &offer.OfferPropertyIDFilterResponse{
		PropertyIds: propertyIDs,
	}, nil
}

func (server *serverImpl) ServiceOfferCount(
	ctx context.Context,
	req *offer.ServiceOfferCountRequest,
) (*offer.ServiceOfferCountResponse, error) {
	counts := make(map[string]int64)

	if req.GetProperties() == nil {
		return &offer.ServiceOfferCountResponse{
			Counts: counts,
		}, nil
	}

	after := req.GetProperties().GetAfter().AsTime().Add(1)
	before := req.GetProperties().GetBefore().AsTime().Add(-1)

	for _, p := range req.GetProperties().GetValue() {
		filters := []*query.QueryParamsFilter{
			{
				Fields:   []string{"property"},
				Value:    []string{p.GetProperty()},
				Operator: query.QueryFilterOperator_EQUALS,
			},
		}

		if p.GetListing() != "" {
			filters = append(filters, &query.QueryParamsFilter{
				Fields:   []string{"property_listing"},
				Value:    []string{p.GetListing()},
				Operator: query.QueryFilterOperator_EQUALS,
			})
		}

		if after.Unix() > 0 {
			filters = append(filters, &query.QueryParamsFilter{
				Fields:   []string{"created_at"},
				Value:    []string{after.Format(time.RFC3339)},
				Operator: query.QueryFilterOperator_MORE,
			})
		}

		if before.Unix() > 0 {
			filters = append(filters, &query.QueryParamsFilter{
				Fields:   []string{"created_at"},
				Value:    []string{before.Format(time.RFC3339)},
				Operator: query.QueryFilterOperator_LESS,
			})
		}

		count, err := model.OfferCount(ctx, &query.QueryParams{
			Filter: filters,
		})
		if err != nil {
			return nil, err
		}

		counts[p.GetProperty()] = count
	}

	return &offer.ServiceOfferCountResponse{
		Counts: counts,
	}, nil
}

func (server *serverImpl) OfferCount(
	ctx context.Context,
	req *offer.OfferCountRequest,
) (*offer.OfferCountResponse, error) {
	resp, err := model.OfferCount(ctx, req.GetParams())
	if err != nil {
		return nil, err
	}

	return &offer.OfferCountResponse{
		Count: resp,
	}, nil
}

// GenerateRentPlanInsurances is just for hotfix on prod
func (server *serverImpl) GenerateRentPlanInsurances(
	ctx context.Context,
	req *offer.GenerateRentPlanInsurancesRequest,
) (*offer.GenerateRentPlanInsurancesResponse, error) {
	id, err := primitive.ObjectIDFromHex(req.GetOfferId())
	if err != nil {
		return nil, err
	}
	old, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return nil, err
	}

	err = status.TrySetPropertyRentedCreateInsurances(ctx, old, req.GetIsImitationMode())
	if err != nil {
		return nil, err
	}

	err = model.UpdateOffer(ctx, old)
	if err != nil {
		return nil, err
	}

	return &offer.GenerateRentPlanInsurancesResponse{}, nil
}

func (server *serverImpl) GenerateRentInvoices(
	ctx context.Context,
	req *offer.GenerateRentInvoicesRequest,
) (*offer.GenerateRentInvoicesResponse, error) {
	id, err := primitive.ObjectIDFromHex(req.GetOfferId())
	if err != nil {
		return nil, err
	}
	old, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return nil, err
	}

	err = status.TryCreateSignedInvoices(ctx, old, req.GetIsImitationMode())
	if err != nil {
		return nil, err
	}

	return &offer.GenerateRentInvoicesResponse{}, nil
}

func (server *serverImpl) checkOfferStatusForUpdate(off *model.OfferModel) bool {
	return off.State != offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES &&
		off.State != offer.OfferState_AWAITING_APPLICANT_CHECK_PAYMENT
}
