package impl

import (
	"context"

	"gitlab.mashroom.com/mashroom/backend/common/api/process"
	contract "gitlab.mashroom.com/mashroom/backend/contract/api"
	invitation "gitlab.mashroom.com/mashroom/backend/invitation/api"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	"gitlab.mashroom.com/mashroom/backend/offer/docs"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/usecase/rto"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/connections"
	"gitlab.mashroom.com/mashroom/backend/offer/model"
	schedule "gitlab.mashroom.com/mashroom/backend/schedule/api"

	"github.com/camunda-cloud/zeebe/clients/go/pkg/zbc"
)

// Server describes Offer service server interface.
type Server interface {
	offer.OfferServiceServer
	invitation.InvitationNotifyServiceServer
	process.ProcessMonitorServiceServer
	invoice.InvoiceNotifyServiceServer
	schedule.ReminderConsumerServiceServer
	RegisterSchedules(ctx context.Context) error
}

type serverImpl struct {
	offer.UnimplementedOfferServiceServer
	bpms zbc.Client
	// when standing order was setup
	updateStateToWaitingForFirstRentPaymentUseCase rto.UpdateStateToWaitingForFirstRentPaymentUseCase
	// after first month paid -> set to active
	updateStateToActiveUseCase rto.UpdateStateToActiveUseCase
	// after last month paid -> set to ending
	updateStateToEndingUseCase rto.UpdateStateToEndingUseCase
}

// NewServer builds Offer server interface and runs background processing.
func NewServer(ctx context.Context,
	updateStateToWaitingForFirstRentPaymentUseCase rto.UpdateStateToWaitingForFirstRentPaymentUseCase,
	updateStateToActiveUseCase rto.UpdateStateToActiveUseCase,
	updateStateToEndingUseCase rto.UpdateStateToEndingUseCase) Server {
	var err error

	// template names are constant now, no need to update them each time template is updated
	// template revision is updated automatically if bindata modification time is after current
	// template revision modification date.
	// see contract.TemplateID for details.

	model.TemplateASTID, err = contract.TemplateID(
		ctx,
		docs.Asset,
		docs.AssetInfo,
		connections.ContractService,
		"offer-contract-template",
		"template.html",
	)
	if err != nil {
		panic(err)
	}

	model.TemplateSOMFID, err = contract.TemplateID(
		ctx,
		docs.Asset,
		docs.AssetInfo,
		connections.ContractService,
		"offer-somf-template",
		"StandingOrderMandateForm.html",
	)
	if err != nil {
		panic(err)
	}

	srv := &serverImpl{
		updateStateToWaitingForFirstRentPaymentUseCase: updateStateToWaitingForFirstRentPaymentUseCase,
		updateStateToActiveUseCase:                     updateStateToActiveUseCase,
		updateStateToEndingUseCase:                     updateStateToEndingUseCase,
	}
	srv.ServeBpms()

	return srv
}
