package usecase

import (
	"errors"
	"fmt"
	"net/mail"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/lib/constants"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/pkg/util"
)

type User struct {
	FirstName string
	LastName  string
	Email     string
}

type PaymentData struct {
	StartDate                     time.Time
	EndDate                       time.Time
	RentAmount                    float64
	PaymentDay                    uint32
	IsPaymentDayWasLastDayInMonth bool
	TermMonths                    uint32
	PaymentFrequency              uint32
}

type CreateTenancyRequest struct {
	PaymentData *PaymentData
	LeadTenant  *User
	ID          string
	LandlordID  string
}

func (ctr *CreateTenancyRequest) GetConditions() *domain.Conditions {
	if ctr.PaymentData != nil {
		return &domain.Conditions{
			Price:                         ctr.PaymentData.RentAmount,
			TermMonths:                    util.GetTermMonths(ctr.PaymentData.StartDate, ctr.PaymentData.EndDate),
			MoveInDate:                    ctr.PaymentData.StartDate,
			RentPaymentDay:                ctr.PaymentData.PaymentDay,
			IsPaymentDayWasLastDayInMonth: ctr.PaymentData.IsPaymentDayWasLastDayInMonth,
			EndDate:                       ctr.PaymentData.EndDate,
			PaymentFrequency:              uint32(invoice.PeriodicPaymentFrequency_FREQUENCY_MONTHLY),
		}
	}

	return nil
}

func (ctr *CreateTenancyRequest) GetLeadTenant() *domain.Tenant {
	if ctr.LeadTenant != nil {
		return &domain.Tenant{
			FirstName:    ctr.LeadTenant.FirstName,
			LastName:     ctr.LeadTenant.LastName,
			EmailPrimary: ctr.LeadTenant.Email,
		}
	}

	return nil
}

func (ctr *CreateTenancyRequest) Validate() error {
	if ctr.ID == "" {
		return fmt.Errorf(constants.ParameterCannotBeEmpty, "id")
	}
	if ctr.LandlordID == "" {
		return fmt.Errorf(constants.ParameterCannotBeEmpty, "landlord_id")
	}
	if ctr.PaymentData == nil {
		return fmt.Errorf(constants.ParameterCannotBeEmpty, "payment_data")
	}
	if ctr.PaymentData.RentAmount <= 0 {
		return fmt.Errorf(constants.ParameterCannotBeEmpty, "rent_amount")
	}
	if ctr.PaymentData.IsPaymentDayWasLastDayInMonth == false && ctr.PaymentData.PaymentDay <= 0 {
		return errors.New("parameters payment_day and is_payment_day_was_last_day_in_month " +
			"can not be empty in same time")
	}
	if ctr.PaymentData.IsPaymentDayWasLastDayInMonth {
		ctr.PaymentData.PaymentDay = 0
	} else if !(ctr.PaymentData.PaymentDay >= 1 && ctr.PaymentData.PaymentDay <= 31) {
		return errors.New("parameter payment_day must be greater or equal 1 and must be less or equal 31")
	}
	if ctr.LeadTenant == nil || ctr.LeadTenant.Email == "" {
		return fmt.Errorf(constants.ParameterCannotBeEmpty, "lead_tenant_email")
	}
	_, err := mail.ParseAddress(ctr.LeadTenant.Email)
	if err != nil {
		return fmt.Errorf("lead_tenant_email: %s not valid email address", ctr.LeadTenant.Email)
	}
	if ctr.PaymentData.StartDate.IsZero() {
		return fmt.Errorf(constants.ParameterCannotBeEmpty, "start_date")
	}
	if ctr.PaymentData.EndDate.IsZero() {
		return fmt.Errorf(constants.ParameterCannotBeEmpty, "end_date")
	}
	if ctr.PaymentData.EndDate.Before(ctr.PaymentData.StartDate) {
		return fmt.Errorf("rto end date (%s) can not be early than start date (%s)",
			ctr.PaymentData.EndDate, ctr.PaymentData.StartDate)
	}

	return nil
}
