package rto

import (
	"context"
	"time"

	account "gitlab.mashroom.com/mashroom/backend/account/api"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	api "gitlab.mashroom.com/mashroom/backend/offer/api/rto"

	"go.mongodb.org/mongo-driver/bson/primitive"
	"google.golang.org/protobuf/types/known/timestamppb"
)

type Rto struct {
	UpdatedAt       time.Time          `bson:"updated_at"`
	CreatedAt       time.Time          `bson:"created_at"`
	Conditions      *Conditions        `bson:"conditions"`
	Landlord        *Landlord          `bson:"landlord"`
	PaidStatuses    *PaidStatuses      `bson:"paid_statuses"`
	Notifications   Notifications      `bson:"notifications"`
	LeadTenant      *Tenant            `bson:"lead_tenant"`
	PropertyIvr     string             `bson:"property_ivr"`
	PropertyAddress string             `bson:"property_address"`
	Property        string             `bson:"property"`
	PropertyListing string             `bson:"property_listing"`
	Tenants         []*Tenant          `bson:"tenants"`
	RentPlan        api.RentPlan       `bson:"rent_plan"`
	State           api.RtoState       `bson:"state"`
	ID              primitive.ObjectID `bson:"_id"`
}

type Repository interface {
	Create(ctx context.Context, rto *Rto) error

	Update(ctx context.Context, rto *Rto) (int64, error)
	MustUpdate(ctx context.Context, rto *Rto) error

	DeleteByID(ctx context.Context, id primitive.ObjectID) (int64, error)

	Drop(ctx context.Context) error

	FindByID(ctx context.Context, id primitive.ObjectID) (*Rto, error)
	FindByPropertyID(ctx context.Context, propertyID string) (*Rto, error)

	FindAllByFilter(ctx context.Context, filter *Filter) ([]*Rto, error)
	FindAllEmptyLeadTenantUsers(ctx context.Context, states []api.RtoState) ([]*Rto, error)
	FindAllByPropertyIDs(ctx context.Context, propertyIDs []string) ([]*Rto, error)
}

func (r *Rto) CreateApiRto() *api.Rto {
	return &api.Rto{
		Id:              r.ID.String(),
		RentPlan:        r.RentPlan,
		State:           r.State,
		CreatedAt:       timestamppb.New(r.CreatedAt),
		UpdatedAt:       timestamppb.New(r.UpdatedAt),
		Landlord:        CreateApiLandlord(r.Landlord),
		Property:        r.Property,
		PropertyIvr:     r.PropertyIvr,
		PropertyListing: r.PropertyListing,
		PropertyAddress: r.PropertyAddress,
		Conditions:      CreateApiConditions(r.Conditions),
		LeadTenant:      CreateApiTenant(r.LeadTenant),
	}
}

func (r *Rto) GetDate(dateColumnName DateColumnName) time.Time {
	switch dateColumnName {
	case DateColumnNameCreatedAt:
		return r.CreatedAt
	case DateColumnNameUpdatedAt:
		return r.UpdatedAt
	case DateColumnNameEndDate:
		if r.Conditions != nil {
			return r.Conditions.EndDate
		}

		return time.Time{}
	default:
		return time.Time{}
	}
}

func (r *Rto) GetLeadTenantID() string {
	if r.LeadTenant != nil {
		return r.LeadTenant.User
	}

	return ""
}

func CreateRto(now time.Time,
	rentPlan api.RentPlan,
	landlordID string,
	landlordAccount *account.Account,
	listingVersion *listing.Version) *Rto {
	return &Rto{
		ID:              primitive.NewObjectID(),
		State:           api.RtoState_RTO_STATE_PENDING_SETUP,
		CreatedAt:       now,
		UpdatedAt:       now,
		RentPlan:        rentPlan,
		Landlord:        CreateLandlord(landlordID, landlordAccount),
		Property:        listingVersion.GetPropertyId(),
		PropertyIvr:     listingVersion.GetIvr(),
		PropertyListing: listingVersion.GetListingId(),
		PropertyAddress: listingVersion.GetLocation().FormatAddress(),
	}
}
