package daemon

import (
	"context"
	"time"

	api "gitlab.mashroom.com/mashroom/backend/offer/api/rto"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"
	usecaserto "gitlab.mashroom.com/mashroom/backend/offer/internal/app/usecase/rto"

	"github.com/pkg/errors"
)

type FinishService interface {
	FinishAll(ctx context.Context) error
	Finish(ctx context.Context) (*response.IDsResponse, error)
}

type finishService struct {
	repository                   domain.Repository
	updateStateToFinishedUseCase usecaserto.UpdateStateToFinishedUseCase
}

func NewFinishService(repository domain.Repository,
	updateStateToFinishedUseCase usecaserto.UpdateStateToFinishedUseCase) FinishService {
	return &finishService{
		repository:                   repository,
		updateStateToFinishedUseCase: updateStateToFinishedUseCase,
	}
}

func (fs *finishService) FinishAll(ctx context.Context) error {
	for {
		select {
		case <-ctx.Done():
			return ctx.Err()
		default:
			go func() {
				_, _ = fs.Finish(ctx)
			}()
			time.Sleep(time.Hour)
		}
	}
}

func (fs *finishService) Finish(ctx context.Context) (*response.IDsResponse, error) {
	now := time.Now()
	states := []api.RtoState{
		api.RtoState_RTO_STATE_ACTIVE,
		api.RtoState_RTO_STATE_ACTIVE_ROLLING_TENANCY,
		api.RtoState_RTO_STATE_ENDING,
	}
	rtos, err := fs.repository.FindAllByFilter(ctx, &domain.Filter{
		States:         states,
		DateFrom:       now.Add(time.Duration(-12) * time.Hour),
		DateTo:         now.Add(time.Duration(12) * time.Hour),
		DateColumnName: domain.DateColumnNameCreatedAt,
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto daemon finishService Finish repository.FindAllByFilter")
	}
	var rtoIDs []string
	for _, rto := range rtos {
		rtoID := rto.ID.Hex()
		_, err = fs.updateStateToFinishedUseCase.UpdateStateToFinished(ctx, &request.UpdateStateToFinishedRequest{
			ID: rtoID,
		})
		if err != nil {
			return nil, errors.Wrap(err, "rto daemon finishService Finish updateStateToFinishedUseCase")
		}
		rtoIDs = append(rtoIDs, rtoID)
	}

	return &response.IDsResponse{
		IDs: rtoIDs,
	}, nil
}
