package daemon

import (
	"context"
	"time"

	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	api "gitlab.mashroom.com/mashroom/backend/offer/api/rto"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/service/daemon"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"

	"github.com/pkg/errors"
)

type NotificationService interface {
	RemindAll(ctx context.Context) error

	RemindAboutPendingSetup(ctx context.Context) error
	SendRemindAboutPendingSetup(ctx context.Context, req *request.SendRemindAboutPendingSetupRequest) (
		*response.IDsResponse, error)

	RemindAboutWaitingForStandingOrderSetup(ctx context.Context) error
	SendRemindAboutWaitingForStandingOrderSetup(ctx context.Context,
		req *request.SendRemindAboutWaitingForStandingOrderSetupRequest) (
		*response.IDsResponse, error)

	RemindAboutEndOfTenancyDate(ctx context.Context) error
	SendRemindAboutEndOfTenancyDate(ctx context.Context, req *request.SendRemindAboutEndOfTenancyDateRequest) (
		*response.IDsResponse, error)
}

type notificationService struct {
	notificationServiceClient notification.NotificationServiceClient
	repository                domain.Repository
}

func NewNotificationService(notificationServiceClient notification.NotificationServiceClient,
	repository domain.Repository) NotificationService {
	return &notificationService{
		notificationServiceClient: notificationServiceClient,
		repository:                repository,
	}
}

func (ns *notificationService) RemindAll(ctx context.Context) error {
	go func() {
		_ = ns.RemindAboutPendingSetup(ctx)
	}()
	go func() {
		_ = ns.RemindAboutWaitingForStandingOrderSetup(ctx)
	}()
	go func() {
		_ = ns.RemindAboutEndOfTenancyDate(ctx)
	}()

	return nil
}

func (ns *notificationService) RemindAboutPendingSetup(ctx context.Context) error {
	for {
		select {
		case <-ctx.Done():
			return ctx.Err()
		default:
			_, err := ns.SendRemindAboutPendingSetup(ctx, &request.SendRemindAboutPendingSetupRequest{
				NotificationType: domain.NotificationPendingSetupOneDay,
				DateFromHours:    -25,
				DateToHours:      -23,
				DateSubHours:     24,
				DateColumnName:   domain.DateColumnNameCreatedAt,
			})
			if err != nil {
				return errors.Wrap(err, "rto daemon notificationService RemindAboutPendingSetup "+
					"SendRemindAboutPendingSetup 1")
			}

			_, err = ns.SendRemindAboutPendingSetup(ctx, &request.SendRemindAboutPendingSetupRequest{
				NotificationType: domain.NotificationPendingSetupTwoDays,
				DateFromHours:    -49,
				DateToHours:      -47,
				DateSubHours:     48,
				DateColumnName:   domain.DateColumnNameCreatedAt,
			})
			if err != nil {
				return errors.Wrap(err, "rto daemon notificationService RemindAboutPendingSetup "+
					"SendRemindAboutPendingSetup 2")
			}

			time.Sleep(time.Hour)
		}
	}
}

func (ns *notificationService) SendRemindAboutPendingSetup(ctx context.Context,
	req *request.SendRemindAboutPendingSetupRequest) (*response.IDsResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto daemon notificationService SendRemindAboutPendingSetup req.Validate")
	}

	now := time.Now()
	rtos, err := ns.repository.FindAllByFilter(ctx, &domain.Filter{
		States:           []api.RtoState{api.RtoState_RTO_STATE_PENDING_SETUP},
		NotificationType: req.NotificationType,
		DateFrom:         now.Add(time.Duration(req.DateFromHours) * time.Hour),
		DateTo:           now.Add(time.Duration(req.DateToHours) * time.Hour),
		DateColumnName:   req.DateColumnName,
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto daemon notificationService SendRemindAboutPendingSetup "+
			"repository.FindAllByFilter")
	}

	resp := &response.IDsResponse{}
	for _, rto := range rtos {
		if int(now.Sub(rto.GetDate(req.DateColumnName)).Hours()) == req.DateSubHours {
			_, err = ns.notificationServiceClient.SendNotification(ctx, &notification.SendNotificationRequest{
				Entity: notification.Entity_ENTITY_OFFER_RTO,
				Action: notification.Action_ACTION_PENDING,
				Targets: []*notification.Target{
					{
						User: rto.Landlord.User,
						Role: notification.Role_ROLE_LANDLORD,
					},
				},
				Embeds: []*notification.EventEmbed{
					{
						EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
						EmbedValue: rto.ID.Hex(),
					},
					{
						EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
						EmbedValue: rto.Property,
					},
				},
			})
			if err != nil {
				return nil, errors.Wrap(err, "rto daemon notificationService SendRemindAboutPendingSetup "+
					"notificationServiceClient.SendNotification "+rto.ID.Hex())
			}

			if rto.Notifications == nil {
				rto.Notifications = domain.Notifications{
					req.NotificationType: now,
				}
			} else {
				rto.Notifications[req.NotificationType] = now
			}
			rto.UpdatedAt = now
			err = ns.repository.MustUpdate(ctx, rto)
			if err != nil {
				return nil, errors.Wrap(err, "rto daemon notificationService SendRemindAboutPendingSetup "+
					"repository.MustUpdate "+rto.ID.Hex())
			}

			resp.IDs = append(resp.IDs, rto.ID.Hex())
		}
	}

	return resp, nil
}

func (ns *notificationService) RemindAboutWaitingForStandingOrderSetup(ctx context.Context) error {
	for {
		select {
		case <-ctx.Done():
			return ctx.Err()
		default:
			_, err := ns.SendRemindAboutWaitingForStandingOrderSetup(ctx,
				&request.SendRemindAboutWaitingForStandingOrderSetupRequest{
					NotificationType: domain.NotificationWaitingForStandingOrderSetupOneDay,
					DateFromHours:    -25,
					DateToHours:      -23,
					DateSubHours:     24,
					DateColumnName:   domain.DateColumnNameUpdatedAt,
				})
			if err != nil {
				return errors.Wrap(err, "rto daemon notificationService "+
					"RemindAboutWaitingForStandingOrderSetup SendRemindAboutWaitingForStandingOrderSetup 1")
			}

			_, err = ns.SendRemindAboutWaitingForStandingOrderSetup(ctx,
				&request.SendRemindAboutWaitingForStandingOrderSetupRequest{
					NotificationType: domain.NotificationWaitingForStandingOrderSetupTwoDays,
					DateFromHours:    -49,
					DateToHours:      -47,
					DateSubHours:     48,
					DateColumnName:   domain.DateColumnNameUpdatedAt,
				})
			if err != nil {
				return errors.Wrap(err, "rto daemon notificationService "+
					"RemindAboutWaitingForStandingOrderSetup SendRemindAboutWaitingForStandingOrderSetup 2")
			}

			time.Sleep(time.Hour)
		}
	}
}

func (ns *notificationService) SendRemindAboutWaitingForStandingOrderSetup(ctx context.Context,
	req *request.SendRemindAboutWaitingForStandingOrderSetupRequest) (
	*response.IDsResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto daemon notificationService "+
			"SendRemindAboutWaitingForStandingOrderSetup req.Validate")
	}

	now := time.Now()
	rtos, err := ns.repository.FindAllByFilter(ctx, &domain.Filter{
		States:           []api.RtoState{api.RtoState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP},
		NotificationType: req.NotificationType,
		DateFrom:         now.Add(time.Duration(req.DateFromHours) * time.Hour),
		DateTo:           now.Add(time.Duration(req.DateToHours) * time.Hour),
		DateColumnName:   req.DateColumnName,
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto daemon notificationService "+
			"SendRemindAboutWaitingForStandingOrderSetup repository.FindAllByFilter")
	}

	resp := &response.IDsResponse{}
	for _, rto := range rtos {
		if int(now.Sub(rto.GetDate(req.DateColumnName)).Hours()) == req.DateSubHours {
			_, err = ns.notificationServiceClient.SendNotification(ctx, &notification.SendNotificationRequest{
				Entity: notification.Entity_ENTITY_OFFER_RTO,
				Action: notification.Action_ACTION_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP,
				Targets: []*notification.Target{
					{
						User: rto.LeadTenant.User,
						Role: notification.Role_ROLE_TENANT,
					},
				},
				Embeds: []*notification.EventEmbed{
					{
						EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
						EmbedValue: rto.ID.Hex(),
					},
					{
						EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
						EmbedValue: rto.Property,
					},
				},
			})
			if err != nil {
				return nil, errors.Wrap(err, "rto daemon notificationService "+
					"SendRemindAboutWaitingForStandingOrderSetup notificationServiceClient.SendNotification "+
					rto.ID.Hex())
			}

			if rto.Notifications == nil {
				rto.Notifications = domain.Notifications{
					req.NotificationType: now,
				}
			} else {
				rto.Notifications[req.NotificationType] = now
			}
			rto.UpdatedAt = now
			err = ns.repository.MustUpdate(ctx, rto)
			if err != nil {
				return nil, errors.Wrap(err, "rto daemon notificationService "+
					"SendRemindAboutWaitingForStandingOrderSetup repository.MustUpdate "+rto.ID.Hex())
			}

			resp.IDs = append(resp.IDs, rto.ID.Hex())
		}
	}

	return resp, nil
}

func (ns *notificationService) RemindAboutEndOfTenancyDate(ctx context.Context) error {
	for {
		select {
		case <-ctx.Done():
			return ctx.Err()
		default:
			_, err := ns.SendRemindAboutEndOfTenancyDate(ctx, &request.SendRemindAboutEndOfTenancyDateRequest{
				NotificationType: domain.NotificationEndOfTenancyDateBeforeTwoMonths,
				DateFromHours:    24 * 59.5,
				DateToHours:      24 * 60.5,
				DateColumnName:   domain.DateColumnNameEndDate,
			})
			if err != nil {
				return errors.Wrap(err, "rto daemon notificationService RemindAboutEndOfTenancyDate "+
					"SendRemindAboutEndOfTenancyDate")
			}

			time.Sleep(24 * time.Hour)
		}
	}
}

func (ns *notificationService) SendRemindAboutEndOfTenancyDate(ctx context.Context,
	req *request.SendRemindAboutEndOfTenancyDateRequest) (*response.IDsResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto daemon notificationService "+
			"SendRemindAboutEndOfTenancyDate req.Validate")
	}

	now := time.Now()
	rtos, err := ns.repository.FindAllByFilter(ctx, &domain.Filter{
		States:           []api.RtoState{api.RtoState_RTO_STATE_ACTIVE},
		NotificationType: req.NotificationType,
		DateFrom:         now.Add(time.Duration(req.DateFromHours) * time.Hour),
		DateTo:           now.Add(time.Duration(req.DateToHours) * time.Hour),
		DateColumnName:   req.DateColumnName,
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto daemon notificationService SendRemindAboutEndOfTenancyDate "+
			"repository.FindAllByFilter")
	}

	resp := &response.IDsResponse{}
	for _, rto := range rtos {
		_, err = ns.notificationServiceClient.SendNotification(ctx, &notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_OFFER_RTO,
			Action: notification.Action_ACTION_RTO_TWO_MONTHS_BEFORE_END_OF_TENANCY_DATE,
			Targets: []*notification.Target{
				{
					User: rto.Landlord.User,
					Role: notification.Role_ROLE_LANDLORD,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: rto.Property,
				},
			},
		})
		if err != nil {
			return nil, errors.Wrap(err, "rto daemon notificationService SendRemindAboutEndOfTenancyDate "+
				"notificationServiceClient.SendNotification "+rto.ID.Hex())
		}

		if rto.Notifications == nil {
			rto.Notifications = domain.Notifications{
				req.NotificationType: now,
			}
		} else {
			rto.Notifications[req.NotificationType] = now
		}
		rto.UpdatedAt = now
		err = ns.repository.MustUpdate(ctx, rto)
		if err != nil {
			return nil, errors.Wrap(err, "rto daemon notificationService SendRemindAboutEndOfTenancyDate "+
				"repository.MustUpdate "+rto.ID.Hex())
		}

		resp.IDs = append(resp.IDs, rto.ID.Hex())
	}

	return resp, nil
}
