package daemon

import (
	"context"
	"time"

	api "gitlab.mashroom.com/mashroom/backend/offer/api/rto"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"
	usecaserto "gitlab.mashroom.com/mashroom/backend/offer/internal/app/usecase/rto"

	"github.com/pkg/errors"
)

type UpdateLeadTenantService interface {
	UpdateAllLeadTenantUsers(ctx context.Context) error
	UpdateLeadTenantUsers(ctx context.Context) (*response.IDsResponse, error)
}

type updateLeadTenantUserService struct {
	repository                  domain.Repository
	updateLeadTenantUserUseCase usecaserto.UpdateLeadTenantUserUseCase
}

func NewUpdateLeadTenantUserService(rtoRepository domain.Repository,
	updateLeadTenantUserUseCase usecaserto.UpdateLeadTenantUserUseCase) UpdateLeadTenantService {
	return &updateLeadTenantUserService{
		repository:                  rtoRepository,
		updateLeadTenantUserUseCase: updateLeadTenantUserUseCase,
	}
}

func (ultus *updateLeadTenantUserService) UpdateAllLeadTenantUsers(ctx context.Context) error {
	for {
		select {
		case <-ctx.Done():
			return ctx.Err()
		default:
			go func() {
				_, _ = ultus.UpdateLeadTenantUsers(ctx)
			}()
			time.Sleep(time.Hour)
		}
	}
}

func (ultus *updateLeadTenantUserService) UpdateLeadTenantUsers(ctx context.Context) (*response.IDsResponse, error) {
	states := []api.RtoState{
		api.RtoState_RTO_STATE_PENDING_SETUP,
		api.RtoState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP,
		api.RtoState_RTO_STATE_WAITING_FOR_FIRST_RENT_PAYMENT,
	}
	rtos, err := ultus.repository.FindAllEmptyLeadTenantUsers(ctx, states)
	if err != nil {
		return nil, errors.Wrap(err, "rto daemon updateLeadTenantUserService "+
			"UpdateLeadTenantUsers repository.FindAllEmptyLeadTenantUsers")
	}
	var rtoIDs []string
	for _, rto := range rtos {
		resp, err := ultus.updateLeadTenantUserUseCase.UpdateLeadTenantUser(ctx,
			&request.UpdateLeadTenantUserRequest{Rto: rto})
		if err != nil {
			return nil, errors.Wrap(err, "rto daemon updateLeadTenantUserService "+
				"UpdateLeadTenantUsers updateLeadTenantUserUseCase.UpdateLeadTenantUser")
		}
		if resp.Rto != nil {
			rtoIDs = append(rtoIDs, rto.ID.Hex())
		}
	}

	return &response.IDsResponse{
		IDs: rtoIDs,
	}, nil
}
