package rto

import (
	"context"

	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/service"

	"github.com/pkg/errors"
)

const IsRtoTrue = `{"is_rto":"TRUE"}`

type NotificationService interface {
	SendWaitingForStandingOrderSetup(ctx context.Context, req *request.SendWaitingForStandingOrderSetupRequest) error
	SendWaitingForFirstRentPayment(ctx context.Context, req *request.SendWaitingForFirstRentPaymentRequest) error
	SendActive(ctx context.Context, req *request.SendActiveRequest) error
}

type notificationService struct {
	notificationServiceClient notification.NotificationServiceClient
	repository                domain.Repository
}

func NewNotificationService(notificationServiceClient notification.NotificationServiceClient,
	repository domain.Repository) NotificationService {
	return &notificationService{
		notificationServiceClient: notificationServiceClient,
		repository:                repository,
	}
}

func (ns *notificationService) SendWaitingForStandingOrderSetup(ctx context.Context,
	req *request.SendWaitingForStandingOrderSetupRequest) error {
	err := req.Validate()
	if err != nil {
		return errors.Wrap(err, "rto notificationService SendWaitingForStandingOrderSetup req.Validate")
	}

	sendNotificationRequest := &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_RTO,
		Action: notification.Action_ACTION_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP,
		Targets: []*notification.Target{
			{
				User: req.LandlordID,
				Role: notification.Role_ROLE_LANDLORD,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: req.PropertyID,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: req.ID,
			},
		},
	}
	_, err = ns.notificationServiceClient.SendNotification(ctx, sendNotificationRequest)
	if err != nil {
		return errors.Wrap(err, "rto notificationService SendWaitingForStandingOrderSetup "+
			"notificationServiceClient.SendNotification for landlord")
	}

	sendNotificationRequest.Targets = []*notification.Target{
		{
			User: req.TenantID,
			Role: notification.Role_ROLE_TENANT,
		},
	}
	_, err = ns.notificationServiceClient.SendNotification(ctx, sendNotificationRequest)
	if err != nil {
		return errors.Wrap(err, "rto notificationService SendWaitingForStandingOrderSetup "+
			"notificationServiceClient.SendNotification for tenant")
	}

	return nil
}

func (ns *notificationService) SendActive(ctx context.Context, req *request.SendActiveRequest) error {
	err := req.Validate()
	if err != nil {
		return errors.Wrap(err, "rto notificationService SendActive req.Validate")
	}

	_, err = ns.notificationServiceClient.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_RTO,
		Action: req.GetNotificationAction(),
		Targets: []*notification.Target{
			{
				User: req.LandlordID,
				Role: notification.Role_ROLE_LANDLORD,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: req.PropertyID,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_JSON,
				EmbedValue: IsRtoTrue,
			},
		},
	})
	if err != nil {
		return errors.Wrap(err, "rto notificationService SendActive "+
			"notificationServiceClient.SendNotification for landlord")
	}

	return nil
}

func (ns *notificationService) SendWaitingForFirstRentPayment(ctx context.Context,
	req *request.SendWaitingForFirstRentPaymentRequest) error {
	err := req.Validate()
	if err != nil {
		return errors.Wrap(err, "rto notificationService SendWaitingForFirstRentPayment req.Validate")
	}

	_, err = ns.notificationServiceClient.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_RTO,
		Action: notification.Action_ACTION_RTO_STATE_WAITING_FOR_FIRST_RENT_PAYMENT,
		Targets: []*notification.Target{
			{
				User: req.LandlordID,
				Role: notification.Role_ROLE_LANDLORD,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: req.PropertyID,
			},
		},
	})
	if err != nil {
		return errors.Wrap(err, "rto notificationService SendWaitingForFirstRentPayment "+
			"notificationServiceClient.SendNotification for landlord")
	}

	return nil
}
