package rto

import (
	"context"
	"fmt"

	account "gitlab.mashroom.com/mashroom/backend/account/api"
	api "gitlab.mashroom.com/mashroom/backend/offer/api/rto"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/service"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

var (
	VeriffStatusNotValidError = errors.New("veriff status not valid error")
)

const (
	IsVeriffValidNotValidRequestErrorText = "rto securityService IsVeriffValid req.Validate"
	IsVeriffValidErrorText                = "rto securityService IsVeriffValid accountServiceClient.IsVeriffValid"
	IsVeriffValidNotSuccessErrorText      = "rto securityService IsVeriffValid !isVeriffValidResponse.Success"

	CanUserReadRtoNotValidUserErrorText  = "rto securityService CanUserReadRto bad user"
	CanUserReadRtoNotValidStateErrorText = "rto securityService CanUserReadRto bad state"
)

type SecurityService interface {
	IsVeriffValid(ctx context.Context, req *request.IsVeriffValidRequest) error
	CanLandlordUpdateRto(ctx context.Context, req *request.CanLandlordUpdateRtoRequest) (*response.RtoResponse, error)
	CanUserReadRto(ctx context.Context, req *request.CanUserReadRtoRequest) error
	CanTenantUpdateRto(ctx context.Context, req *request.CanTenantUpdateRtoRequest) (*response.RtoResponse, error)
}

type securityService struct {
	repository           domain.Repository
	accountServiceClient account.AccountServiceClient
}

func NewSecurityService(repository domain.Repository,
	accountServiceClient account.AccountServiceClient) SecurityService {
	return &securityService{
		repository:           repository,
		accountServiceClient: accountServiceClient,
	}
}

func (ss *securityService) IsVeriffValid(ctx context.Context, req *request.IsVeriffValidRequest) error {
	err := req.Validate()
	if err != nil {
		return errors.Wrap(err, IsVeriffValidNotValidRequestErrorText)
	}

	isVeriffValidResponse, err := ss.accountServiceClient.IsVeriffValid(ctx, &account.IsVeriffValidRequest{
		AccountId: req.LandlordID,
	})
	if err != nil {
		return errors.Wrap(err, IsVeriffValidErrorText)
	}
	if !isVeriffValidResponse.Success {
		return errors.Wrap(VeriffStatusNotValidError, IsVeriffValidNotSuccessErrorText)
	}

	return nil
}

func (ss *securityService) CanLandlordUpdateRto(ctx context.Context, req *request.CanLandlordUpdateRtoRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto securityService CanLandlordUpdateRto req.Validate")
	}

	if req.NeedCheckVeriff {
		err = ss.IsVeriffValid(ctx, &request.IsVeriffValidRequest{
			LandlordID: req.LandlordID,
		})
		if err != nil {
			return nil, errors.Wrap(err, "rto securityService CanLandlordUpdateRto IsVeriffValid")
		}
	}

	rtoID, err := primitive.ObjectIDFromHex(req.ID)
	if err != nil {
		return nil, errors.Wrap(err, "rto securityService CanLandlordUpdateRto primitive.ObjectIDFromHex")
	}

	rto, err := ss.repository.FindByID(ctx, rtoID)
	if err != nil {
		return nil, errors.Wrap(err, "rto securityService CanLandlordUpdateRto repository.FindByID")
	}

	if rto.Landlord.User != req.LandlordID {
		return nil, errors.Wrap(fmt.Errorf("403 error rto.Landlord.User (%s) != req.LandlordID (%s)",
			rto.Landlord.User, req.LandlordID),
			"rto securityService CanLandlordUpdateRto rto.Landlord.User != req.LandlordID")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}

func (ss *securityService) CanUserReadRto(ctx context.Context, req *request.CanUserReadRtoRequest) error {
	err := req.Validate()
	if err != nil {
		return errors.Wrap(err, "rto securityService CanUserReadRto req.Validate")
	}

	if req.Rto.Landlord.User == req.UserID {
		return nil
	}

	if req.Rto.GetLeadTenantID() == req.UserID {
		if req.Rto.State == api.RtoState_RTO_STATE_PENDING_SETUP {
			return errors.Wrap(errors.New(CanUserReadRtoNotValidStateErrorText),
				"rto securityService CanUserReadRto rto.State == api.RtoState_RTO_STATE_PENDING_SETUP")
		}

		return nil
	}

	return errors.Wrap(errors.New(CanUserReadRtoNotValidUserErrorText), "rto securityService CanUserReadRto")
}

func (ss *securityService) CanTenantUpdateRto(ctx context.Context, req *request.CanTenantUpdateRtoRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto securityService CanTenantUpdateRto req.Validate")
	}

	rtoID, err := primitive.ObjectIDFromHex(req.ID)
	if err != nil {
		return nil, errors.Wrap(err, "rto securityService CanTenantUpdateRto primitive.ObjectIDFromHex")
	}

	rto, err := ss.repository.FindByID(ctx, rtoID)
	if err != nil {
		return nil, errors.Wrap(err, "rto securityService CanTenantUpdateRto repository.FindByID")
	}

	if rto.GetLeadTenantID() != req.TenantID {
		return nil, errors.Wrap(fmt.Errorf("403 error rto.GetLeadTenantID() (%s) != req.TenantID (%s)",
			rto.GetLeadTenantID(), req.TenantID),
			"rto securityService CanTenantUpdateRto rto.GetLeadTenantID() != req.TenantID")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
