package grpc

import (
	"context"

	api "gitlab.mashroom.com/mashroom/backend/offer/api/rto"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	usecase "gitlab.mashroom.com/mashroom/backend/offer/internal/app/usecase/rto"
)

type RtoServer interface {
	api.RtoServiceServer
}

type rtoServer struct {
	api.UnimplementedRtoServiceServer
	createUseCase                                    usecase.CreateUseCase
	createByGuestUseCase                             usecase.CreateByGuestUseCase
	updateLandlordUseCase                            usecase.UpdateLandlordUseCase
	createTenancyUseCase                             usecase.CreateTenancyUseCase
	updateStateUseCase                               usecase.UpdateStateUseCase
	updateRentPlanUseCase                            usecase.UpdateRentPlanUseCase
	createInvoiceUseCase                             usecase.CreateInvoiceUseCase
	updateEndDateUseCase                             usecase.UpdateEndDateUseCase
	renewalUseCase                                   usecase.RenewalUseCase
	updateStateToWaitingForStandingOrderSetupUseCase usecase.UpdateStateToWaitingForStandingOrderSetupUseCase
	updateStateToCancelledUseCase                    usecase.UpdateStateToCancelledUseCase
	getByPropertyIDUseCase                           usecase.GetByPropertyIDUseCase
	getByPropertyIDsUseCase                          usecase.GetByPropertyIDsUseCase
	answerFromTenantUseCase                          usecase.AnswerFromTenantUseCase
}

func NewRtoServer(createUseCase usecase.CreateUseCase,
	createByGuestUseCase usecase.CreateByGuestUseCase,
	updateLandlordUseCase usecase.UpdateLandlordUseCase,
	createTenancyUseCase usecase.CreateTenancyUseCase,
	updateStateUseCase usecase.UpdateStateUseCase,
	updateRentPlanUseCase usecase.UpdateRentPlanUseCase,
	createInvoiceUseCase usecase.CreateInvoiceUseCase,
	updateEndDateUseCase usecase.UpdateEndDateUseCase,
	renewalUseCase usecase.RenewalUseCase,
	updateStateToWaitingForStandingOrderSetupUseCase usecase.UpdateStateToWaitingForStandingOrderSetupUseCase,
	updateStateToCancelledUseCase usecase.UpdateStateToCancelledUseCase,
	getByPropertyIDUseCase usecase.GetByPropertyIDUseCase,
	getByPropertyIDsUseCase usecase.GetByPropertyIDsUseCase,
	answerFromTenantUseCase usecase.AnswerFromTenantUseCase) RtoServer {
	return &rtoServer{
		createUseCase:         createUseCase,
		createByGuestUseCase:  createByGuestUseCase,
		updateLandlordUseCase: updateLandlordUseCase,
		createTenancyUseCase:  createTenancyUseCase,
		updateStateUseCase:    updateStateUseCase,
		updateRentPlanUseCase: updateRentPlanUseCase,
		createInvoiceUseCase:  createInvoiceUseCase,
		updateEndDateUseCase:  updateEndDateUseCase,
		renewalUseCase:        renewalUseCase,
		updateStateToWaitingForStandingOrderSetupUseCase: updateStateToWaitingForStandingOrderSetupUseCase,
		updateStateToCancelledUseCase:                    updateStateToCancelledUseCase,
		getByPropertyIDUseCase:                           getByPropertyIDUseCase,
		getByPropertyIDsUseCase:                          getByPropertyIDsUseCase,
		answerFromTenantUseCase:                          answerFromTenantUseCase,
	}
}

func (rs *rtoServer) Create(ctx context.Context, req *api.CreateRequest) (*api.RtoResponse, error) {
	resp, err := rs.createUseCase.Create(ctx, &request.CreateRequest{
		LandlordID: req.GetLandlordId(),
		RentPlan:   req.GetRentPlan(),
		PropertyID: req.GetPropertyId(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiRto(),
	}, nil
}

func (rs *rtoServer) CreateByGuest(ctx context.Context, req *api.CreateByGuestRequest) (*api.RtoResponse, error) {
	resp, err := rs.createByGuestUseCase.CreateByGuest(ctx, &request.CreateByGuestRequest{
		RentPlan:   req.GetRentPlan(),
		PropertyID: req.GetPropertyId(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiRto(),
	}, nil
}

func (rs *rtoServer) UpdateLandlord(ctx context.Context, req *api.UpdateLandlordRequest) (*api.RtoResponse, error) {
	resp, err := rs.updateLandlordUseCase.UpdateLandlord(ctx, &request.UpdateLandlordRequest{
		ID:         req.GetId(),
		LandlordID: req.GetLandlordId(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiRto(),
	}, nil
}

func (rs *rtoServer) CreateTenancy(ctx context.Context, req *api.CreateTenancyRequest) (*api.RtoResponse, error) {
	resp, err := rs.createTenancyUseCase.CreateTenancy(ctx, &request.CreateTenancyRequest{
		ID:         req.GetId(),
		LandlordID: req.GetLandlordId(),
		PaymentData: &request.PaymentData{
			StartDate:                     req.GetPaymentData().GetStartDate().AsTime(),
			EndDate:                       req.GetPaymentData().GetEndDate().AsTime(),
			RentAmount:                    req.GetPaymentData().GetRentAmount(),
			PaymentDay:                    req.GetPaymentData().GetPaymentDay(),
			IsPaymentDayWasLastDayInMonth: req.GetPaymentData().GetIsPaymentDayWasLastDayInMonth(),
		},
		LeadTenant: &request.User{
			FirstName: req.GetLeadTenant().GetFirstNameValue(),
			LastName:  req.GetLeadTenant().GetLastNameValue(),
			Email:     req.GetLeadTenant().GetEmail(),
		},
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiRto(),
	}, nil
}

func (rs *rtoServer) UpdateState(ctx context.Context, req *api.UpdateStateRequest) (*api.RtoResponse, error) {
	resp, err := rs.updateStateUseCase.UpdateState(ctx, &request.UpdateStateRequest{
		ID:         req.GetId(),
		LandlordID: req.GetLandlordId(),
		State:      req.GetState(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiRto(),
	}, nil
}

func (rs *rtoServer) UpdateRentPlan(ctx context.Context, req *api.UpdateRentPlanRequest) (*api.RtoResponse, error) {
	resp, err := rs.updateRentPlanUseCase.UpdateRentPlan(ctx, &request.UpdateRentPlanRequest{
		ID:         req.GetId(),
		LandlordID: req.GetLandlordId(),
		RentPlan:   req.GetRentPlan(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiRto(),
	}, nil
}

func (rs *rtoServer) CreateInvoice(ctx context.Context, req *api.CreateInvoiceRequest) (*api.RtoResponse, error) {
	resp, err := rs.createInvoiceUseCase.CreateInvoice(ctx, &request.CreateInvoiceRequest{
		ID:         req.GetId(),
		LandlordID: req.GetLandlordId(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiRto(),
	}, nil
}

func (rs *rtoServer) UpdateEndDate(ctx context.Context, req *api.UpdateRtoEndDateRequest) (*api.RtoResponse, error) {
	resp, err := rs.updateEndDateUseCase.UpdateEndDate(ctx, &request.UpdateEndDateRequest{
		ID:         req.GetId(),
		LandlordID: req.GetLandlordId(),
		EndDate:    req.GetEndDate().AsTime(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiRto(),
	}, nil
}

func (rs *rtoServer) Renewal(ctx context.Context, req *api.RenewalRequest) (*api.RtoResponse, error) {
	resp, err := rs.renewalUseCase.Renewal(ctx, &request.RenewalRequest{
		ID:         req.GetId(),
		LandlordID: req.GetLandlordId(),
		EndDate:    req.GetEndDate().AsTime(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiRto(),
	}, nil
}

func (rs *rtoServer) UpdateStateToWaitingForStandingOrderSetup(ctx context.Context,
	req *api.UpdateStateToWaitingForStandingOrderSetupRequest) (*api.RtoResponse, error) {
	resp, err := rs.updateStateToWaitingForStandingOrderSetupUseCase.UpdateStateToWaitingForStandingOrderSetup(ctx,
		&request.UpdateStateToWaitingForStandingOrderSetupRequest{
			ID:         req.GetId(),
			LandlordID: req.GetLandlordId(),
		})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiRto(),
	}, nil
}

func (rs *rtoServer) Cancel(ctx context.Context, req *api.CancelRequest) (*api.RtoResponse, error) {
	resp, err := rs.updateStateToCancelledUseCase.UpdateStateToCancelled(ctx, &request.UpdateStateToCancelledRequest{
		ID: req.GetId(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiRto(),
	}, nil
}

func (rs *rtoServer) GetByPropertyID(ctx context.Context, req *api.GetByPropertyIDRequest) (*api.RtoResponse, error) {
	resp, err := rs.getByPropertyIDUseCase.GetByPropertyID(ctx, &request.GetByPropertyIDRequest{
		PropertyID: req.GetPropertyId(),
		UserID:     req.GetUserId(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiRto(),
	}, nil
}

func (rs *rtoServer) GetByPropertyIDs(ctx context.Context, req *api.GetByPropertyIDsRequest) (
	*api.RtosResponse, error) {
	resp, err := rs.getByPropertyIDsUseCase.GetByPropertyIDs(ctx, &request.GetByPropertyIDsRequest{
		PropertyIDs: req.GetPropertyIds(),
		UserID:      req.GetUserId(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtosResponse{
		Rtos: resp.GetApiRtos(),
	}, nil
}

func (rs *rtoServer) AnswerFromTenant(ctx context.Context, req *api.AnswerFromTenantRequest) (*api.RtoResponse, error) {
	resp, err := rs.answerFromTenantUseCase.AnswerFromTenant(ctx, &request.AnswerFromTenantRequest{
		ID:       req.GetId(),
		TenantID: req.GetTenantId(),
		Accept:   req.GetAcceptValue(),
	})
	if err != nil {
		return nil, err
	}

	return &api.RtoResponse{
		Rto: resp.Rto.CreateApiRto(),
	}, nil
}
