package rto

import (
	"context"
	"fmt"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/lib/mongolocal"
	api "gitlab.mashroom.com/mashroom/backend/offer/api/rto"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	requestservice "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/service"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"
	service "gitlab.mashroom.com/mashroom/backend/offer/internal/app/service/rto"

	"github.com/pkg/errors"
)

const AnswerFromTenantNotValidStateErrorText = "bad rto state for AnswerFromTenant"

type AnswerFromTenantUseCase interface {
	AnswerFromTenant(ctx context.Context, req *request.AnswerFromTenantRequest) (*response.RtoResponse, error)
}

type answerFromTenantUseCase struct {
	repository        domain.Repository
	historyRepository domain.HistoryRepository
	securityService   service.SecurityService
}

func NewAnswerFromTenantUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	securityService service.SecurityService) AnswerFromTenantUseCase {
	return &answerFromTenantUseCase{
		repository:        repository,
		historyRepository: historyRepository,
		securityService:   securityService,
	}
}

func (aftuc *answerFromTenantUseCase) AnswerFromTenant(ctx context.Context, req *request.AnswerFromTenantRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto answerFromTenantUseCase req.Validate")
	}

	resp, err := aftuc.securityService.CanTenantUpdateRto(ctx, &requestservice.CanTenantUpdateRtoRequest{
		ID:       req.ID,
		TenantID: req.TenantID,
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto answerFromTenantUseCase securityService.CanTenantUpdateRto")
	}

	rto := resp.Rto
	if rto.State != api.RtoState_RTO_STATE_WAITING_TENANT_CONFIRMATION {
		return nil, errors.Wrap(fmt.Errorf("rto state must be in state %s, not in state %s",
			api.RtoState_RTO_STATE_WAITING_TENANT_CONFIRMATION, rto.State),
			AnswerFromTenantNotValidStateErrorText)
	}

	if req.Accept {
		rto.State = api.RtoState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP
	} else {
		rto.State = api.RtoState_RTO_STATE_CANCELLED
	}
	now := time.Now()
	rto.UpdatedAt = now
	err = mongolocal.Transaction(ctx, func() error {
		err = aftuc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto answerFromTenantUseCase repository.MustUpdate")
		}

		err = aftuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err, "rto answerFromTenantUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto answerFromTenantUseCase mongolocal.Transaction")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
