package rto

import (
	"context"
	"strings"
	"time"

	account "gitlab.mashroom.com/mashroom/backend/account/api"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	"gitlab.mashroom.com/mashroom/backend/common/lib/mongolocal"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/mongo"
)

const (
	PropertyNotFoundErrorText                   = "property not found"
	WithThisPropertyIDRtoAlreadyExistsErrorText = "with this property id rto already exists"
)

type CreateUseCase interface {
	Create(ctx context.Context, req *request.CreateRequest) (*response.RtoResponse, error)
}

type createUseCase struct {
	repository           domain.Repository
	historyRepository    domain.HistoryRepository
	accountServiceClient account.AccountServiceClient
	listingServiceClient listing.ListingServiceClient
}

func NewCreateUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	accountServiceClient account.AccountServiceClient,
	listingServiceClient listing.ListingServiceClient) CreateUseCase {
	return &createUseCase{
		repository:           repository,
		historyRepository:    historyRepository,
		accountServiceClient: accountServiceClient,
		listingServiceClient: listingServiceClient,
	}
}

func (cuc *createUseCase) Create(ctx context.Context, req *request.CreateRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto createUseCase req.Validate")
	}

	propertyVersion, err := cuc.listingServiceClient.VersionGet(
		grpcctx.WithServiceUser(ctx),
		&listing.VersionGetRequest{
			Spec: listing.VersionSpecPropertyID("", req.PropertyID),
		},
	)
	if err != nil {
		return nil, errors.Wrap(errors.New(PropertyNotFoundErrorText),
			"rto createUseCase listingServiceClient.VersionGet")
	}

	rto, err := cuc.repository.FindByPropertyID(ctx, req.PropertyID)
	if err != nil && !strings.Contains(err.Error(), mongo.ErrNoDocuments.Error()) {
		return nil, errors.Wrap(err, "rto createUseCase repository.FindByPropertyID")
	}
	if rto != nil {
		return nil, errors.Wrap(errors.New(WithThisPropertyIDRtoAlreadyExistsErrorText),
			"rto createUseCase rto != nil")
	}

	landlord, err := cuc.accountServiceClient.Get(ctx, &account.GetRequest{
		Id: req.LandlordID,
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto createUseCase accountServiceClient.Get")
	}

	now := time.Now()
	rto = domain.CreateRto(now, req.RentPlan, req.LandlordID, landlord.GetAccount(), propertyVersion.GetVersion())
	err = mongolocal.Transaction(ctx, func() error {
		err = cuc.repository.Create(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto createUseCase repository.Create")
		}

		err = cuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionCreate, now, rto))
		if err != nil {
			return errors.Wrap(err, "rto createUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto createUseCase mongolocal.Transaction")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
