package rto

import (
	"context"
	"strings"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	"gitlab.mashroom.com/mashroom/backend/common/lib/mongolocal"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/mongo"
)

type CreateByGuestUseCase interface {
	CreateByGuest(ctx context.Context, req *request.CreateByGuestRequest) (*response.RtoResponse, error)
}

type createByGuestUseCase struct {
	repository           domain.Repository
	historyRepository    domain.HistoryRepository
	listingServiceClient listing.ListingServiceClient
}

func NewCreateByGuestUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	listingServiceClient listing.ListingServiceClient) CreateByGuestUseCase {
	return &createByGuestUseCase{
		repository:           repository,
		historyRepository:    historyRepository,
		listingServiceClient: listingServiceClient,
	}
}

func (cbguc *createByGuestUseCase) CreateByGuest(ctx context.Context, req *request.CreateByGuestRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto createByGuestUseCase req.Validate")
	}

	propertyVersion, err := cbguc.listingServiceClient.VersionGet(
		grpcctx.WithServiceUser(ctx),
		&listing.VersionGetRequest{
			Spec: listing.VersionSpecPropertyID("", req.PropertyID),
		},
	)
	if err != nil {
		return nil, errors.Wrap(errors.New(PropertyNotFoundErrorText),
			"rto createByGuestUseCase listingServiceClient.VersionGet")
	}

	rto, err := cbguc.repository.FindByPropertyID(ctx, req.PropertyID)
	if err != nil && !strings.Contains(err.Error(), mongo.ErrNoDocuments.Error()) {
		return nil, errors.Wrap(err, "rto createByGuestUseCase repository.FindByPropertyID")
	}
	if rto != nil {
		return nil, errors.Wrap(errors.New(WithThisPropertyIDRtoAlreadyExistsErrorText),
			"rto createByGuestUseCase rto != nil")
	}

	now := time.Now()
	rto = domain.CreateRto(now, req.RentPlan, "", nil, propertyVersion.GetVersion())
	err = mongolocal.Transaction(ctx, func() error {
		err = cbguc.repository.Create(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto createByGuestUseCase repository.CreateByGuest")
		}

		err = cbguc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionCreate, now, rto))
		if err != nil {
			return errors.Wrap(err, "rto createByGuestUseCase historyRepository.CreateByGuest")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto createByGuestUseCase mongolocal.Transaction")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
