package rto

import (
	"context"

	extras "gitlab.mashroom.com/mashroom/backend/extras/api"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	requestservice "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/service"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"
	appservice "gitlab.mashroom.com/mashroom/backend/offer/internal/app/service"
	service "gitlab.mashroom.com/mashroom/backend/offer/internal/app/service/rto"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/connections"

	"github.com/pkg/errors"
)

const CreateInvoiceTenantNotFound = "rto createInvoiceUseCase tenant not found"

type CreateInvoiceUseCase interface {
	CreateInvoice(ctx context.Context, req *request.CreateInvoiceRequest) (*response.RtoResponse, error)
}

type createInvoiceUseCase struct {
	securityService             service.SecurityService
	invoiceServiceClient        invoice.InvoiceServiceClient
	extrasService               appservice.ExtrasService
	updateLeadTenantUserUseCase UpdateLeadTenantUserUseCase
}

func NewCreateInvoiceUseCase(securityService service.SecurityService,
	invoiceServiceClient invoice.InvoiceServiceClient,
	extrasService appservice.ExtrasService,
	updateLeadTenantUserUseCase UpdateLeadTenantUserUseCase) CreateInvoiceUseCase {
	return &createInvoiceUseCase{
		securityService:             securityService,
		invoiceServiceClient:        invoiceServiceClient,
		extrasService:               extrasService,
		updateLeadTenantUserUseCase: updateLeadTenantUserUseCase,
	}
}

func (ciuc *createInvoiceUseCase) CreateInvoice(ctx context.Context, req *request.CreateInvoiceRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto createInvoiceUseCase req.Validate")
	}

	canLandlordUpdateRtoResponse, err := ciuc.securityService.CanLandlordUpdateRto(ctx,
		&requestservice.CanLandlordUpdateRtoRequest{
			ID:              req.ID,
			LandlordID:      req.LandlordID,
			NeedCheckVeriff: true,
		})
	if err != nil {
		return nil, errors.Wrap(err, "rto createInvoiceUseCase securityService.CanLandlordUpdateRto")
	}
	rto := canLandlordUpdateRtoResponse.Rto

	servicePriceSnapshot, err := ciuc.extrasService.GetServicePriceSnapshot(
		ctx, extras.ProductTenantStandingOrder, rto.Conditions.Price)
	if err != nil {
		return nil, errors.Wrap(err, "rto createInvoiceUseCase extrasService.GetServicePriceSnapshot")
	}

	leadTenantID := rto.GetLeadTenantID()
	if leadTenantID == "" {
		res, err := ciuc.updateLeadTenantUserUseCase.UpdateLeadTenantUser(ctx,
			&request.UpdateLeadTenantUserRequest{Rto: rto})
		if err != nil {
			return nil, errors.Wrap(err, "rto createInvoiceUseCase updateLeadTenantUserUseCase")
		}
		if res == nil {
			return nil, errors.Wrap(err, CreateInvoiceTenantNotFound)
		}
		leadTenantID = rto.GetLeadTenantID()
	}

	_, err = ciuc.invoiceServiceClient.InvoiceCreate(ctx, &invoice.InvoiceCreateRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     rto.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetStandingOrder,
		Periodic:          true,
		Property:          rto.Property,
		Ivr:               rto.PropertyIvr,
		Listing:           rto.PropertyListing,
		User:              leadTenantID,
		Services: []*extras.ServiceSnapshot{
			servicePriceSnapshot,
		},
		NotifyCallback: connections.SelfRemote,
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto createInvoiceUseCase invoiceServiceClient.InvoiceCreate")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
