package rto

import (
	"context"
	"time"

	account "gitlab.mashroom.com/mashroom/backend/account/api"
	"gitlab.mashroom.com/mashroom/backend/common/lib/mongolocal"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	requestservice "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/service"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"
	service "gitlab.mashroom.com/mashroom/backend/offer/internal/app/service/rto"

	"github.com/pkg/errors"
)

type CreateTenancyUseCase interface {
	CreateTenancy(ctx context.Context, req *request.CreateTenancyRequest) (*response.RtoResponse, error)
}

type createTenancyUseCase struct {
	repository           domain.Repository
	historyRepository    domain.HistoryRepository
	securityService      service.SecurityService
	accountServiceClient account.AccountServiceClient
}

func NewCreateTenancyUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	securityService service.SecurityService,
	accountServiceClient account.AccountServiceClient) CreateTenancyUseCase {
	return &createTenancyUseCase{
		repository:           repository,
		historyRepository:    historyRepository,
		securityService:      securityService,
		accountServiceClient: accountServiceClient,
	}
}

func (ctuc *createTenancyUseCase) CreateTenancy(ctx context.Context, req *request.CreateTenancyRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto createTenancyUseCase req.Validate")
	}

	canLandlordUpdateRtoResponse, err := ctuc.securityService.
		CanLandlordUpdateRto(ctx, &requestservice.CanLandlordUpdateRtoRequest{
			ID:         req.ID,
			LandlordID: req.LandlordID,
		})
	if err != nil {
		return nil, errors.Wrap(err, "rto createTenancyUseCase securityService.CanLandlordUpdateRto")
	}
	rto := canLandlordUpdateRtoResponse.Rto

	tenant, err := ctuc.accountServiceClient.GetByEmail(ctx, &account.GetByEmailRequest{
		Email: req.LeadTenant.Email,
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto createTenancyUseCase accountServiceClient.GetByEmail")
	}
	rto.LeadTenant = req.GetLeadTenant()
	if tenant != nil {
		rto.LeadTenant.User = tenant.GetAccount().GetIdValue()
	}
	rto.Tenants = rto.LeadTenant.GetTenants()
	rto.Conditions = req.GetConditions()
	now := time.Now()
	rto.UpdatedAt = now

	err = mongolocal.Transaction(ctx, func() error {
		err = ctuc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto createTenancyUseCase repository.MustUpdate")
		}

		err = ctuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err, "rto createTenancyUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto createTenancyUseCase mongolocal.Transaction")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
