package rto

import (
	"context"
	"fmt"
	"time"

	account "gitlab.mashroom.com/mashroom/backend/account/api"
	"gitlab.mashroom.com/mashroom/backend/common/lib/mongolocal"
	api "gitlab.mashroom.com/mashroom/backend/offer/api/rto"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	requestservice "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/service"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"
	service "gitlab.mashroom.com/mashroom/backend/offer/internal/app/service/rto"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

const (
	UpdateLandlordNotValidStateErrorText               = "bad rto state for Update Landlord"
	UpdateLandlordCanNotChangeCurrentLandlordErrorText = "can't change current landlord"
)

type UpdateLandlordUseCase interface {
	UpdateLandlord(ctx context.Context, req *request.UpdateLandlordRequest) (*response.RtoResponse, error)
}

type updateLandlordUseCase struct {
	repository           domain.Repository
	historyRepository    domain.HistoryRepository
	securityService      service.SecurityService
	accountServiceClient account.AccountServiceClient
}

func NewUpdateLandlordUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	securityService service.SecurityService,
	accountServiceClient account.AccountServiceClient) UpdateLandlordUseCase {
	return &updateLandlordUseCase{
		repository:           repository,
		historyRepository:    historyRepository,
		securityService:      securityService,
		accountServiceClient: accountServiceClient,
	}
}

func (uluc *updateLandlordUseCase) UpdateLandlord(ctx context.Context, req *request.UpdateLandlordRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto updateLandlordUseCase req.Validate")
	}

	rtoID, err := primitive.ObjectIDFromHex(req.ID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateLandlordUseCase primitive.ObjectIDFromHex")
	}

	rto, err := uluc.repository.FindByID(ctx, rtoID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateLandlordUseCase repository.FindByID")
	}

	if rto.State != api.RtoState_RTO_STATE_PENDING_SETUP {
		return nil, errors.Wrap(fmt.Errorf("rto state must be in state %s, not in state %s",
			api.RtoState_RTO_STATE_PENDING_SETUP, rto.State),
			UpdateLandlordNotValidStateErrorText)
	}

	if rto.Landlord != nil && rto.Landlord.User != "" {
		return nil, errors.Wrap(errors.New(UpdateLandlordCanNotChangeCurrentLandlordErrorText),
			`rto.Landlord != nil && rto.Landlord.User != ""`)
	}

	err = uluc.securityService.IsVeriffValid(ctx, &requestservice.IsVeriffValidRequest{
		LandlordID: req.LandlordID,
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateLandlordUseCase securityService.IsVeriffValid")
	}

	landlord, err := uluc.accountServiceClient.Get(ctx, &account.GetRequest{
		Id: req.LandlordID,
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateLandlordUseCase accountServiceClient.Get")
	}

	now := time.Now()
	rto.UpdatedAt = now
	rto.Landlord = domain.CreateLandlord(req.LandlordID, landlord.GetAccount())
	err = mongolocal.Transaction(ctx, func() error {
		err = uluc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto updateLandlordUseCase repository.MustUpdate")
		}

		err = uluc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err, "rto updateLandlordUseCase historyRepository.UpdateLandlord")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateLandlordUseCase mongolocal.Transaction")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
