package rto

import (
	"context"
	"time"

	account "gitlab.mashroom.com/mashroom/backend/account/api"
	"gitlab.mashroom.com/mashroom/backend/common/lib/mongolocal"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"

	"github.com/pkg/errors"
)

const UpdateLeadTenantUserTenantNotFound = "rto updateLeadTenantUserUseCase tenant not found"

type UpdateLeadTenantUserUseCase interface {
	UpdateLeadTenantUser(ctx context.Context, req *request.UpdateLeadTenantUserRequest) (*response.RtoResponse, error)
}

type updateLeadTenantUserUseCase struct {
	repository           domain.Repository
	historyRepository    domain.HistoryRepository
	accountServiceClient account.AccountServiceClient
}

func NewUpdateLeadTenantUserUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	accountServiceClient account.AccountServiceClient) UpdateLeadTenantUserUseCase {
	return &updateLeadTenantUserUseCase{
		repository:           repository,
		historyRepository:    historyRepository,
		accountServiceClient: accountServiceClient,
	}
}

func (ultuuc *updateLeadTenantUserUseCase) UpdateLeadTenantUser(ctx context.Context,
	req *request.UpdateLeadTenantUserRequest) (*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto updateLeadTenantUserUseCase req.Validate")
	}

	tenant, err := ultuuc.accountServiceClient.GetByEmail(ctx, &account.GetByEmailRequest{
		Email: req.Rto.LeadTenant.EmailPrimary,
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateLeadTenantUserUseCase accountServiceClient.GetByEmail")
	}
	if tenant != nil && tenant.GetAccount().GetIdValue() != "" {
		req.Rto.LeadTenant.User = tenant.GetAccount().GetIdValue()
		if len(req.Rto.Tenants) > 0 {
			req.Rto.Tenants[0].User = tenant.GetAccount().GetIdValue()
		}
		now := time.Now()
		req.Rto.UpdatedAt = now

		err = mongolocal.Transaction(ctx, func() error {
			err = ultuuc.repository.MustUpdate(ctx, req.Rto)
			if err != nil {
				return errors.Wrap(err, "rto updateLeadTenantUserUseCase repository.MustUpdate")
			}

			err = ultuuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, req.Rto))
			if err != nil {
				return errors.Wrap(err, "rto updateLeadTenantUserUseCase historyRepository.Create")
			}

			return nil
		})
		if err != nil {
			return nil, errors.Wrap(err, "rto updateLeadTenantUserUseCase mongolocal.Transaction")
		}

		return &response.RtoResponse{
			Rto: req.Rto,
		}, nil
	}

	return nil, errors.Wrap(errors.New(UpdateLeadTenantUserTenantNotFound), UpdateLeadTenantUserTenantNotFound)
}
