package rto

import (
	"context"
	"fmt"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/lib/mongolocal"
	api "gitlab.mashroom.com/mashroom/backend/offer/api/rto"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	requestservice "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/service"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"
	service "gitlab.mashroom.com/mashroom/backend/offer/internal/app/service/rto"

	"github.com/pkg/errors"
)

const UpdateRentPlanNotValidStateErrorText = "bad rto state for UpdateRentPlan"

type UpdateRentPlanUseCase interface {
	UpdateRentPlan(ctx context.Context, req *request.UpdateRentPlanRequest) (*response.RtoResponse, error)
}

type updateRentPlanUseCase struct {
	repository        domain.Repository
	historyRepository domain.HistoryRepository
	securityService   service.SecurityService
}

func NewUpdateRentPlanUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	securityService service.SecurityService) UpdateRentPlanUseCase {
	return &updateRentPlanUseCase{
		repository:        repository,
		historyRepository: historyRepository,
		securityService:   securityService,
	}
}

func (urpuc *updateRentPlanUseCase) UpdateRentPlan(ctx context.Context, req *request.UpdateRentPlanRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto updateRentPlanUseCase req.Validate")
	}

	canLandlordUpdateRtoResponse, err := urpuc.securityService.CanLandlordUpdateRto(ctx,
		&requestservice.CanLandlordUpdateRtoRequest{
			ID:         req.ID,
			LandlordID: req.LandlordID,
		})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateRentPlanUseCase securityService.CanLandlordUpdateRto")
	}
	rto := canLandlordUpdateRtoResponse.Rto

	if rto.State != api.RtoState_RTO_STATE_PENDING_SETUP {
		return nil, errors.Wrap(fmt.Errorf("rto state must be in state %s, not in state %s",
			api.RtoState_RTO_STATE_PENDING_SETUP, rto.State), UpdateRentPlanNotValidStateErrorText)
	}

	rto.RentPlan = req.RentPlan
	now := time.Now()
	rto.UpdatedAt = now

	err = mongolocal.Transaction(ctx, func() error {
		err = urpuc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto updateRentPlanUseCase repository.MustUpdate")
		}

		err = urpuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err, "rto updateRentPlanUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateRentPlanUseCase mongolocal.Transaction")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
