package rto

import (
	"context"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/lib/mongolocal"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	requestservice "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/service"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"
	service "gitlab.mashroom.com/mashroom/backend/offer/internal/app/service/rto"

	"github.com/pkg/errors"
)

type UpdateStateUseCase interface {
	UpdateState(ctx context.Context, req *request.UpdateStateRequest) (*response.RtoResponse, error)
}

type updateStateUseCase struct {
	repository        domain.Repository
	historyRepository domain.HistoryRepository
	securityService   service.SecurityService
}

func NewUpdateStateUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	securityService service.SecurityService) UpdateStateUseCase {
	return &updateStateUseCase{
		repository:        repository,
		historyRepository: historyRepository,
		securityService:   securityService,
	}
}

func (usuc *updateStateUseCase) UpdateState(ctx context.Context, req *request.UpdateStateRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateUseCase req.Validate")
	}

	canLandlordUpdateRtoResponse, err := usuc.securityService.CanLandlordUpdateRto(ctx,
		&requestservice.CanLandlordUpdateRtoRequest{
			ID:              req.ID,
			LandlordID:      req.LandlordID,
			NeedCheckVeriff: true,
		})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateUseCase securityService.CanLandlordUpdateRto")
	}
	rto := canLandlordUpdateRtoResponse.Rto

	rto.State = req.State
	now := time.Now()
	rto.UpdatedAt = now

	err = mongolocal.Transaction(ctx, func() error {
		err = usuc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto updateStateUseCase repository.MustUpdate")
		}

		err = usuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err, "rto updateStateUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateUseCase mongolocal.Transaction")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
