package rto

import (
	"context"
	"fmt"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/lib/mongolocal"
	api "gitlab.mashroom.com/mashroom/backend/offer/api/rto"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	requestservice "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/service"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"
	servicerto "gitlab.mashroom.com/mashroom/backend/offer/internal/app/service/rto"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

const UpdateStateToActiveNotValidStateErrorText = "bad rto state for Active"

// todo: RtoState_RTO_STATE_WAITING_FOR_FIRST_RENT_PAYMENT
// first month rent
type UpdateStateToActiveUseCase interface {
	UpdateStateToActive(ctx context.Context, req *request.UpdateStateToActiveRequest) (*response.RtoResponse, error)
}

type updateStateToActiveUseCase struct {
	repository          domain.Repository
	historyRepository   domain.HistoryRepository
	notificationService servicerto.NotificationService
}

func NewUpdateStateToActiveUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	notificationService servicerto.NotificationService) UpdateStateToActiveUseCase {
	return &updateStateToActiveUseCase{
		repository:          repository,
		historyRepository:   historyRepository,
		notificationService: notificationService,
	}
}

func (ustauc *updateStateToActiveUseCase) UpdateStateToActive(ctx context.Context,
	req *request.UpdateStateToActiveRequest) (*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateToActiveUseCase req.Validate")
	}

	rtoID, err := primitive.ObjectIDFromHex(req.ID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToActiveUseCase primitive.ObjectIDFromHex")
	}

	rto, err := ustauc.repository.FindByID(ctx, rtoID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToActiveUseCase repository.FindByID")
	}

	if rto.State != api.RtoState_RTO_STATE_WAITING_FOR_FIRST_RENT_PAYMENT {
		return nil, errors.Wrap(fmt.Errorf("rto state must be in state %s, not in state %s",
			api.RtoState_RTO_STATE_WAITING_FOR_FIRST_RENT_PAYMENT, rto.State),
			UpdateStateToActiveNotValidStateErrorText)
	}

	rto.State = api.RtoState_RTO_STATE_ACTIVE
	now := time.Now()
	rto.UpdatedAt = now

	err = mongolocal.Transaction(ctx, func() error {
		err = ustauc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToActiveUseCase repository.MustUpdate")
		}

		err = ustauc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToActiveUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToActiveUseCase mongolocal.Transaction")
	}

	err = ustauc.notificationService.SendActive(ctx,
		&requestservice.SendActiveRequest{
			LandlordID: rto.Landlord.User,
			PropertyID: rto.Property,
			RentPlan:   rto.RentPlan,
		})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateToActiveUseCase notificationService.SendActive")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
