package rto

import (
	"context"
	"fmt"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/lib/mongolocal"
	api "gitlab.mashroom.com/mashroom/backend/offer/api/rto"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

const (
	UpdateStateToCancelledNotValidStateErrorText      = "bad state for Cancelled"
	UpdateStateToCancelledNotValidRtoEndDateErrorText = "bad rto end date for Cancelled"
)

type UpdateStateToCancelledUseCase interface {
	UpdateStateToCancelled(ctx context.Context, req *request.UpdateStateToCancelledRequest) (
		*response.RtoResponse, error)
}

type updateStateToCancelledUseCase struct {
	repository        domain.Repository
	historyRepository domain.HistoryRepository
}

func NewUpdateStateToCancelledUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository) UpdateStateToCancelledUseCase {
	return &updateStateToCancelledUseCase{
		repository:        repository,
		historyRepository: historyRepository,
	}
}

func (ustfuc *updateStateToCancelledUseCase) UpdateStateToCancelled(ctx context.Context,
	req *request.UpdateStateToCancelledRequest) (*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateToCancelledUseCase req.Validate")
	}

	rtoID, err := primitive.ObjectIDFromHex(req.ID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToCancelledUseCase primitive.ObjectIDFromHex")
	}

	rto, err := ustfuc.repository.FindByID(ctx, rtoID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToCancelledUseCase repository.FindByID")
	}

	if !(rto.State == api.RtoState_RTO_STATE_ACTIVE ||
		rto.State == api.RtoState_RTO_STATE_ACTIVE_ROLLING_TENANCY ||
		rto.State == api.RtoState_RTO_STATE_ENDING) {
		return nil, errors.Wrap(fmt.Errorf("rto state must be in one of the next states "+
			"(%s, %s, %s), not in state %s", api.RtoState_RTO_STATE_ACTIVE,
			api.RtoState_RTO_STATE_ACTIVE_ROLLING_TENANCY, api.RtoState_RTO_STATE_ENDING, rto.State),
			UpdateStateToCancelledNotValidStateErrorText)
	}

	now := time.Now()
	if !rto.Conditions.EndDate.IsZero() && rto.Conditions.EndDate.Sub(now).Round(time.Hour) < 60*24*time.Hour {
		return nil, errors.Wrap(fmt.Errorf("rto end date (%s) must be greater than current date (%s) on 2 months",
			rto.Conditions.EndDate.String(), now.String()),
			UpdateStateToCancelledNotValidRtoEndDateErrorText)
	}

	rto.State = api.RtoState_RTO_STATE_CANCELLED
	rto.UpdatedAt = now

	err = mongolocal.Transaction(ctx, func() error {
		err = ustfuc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto updateStateToCancelledUseCase repository.MustUpdate")
		}

		err = ustfuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err, "rto updateStateToCancelledUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateToCancelledUseCase mongolocal.Transaction")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
