package rto

import (
	"context"
	"fmt"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/lib/mongolocal"
	api "gitlab.mashroom.com/mashroom/backend/offer/api/rto"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

const UpdateStateToEndingNotValidStateErrorText = "bad rto state for Ending"

type UpdateStateToEndingUseCase interface {
	UpdateStateToEnding(ctx context.Context, req *request.UpdateStateToEndingRequest) (*response.RtoResponse, error)
}

type updateStateToEndingUseCase struct {
	repository        domain.Repository
	historyRepository domain.HistoryRepository
}

func NewUpdateStateToEndingUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository) UpdateStateToEndingUseCase {
	return &updateStateToEndingUseCase{
		repository:        repository,
		historyRepository: historyRepository,
	}
}

/*
todo: Your rent collection service will end after the last rent payment has been received
*/
func (usteuc *updateStateToEndingUseCase) UpdateStateToEnding(ctx context.Context,
	req *request.UpdateStateToEndingRequest) (*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateToEndingUseCase req.Validate")
	}

	rtoID, err := primitive.ObjectIDFromHex(req.ID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToEndingUseCase primitive.ObjectIDFromHex")
	}

	rto, err := usteuc.repository.FindByID(ctx, rtoID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToEndingUseCase repository.FindByID")
	}

	if rto.State != api.RtoState_RTO_STATE_ACTIVE_ROLLING_TENANCY {
		return nil, errors.Wrap(fmt.Errorf("rto state must be in state %s, not in state %s",
			api.RtoState_RTO_STATE_ACTIVE_ROLLING_TENANCY, rto.State),
			UpdateStateToEndingNotValidStateErrorText)
	}

	rto.State = api.RtoState_RTO_STATE_ENDING
	now := time.Now()
	rto.UpdatedAt = now

	err = mongolocal.Transaction(ctx, func() error {
		err = usteuc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToEndingUseCase repository.MustUpdate")
		}

		err = usteuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToEndingUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToEndingUseCase mongolocal.Transaction")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
