package rto

import (
	"context"
	"fmt"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/lib/mongolocal"
	api "gitlab.mashroom.com/mashroom/backend/offer/api/rto"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

const UpdateStateToFinishedNotValidStateErrorText = "bad state for Finished"

type UpdateStateToFinishedUseCase interface {
	UpdateStateToFinished(ctx context.Context, req *request.UpdateStateToFinishedRequest) (*response.RtoResponse, error)
}

type updateStateToFinishedUseCase struct {
	repository        domain.Repository
	historyRepository domain.HistoryRepository
}

func NewUpdateStateToFinishedUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository) UpdateStateToFinishedUseCase {
	return &updateStateToFinishedUseCase{
		repository:        repository,
		historyRepository: historyRepository,
	}
}

/*
todo: Your rent collection service has ended
*/
func (ustfuc *updateStateToFinishedUseCase) UpdateStateToFinished(ctx context.Context,
	req *request.UpdateStateToFinishedRequest) (*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateToFinishedUseCase req.Validate")
	}

	rtoID, err := primitive.ObjectIDFromHex(req.ID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToFinishedUseCase primitive.ObjectIDFromHex")
	}

	rto, err := ustfuc.repository.FindByID(ctx, rtoID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToFinishedUseCase repository.FindByID")
	}

	if !(rto.State == api.RtoState_RTO_STATE_ACTIVE ||
		rto.State == api.RtoState_RTO_STATE_ACTIVE_ROLLING_TENANCY ||
		rto.State == api.RtoState_RTO_STATE_ENDING) {
		return nil, errors.Wrap(fmt.Errorf("rto state must be in one of the next states "+
			"(%s, %s, %s), not in state %s", api.RtoState_RTO_STATE_ACTIVE,
			api.RtoState_RTO_STATE_ACTIVE_ROLLING_TENANCY, api.RtoState_RTO_STATE_ENDING, rto.State),
			UpdateStateToFinishedNotValidStateErrorText)
	}

	rto.State = api.RtoState_RTO_STATE_FINISHED
	now := time.Now()
	rto.UpdatedAt = now

	err = mongolocal.Transaction(ctx, func() error {
		err = ustfuc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err, "rto updateStateToFinishedUseCase repository.MustUpdate")
		}

		err = ustfuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err, "rto updateStateToFinishedUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToFinishedUseCase mongolocal.Transaction")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
