package rto

import (
	"context"
	"fmt"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/lib/mongolocal"
	api "gitlab.mashroom.com/mashroom/backend/offer/api/rto"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	requestservice "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/service"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"
	servicerto "gitlab.mashroom.com/mashroom/backend/offer/internal/app/service/rto"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

const UpdateStateToWaitingForFirstRentPaymentNotValidStateErrorText = "bad rto state for WaitingForFirstRentPayment"

type UpdateStateToWaitingForFirstRentPaymentUseCase interface {
	UpdateStateToWaitingForFirstRentPayment(ctx context.Context,
		req *request.UpdateStateToWaitingForFirstRentPaymentRequest) (*response.RtoResponse, error)
}

type updateStateToWaitingForFirstRentPaymentUseCase struct {
	repository          domain.Repository
	historyRepository   domain.HistoryRepository
	notificationService servicerto.NotificationService
}

func NewUpdateStateToWaitingForFirstRentPaymentUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	notificationService servicerto.NotificationService) UpdateStateToWaitingForFirstRentPaymentUseCase {
	return &updateStateToWaitingForFirstRentPaymentUseCase{
		repository:          repository,
		historyRepository:   historyRepository,
		notificationService: notificationService,
	}
}

func (ustwffrpuc *updateStateToWaitingForFirstRentPaymentUseCase) UpdateStateToWaitingForFirstRentPayment(
	ctx context.Context, req *request.UpdateStateToWaitingForFirstRentPaymentRequest) (
	*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateToWaitingForFirstRentPaymentUseCase req.Validate")
	}

	rtoID, err := primitive.ObjectIDFromHex(req.ID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToWaitingForFirstRentPaymentUseCase primitive.ObjectIDFromHex")
	}

	rto, err := ustwffrpuc.repository.FindByID(ctx, rtoID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToWaitingForFirstRentPaymentUseCase repository.FindByID")
	}

	if rto.State != api.RtoState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP {
		return nil, errors.Wrap(fmt.Errorf("rto state must be in state %s, not in state %s",
			api.RtoState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP, rto.State),
			UpdateStateToWaitingForFirstRentPaymentNotValidStateErrorText)
	}

	rto.State = api.RtoState_RTO_STATE_WAITING_FOR_FIRST_RENT_PAYMENT
	now := time.Now()
	rto.UpdatedAt = now

	err = mongolocal.Transaction(ctx, func() error {
		err = ustwffrpuc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToWaitingForFirstRentPaymentUseCase repository.MustUpdate")
		}

		err = ustwffrpuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToWaitingForFirstRentPaymentUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToWaitingForFirstRentPaymentUseCase mongolocal.Transaction")
	}

	err = ustwffrpuc.notificationService.SendWaitingForFirstRentPayment(ctx,
		&requestservice.SendWaitingForFirstRentPaymentRequest{
			LandlordID: rto.Landlord.User,
			PropertyID: rto.Property,
		})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateToWaitingForFirstRentPaymentUseCase "+
			"notificationService.SendWaitingForFirstRentPayment")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
