package rto

import (
	"context"
	"fmt"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/lib/mongolocal"
	api "gitlab.mashroom.com/mashroom/backend/offer/api/rto"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	requestservice "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/service"
	request "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/request/usecase"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/resource/response"
	servicerto "gitlab.mashroom.com/mashroom/backend/offer/internal/app/service/rto"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

const (
	UpdateStateToWaitingForStandingOrderSetupNotValidStateErrorText = "bad rto state for WaitingForStandingOrderSetup"
	UpdateStateToWaitingForStandingOrderSetupTenantNotFound         = "WaitingForStandingOrderSetup tenant not found"
)

type UpdateStateToWaitingForStandingOrderSetupUseCase interface {
	UpdateStateToWaitingForStandingOrderSetup(ctx context.Context,
		req *request.UpdateStateToWaitingForStandingOrderSetupRequest) (*response.RtoResponse, error)
}

type updateStateToWaitingForStandingOrderSetupUseCase struct {
	repository                  domain.Repository
	historyRepository           domain.HistoryRepository
	updateLeadTenantUserUseCase UpdateLeadTenantUserUseCase
	notificationService         servicerto.NotificationService
}

func NewUpdateStateToWaitingForStandingOrderSetupUseCase(repository domain.Repository,
	historyRepository domain.HistoryRepository,
	updateLeadTenantUserUseCase UpdateLeadTenantUserUseCase,
	notificationService servicerto.NotificationService) UpdateStateToWaitingForStandingOrderSetupUseCase {
	return &updateStateToWaitingForStandingOrderSetupUseCase{
		repository:                  repository,
		historyRepository:           historyRepository,
		updateLeadTenantUserUseCase: updateLeadTenantUserUseCase,
		notificationService:         notificationService,
	}
}

func (ustwfsosuc *updateStateToWaitingForStandingOrderSetupUseCase) UpdateStateToWaitingForStandingOrderSetup(
	ctx context.Context, req *request.UpdateStateToWaitingForStandingOrderSetupRequest) (*response.RtoResponse, error) {
	err := req.Validate()
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateToWaitingForStandingOrderSetupUseCase req.Validate")
	}

	rtoID, err := primitive.ObjectIDFromHex(req.ID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToWaitingForStandingOrderSetupUseCase primitive.ObjectIDFromHex")
	}

	rto, err := ustwfsosuc.repository.FindByID(ctx, rtoID)
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToWaitingForStandingOrderSetupUseCase repository.FindByID")
	}

	if rto.State != api.RtoState_RTO_STATE_PENDING_SETUP {
		return nil, errors.Wrap(fmt.Errorf("rto state must be in state %s, not in state %s",
			api.RtoState_RTO_STATE_PENDING_SETUP, rto.State),
			UpdateStateToWaitingForStandingOrderSetupNotValidStateErrorText)
	}

	rto.State = api.RtoState_RTO_STATE_WAITING_FOR_STANDING_ORDER_SETUP
	now := time.Now()
	rto.UpdatedAt = now

	err = mongolocal.Transaction(ctx, func() error {
		err = ustwfsosuc.repository.MustUpdate(ctx, rto)
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToWaitingForStandingOrderSetupUseCase repository.MustUpdate")
		}

		err = ustwfsosuc.historyRepository.Create(ctx, domain.CreateHistory(domain.HistoryActionUpdate, now, rto))
		if err != nil {
			return errors.Wrap(err,
				"rto updateStateToWaitingForStandingOrderSetupUseCase historyRepository.Create")
		}

		return nil
	})
	if err != nil {
		return nil, errors.Wrap(err,
			"rto updateStateToWaitingForStandingOrderSetupUseCase mongolocal.Transaction")
	}

	leadTenantID := rto.GetLeadTenantID()
	if leadTenantID == "" {
		res, err := ustwfsosuc.updateLeadTenantUserUseCase.UpdateLeadTenantUser(ctx,
			&request.UpdateLeadTenantUserRequest{Rto: rto})
		if err != nil {
			return nil, errors.Wrap(err,
				"rto updateStateToWaitingForStandingOrderSetupUseCase updateLeadTenantUserUseCase")
		}
		if res == nil {
			return nil, errors.Wrap(err, UpdateStateToWaitingForStandingOrderSetupTenantNotFound)
		}
		leadTenantID = rto.GetLeadTenantID()
	}

	err = ustwfsosuc.notificationService.SendWaitingForStandingOrderSetup(ctx,
		&requestservice.SendWaitingForStandingOrderSetupRequest{
			ID:         rto.ID.Hex(),
			LandlordID: rto.Landlord.User,
			TenantID:   leadTenantID,
			PropertyID: rto.Property,
		})
	if err != nil {
		return nil, errors.Wrap(err, "rto updateStateToWaitingForStandingOrderSetupUseCase "+
			"notificationService.SendWaitingForStandingOrderSetup")
	}

	return &response.RtoResponse{
		Rto: rto,
	}, nil
}
