package oneoff

import (
	"context"
	"encoding/json"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"

	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/connections"
	"gitlab.mashroom.com/mashroom/backend/offer/model"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"google.golang.org/protobuf/types/known/durationpb"
)

// InvocationCheckHoldingDeposit holds data for holding deposit check payment status invocation
type InvocationCheckHoldingDeposit struct {
	OfferID string        `json:"offer_id"`
	Expiry  time.Duration `json:"expirty"`
}

func consumeReminderInvocationCheckHoldingDeposit(
	ctx context.Context,
	inv *Invocation,
) (err error) {
	var depositcheck InvocationCheckHoldingDeposit

	err = json.Unmarshal(inv.Data, &depositcheck)
	if err != nil {
		return
	}

	id, err := primitive.ObjectIDFromHex(depositcheck.OfferID)
	if err != nil {
		return
	}

	off, err := model.GetOfferByID(ctx, id)
	if err != nil {
		return
	}

	if off.Test {
		ctx = context.WithValue(ctx, grpcctx.CtxIsTestUser, true)
	}

	switch off.State {
	case offer.OfferState_CANCELED, offer.OfferState_DENIED:
		log.Infof(
			"holding deposit: offer %v has state %v, %v notification skipped",
			id.Hex(),
			off.State,
			depositcheck.Expiry,
		)

		return nil
	}

	res, _ := connections.InvoiceService.InvoiceList(ctx, &invoice.InvoiceListRequest{
		CorrelationType:   "offer",
		CorrelationId:     off.ID.Hex(),
		CorrelationTarget: "holding-deposit",
		Statuses: []invoice.InvoiceStatus{
			invoice.InvoiceStatus_INVOICE_FULFILLED,
		},
	})

	if len(res.GetInvoices()) > 0 {
		return nil
	}

	var target notification.Role

	var userID, targetID string

	var action notification.Action

	var dur *durationpb.Duration

	switch depositcheck.Expiry {
	case time.Hour * 12:
		action = notification.Action_ACTION_REMIND
		target = notification.Role_ROLE_TENANT
		targetID = off.TenantUser()
		userID = off.LandlordUser()
		dur = durationpb.New(depositcheck.Expiry)
	case time.Hour * 24:
		// TODO remove it in future. just to handle old reminders
	}

	_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity:   notification.Entity_ENTITY_OFFER_HOLDING_DEPOSIT_PAYMENT,
		Action:   action,
		Duration: dur,
		Targets: []*notification.Target{
			{
				User: targetID,
				Role: target,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: off.Property,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
				EmbedValue: off.ID.Hex(),
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
				EmbedValue: userID,
			},
		},
	})

	return err
}
