package status

import (
	"context"
	"errors"
	"fmt"
	"math"
	"strings"
	"time"

	closecom "gitlab.mashroom.com/mashroom/backend/closecom/api"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/util"
	extras "gitlab.mashroom.com/mashroom/backend/extras/api"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/connections"
	"gitlab.mashroom.com/mashroom/backend/offer/model"
	"google.golang.org/protobuf/types/known/timestamppb"
)

// TryCreateSignedInvoices will attempt to create invoices for signed or
// for awaiting tenancy offer if it has rent plan selected
func TryCreateSignedInvoices(ctx context.Context, m *model.OfferModel, isImitationMode bool) error {
	if m.State != offer.OfferState_SIGNED && m.State != offer.OfferState_AWAITING_TENANCY {
		if isImitationMode {
			return errors.New("available only for signed or awaiting tenancy states")
		}

		return nil
	}

	if m.RentPlan == 0 || (m.RentPlan == offer.RentPlan_RENT_PLAN_LEGACY && !isImitationMode) {
		if isImitationMode {
			return errors.New("rent plan is not selected")
		}

		return nil
	}

	if m.SignedInvoicesGenerated {
		if isImitationMode {
			return errors.New("invoices already generated")
		}

		return nil
	}

	return CreateSignedInvoices(ctx, m, isImitationMode)
}

// CreateSignedInvoices creates invoices for signed offer with rent plan selected
func CreateSignedInvoices(ctx context.Context, m *model.OfferModel, isImitationMode bool) error {
	if isImitationMode {
		setDefaultValuesInOldOffersForRentInvoices(m)
	}

	if m.Conditions == nil ||
		len(m.Tenants) == 0 ||
		m.Conditions.PaymentFrequency == nil ||
		m.Conditions.Price == nil ||
		m.Conditions.TermMonths == nil {
		return fmt.Errorf(
			"%w: empty condition payment frequency/price/term months/tenants",
			grpcerr.InvalidState,
		)
	}

	var errs []string

	err := createDepositReplacementInvoice(ctx, m, isImitationMode)
	if err != nil {
		errs = append(errs, "generating deposit replacement: "+err.Error())
	}

	if m.RentPlan == offer.RentPlan_RENT_PLAN_MANUAL {
		if len(errs) > 0 {
			return errors.New(strings.Join(errs, "; "))
		}

		m.SignedInvoicesGenerated = true

		err = model.UpdateOffer(ctx, m)
		if err != nil {
			return err
		}

		return SetPropertyRented(ctx, m)
	}

	err = createFirstMonthPaymentInvoice(ctx, m, isImitationMode)
	if err != nil {
		errs = append(errs, "generating first month rent: "+err.Error())
	}

	// create standing order request invoice
	err = createStandingOrderInvoice(ctx, m, isImitationMode)
	if err != nil {
		errs = append(errs, "generating standing order: "+err.Error())
	}

	if len(errs) > 0 {
		return errors.New(strings.Join(errs, "; "))
	}

	ExportClosecomOpportunity(ctx, m, closecom.OfferStatus_OFFER_STATUS_AWAITING_PAYMENTS)

	m.SignedInvoicesGenerated = true

	return nil
}

func setDefaultValuesInOldOffersForRentInvoices(m *model.OfferModel) {
	if m.Conditions == nil {
		m.Conditions = &model.OfferConditionsModel{}
	}

	if m.Conditions.MoveinDate == nil {
		m.Conditions.MoveinDate = util.PtrTime(time.Now())
	}

	if m.Conditions.TermMonths == nil {
		m.Conditions.TermMonths = util.PtrInt32(12)
	}

	if m.Conditions.PaymentFrequency == nil {
		m.Conditions.PaymentFrequency = util.PtrInt32(1)
	}

	if m.Conditions.FirstRentPaymentDate == nil {
		m.Conditions.FirstRentPaymentDate = m.Conditions.MoveinDate
	}

	if m.Conditions.RentPaymentDay == nil {
		m.Conditions.RentPaymentDay = util.PtrInt32(paymentDayClamp(m.Conditions.MoveinDate.Day()))
	}

	if m.Conditions.BreakClause == nil {
		m.Conditions.BreakClause = util.PtrInt32(0)
	}

	if m.Conditions.DepositType == offer.DepositType_UNKNOWN_DEPOSIT_TYPE {
		m.Conditions.DepositType = offer.DepositType_DEPOSIT_TRADITIONAL_5WEEK
	}
}

func paymentDayClamp(d int) (day int32) {
	if d < int(model.FirstRentPaymentDay) {
		day = model.FirstRentPaymentDay
	}

	if d > int(model.LastRentPaymentDay) {
		day = model.LastRentPaymentDay
	}

	return
}

var dataPeriodicPaymentFrequency = map[int32]invoice.PeriodicPaymentFrequency{
	1:  invoice.PeriodicPaymentFrequency_FREQUENCY_MONTHLY,
	3:  invoice.PeriodicPaymentFrequency_FREQUENCY_QUARTERLY,
	6:  invoice.PeriodicPaymentFrequency_FREQUENCY_TWICE_A_YEAR,
	12: invoice.PeriodicPaymentFrequency_FREQUENCY_YEARLY,
}

func firstMonthPayment(
	ctx context.Context,
	off *model.OfferModel,
) (amount float64, isTraditionalDeposit bool, err error) {
	if off.Conditions.MoveinDate == nil || util.UnPtrInt32(off.Conditions.RentPaymentDay) == 0 {
		err = fmt.Errorf(
			"%w: empty condition movein date/rent payment day",
			grpcerr.InvalidState,
		)

		return
	}

	condPrice := util.UnPtrFloat64(off.Conditions.Price)
	invoiceResp, err := connections.InvoiceService.InvoiceGet(ctx, &invoice.InvoiceGetRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     off.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetHoldingDeposit,
		Type:              invoice.InvoiceType_INVOICE_TYPE_PAYMENT,
		Property:          off.Property,
	})
	if err != nil || invoiceResp.GetInvoice() == nil {
		err = ErrPaidDepositNotFound

		return
	}

	holding := invoiceResp.GetInvoice().GetAmount()
	firstMonth := condPrice - holding
	weekly := condPrice * MonthlyToWeeklyFactor

	if *off.Conditions.PaymentFrequency > *off.Conditions.TermMonths {
		if *off.Conditions.TermMonths > 1 {
			firstMonth = condPrice * (float64(*off.Conditions.TermMonths) - 1)
		}
	} else {
		if *off.Conditions.PaymentFrequency > 1 {
			firstMonth += condPrice * (float64(*off.Conditions.PaymentFrequency) - 1)
		}
	}

	amount = firstMonth

	if off.Conditions.DepositType == offer.DepositType_DEPOSIT_TRADITIONAL_5WEEK {
		amount += weekly * 5.0
		isTraditionalDeposit = true
	}

	return
}

func createHoldingDepositPayoutInvoice(
	ctx context.Context,
	old *model.OfferModel,
) error {
	inv, err := connections.InvoiceService.InvoiceList(ctx, &invoice.InvoiceListRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     old.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetHoldingDeposit,
		Type:              invoice.InvoiceType_INVOICE_TYPE_PAYOUT,
		Statuses: []invoice.InvoiceStatus{
			invoice.InvoiceStatus_INVOICE_OPEN,
			invoice.InvoiceStatus_INVOICE_FULFILLED,
		},
	})
	if err != nil {
		return err
	}

	if len(inv.GetInvoices()) > 0 {
		return nil
	}

	invoiceResp, err := connections.InvoiceService.InvoiceGet(ctx, &invoice.InvoiceGetRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     old.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetHoldingDeposit,
		Type:              invoice.InvoiceType_INVOICE_TYPE_PAYMENT,
		Property:          old.Property,
	})
	if err != nil || invoiceResp.GetInvoice() == nil {
		err = ErrPaidDepositNotFound

		return err
	}

	var svcs []*extras.ServiceSnapshot

	for _, s := range invoiceResp.GetInvoice().GetServices() {
		svcs = append(svcs, s.GetService())
	}

	_, err = connections.InvoiceService.InvoiceCreate(ctx, &invoice.InvoiceCreateRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     old.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetHoldingDeposit,
		Type:              invoice.InvoiceType_INVOICE_TYPE_PAYOUT,
		Property:          old.Property,
		Ivr:               old.PropertyIvr,
		Listing:           old.PropertyListing,
		User:              old.TenantUser(),
		Services:          svcs,
		NotifyCallback:    connections.SelfRemote,
	})
	if err != nil {
		return err
	}

	return nil
}

// createFirstMonthPaymentInvoice creates tenant first month payment invoice after signing
func createFirstMonthPaymentInvoice(
	ctx context.Context,
	offerModel *model.OfferModel,
	isImitationMode bool,
) error {
	if offerModel.OriginalOffer != "" {
		return nil
	}

	amount, isTraditionalDeposit, err := firstMonthPayment(ctx, offerModel)
	if err != nil {
		return err
	}

	title := "First month rent minus holding deposit"

	if isTraditionalDeposit {
		title = "First month rent plus 5 weeks rent, minus holding deposit"
	}

	if amount == 0 {
		return nil
	}

	service, err := extras.GetServicePriceSnapshot(
		ctx,
		connections.ExtrasService,
		extras.ProductTenantFirstMonthRent,
		amount,
	)
	if err != nil {
		return err
	}

	service.Title = title

	// check invoice existence
	inv, err := connections.InvoiceService.InvoiceList(ctx, &invoice.InvoiceListRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     offerModel.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetTenantServices,
		Statuses: []invoice.InvoiceStatus{
			invoice.InvoiceStatus_INVOICE_OPEN,
			invoice.InvoiceStatus_INVOICE_FULFILLED,
		},
	})
	if err != nil {
		return err
	}

	if len(inv.GetInvoices()) > 0 {
		if isImitationMode {
			_, err = connections.InvoiceService.GenerateInvoicePayments(ctx, &invoice.GenerateInvoicePaymentsRequest{
				InvoiceId: inv.GetInvoices()[0].GetInvoiceId(),
			})
			if err != nil {
				return err
			}
		}

		return nil
	}

	resp, err := connections.InvoiceService.InvoiceCreate(ctx, &invoice.InvoiceCreateRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     offerModel.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetTenantServices,
		Property:          offerModel.Property,
		Ivr:               offerModel.PropertyIvr,
		Listing:           offerModel.PropertyListing,
		User:              offerModel.TenantUser(),
		Services: []*extras.ServiceSnapshot{
			service,
		},
		NotifyCallback: connections.SelfRemote,
	})
	if err != nil {
		return err
	}

	if isImitationMode {
		_, err = connections.InvoiceService.GenerateInvoicePayments(ctx, &invoice.GenerateInvoicePaymentsRequest{
			InvoiceId: resp.GetInvoice().GetInvoiceId(),
		})
		if err != nil {
			return err
		}
	}

	return nil
}

// DepositSelect - create first month deposit replacement invoice, create deposit schedule
func createDepositReplacementInvoice(
	ctx context.Context,
	offerModel *model.OfferModel,
	isImitationMode bool,
) error {
	if !offerModel.DepositReplacement() || offerModel.OriginalOffer != "" {
		return nil
	}

	price := model.GetMonthlyDepositPayment(offerModel.Conditions)

	service, err := extras.GetServicePriceSnapshot(
		ctx,
		connections.ExtrasService,
		extras.ProductTenantDepositReplacement,
		price,
	)
	if err != nil {
		return err
	}

	// check invoice existence
	inv, err := connections.InvoiceService.InvoiceList(ctx, &invoice.InvoiceListRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     offerModel.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetDepositReplacement,
		Statuses: []invoice.InvoiceStatus{
			invoice.InvoiceStatus_INVOICE_OPEN,
			invoice.InvoiceStatus_INVOICE_FULFILLED,
		},
	})
	if err != nil {
		return err
	}

	if len(inv.GetInvoices()) > 0 {
		if isImitationMode {
			_, err = connections.InvoiceService.GenerateInvoicePayments(ctx, &invoice.GenerateInvoicePaymentsRequest{
				InvoiceId: inv.GetInvoices()[0].GetInvoiceId(),
			})
			if err != nil {
				return err
			}
		}

		return nil
	}

	resp, err := connections.InvoiceService.InvoiceCreate(ctx, &invoice.InvoiceCreateRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     offerModel.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetDepositReplacement,
		Property:          offerModel.Property,
		Ivr:               offerModel.PropertyIvr,
		Listing:           offerModel.PropertyListing,
		User:              offerModel.TenantUser(),
		Services: []*extras.ServiceSnapshot{
			service,
		},
		NotifyCallback: connections.SelfRemote,
	})
	if err != nil {
		return err
	}

	if isImitationMode {
		_, err = connections.InvoiceService.GenerateInvoicePayments(ctx, &invoice.GenerateInvoicePaymentsRequest{
			InvoiceId: resp.GetInvoice().GetInvoiceId(),
		})
		if err != nil {
			return err
		}
	}

	return nil
}

func createStandingOrderInvoice(ctx context.Context, m *model.OfferModel, isImitationMode bool) error {
	freq := *m.Conditions.PaymentFrequency
	offerPrice := *m.Conditions.Price
	frequency := dataPeriodicPaymentFrequency[freq]

	priceMonth := math.Round(offerPrice*100) / 100
	pricePayment := priceMonth * float64(freq)

	var lastPayMonths int32

	termMonths := *m.Conditions.TermMonths

	if termMonths%freq == 0 {
		lastPayMonths = termMonths - freq
	} else {
		lastPayMonths = termMonths - (termMonths % freq) - freq
	}

	movein := *m.Conditions.MoveinDate
	day := paymentDayClamp(int(util.UnPtrInt32(m.Conditions.RentPaymentDay)))
	startAt := timestamppb.New(movein.AddDate(0, 1, -movein.Day()+int(day)))
	endAt := timestamppb.New(movein.AddDate(0, int(lastPayMonths)+1, 0))

	service, err := extras.GetServicePriceSnapshot(
		ctx,
		connections.ExtrasService,
		extras.ProductTenantStandingOrder,
		pricePayment,
	)
	if err != nil {
		return err
	}

	// check invoice existence
	inv, err := connections.InvoiceService.InvoiceList(ctx, &invoice.InvoiceListRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     m.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetStandingOrder,
		Statuses: []invoice.InvoiceStatus{
			invoice.InvoiceStatus_INVOICE_OPEN,
			invoice.InvoiceStatus_INVOICE_FULFILLED,
		},
	})
	if err != nil {
		return err
	}

	if len(inv.GetInvoices()) > 0 {
		if isImitationMode {
			_, err = connections.InvoiceService.GenerateInvoicePayments(ctx, &invoice.GenerateInvoicePaymentsRequest{
				InvoiceId: inv.GetInvoices()[0].GetInvoiceId(),
				Periodic: &invoice.GenerateInvoicePaymentsPeriodic{
					StartAt:   startAt,
					EndAt:     endAt,
					Frequency: frequency,
				},
			})
			if err != nil {
				return err
			}
		}
	}

	resp, err := connections.InvoiceService.InvoiceCreate(ctx, &invoice.InvoiceCreateRequest{
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     m.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetStandingOrder,
		Periodic:          true,
		Property:          m.Property,
		Ivr:               m.PropertyIvr,
		Listing:           m.PropertyListing,
		User:              m.Tenants[0].User,
		Services: []*extras.ServiceSnapshot{
			service,
		},
		NotifyCallback: connections.SelfRemote,
	})

	if isImitationMode {
		_, err = connections.InvoiceService.GenerateInvoicePayments(ctx, &invoice.GenerateInvoicePaymentsRequest{
			InvoiceId: resp.GetInvoice().GetInvoiceId(),
			Periodic: &invoice.GenerateInvoicePaymentsPeriodic{
				StartAt:   startAt,
				EndAt:     endAt,
				Frequency: frequency,
			},
		})
		if err != nil {
			return err
		}
	}

	return err
}
