package status

import (
	"context"
	"errors"
	"fmt"
	"strings"
	"time"

	extras "gitlab.mashroom.com/mashroom/backend/extras/api"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"

	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/common/generic/util"
	"gitlab.mashroom.com/mashroom/backend/common/generic/wrap"
	document "gitlab.mashroom.com/mashroom/backend/document/api"
	expenses "gitlab.mashroom.com/mashroom/backend/expenses/api"
	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/connections"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/mixpanel"
	"gitlab.mashroom.com/mashroom/backend/offer/model"
	property "gitlab.mashroom.com/mashroom/backend/property/api"
	signature "gitlab.mashroom.com/mashroom/backend/signature/api"
	"google.golang.org/protobuf/types/known/timestamppb"
	"google.golang.org/protobuf/types/known/wrapperspb"
)

func trySetPropertyRentedTenancyUpdate(bgctx context.Context, old *model.OfferModel) {
	tn, err := model.UpdateTenancy(bgctx, old)
	if err != nil {
		log.WithError(err).Error("updating tenancy")
	}

	movein := util.UnPtrTime(old.Conditions.MoveinDate)

	var tenants []*property.Tenant

	for _, t := range old.Tenants {
		var risk property.TenantRisk

		switch t.Risk {
		case offer.TenantRisk_LOW_RISK:
			risk = property.TenantRisk_LOW_RISK
		case offer.TenantRisk_MEDIUM_RISK:
			risk = property.TenantRisk_MEDIUM_RISK
		case offer.TenantRisk_HIGH_RISK:
			risk = property.TenantRisk_HIGH_RISK
		}

		tenants = append(tenants, &property.Tenant{
			User:     wrapperspb.String(t.User),
			Contract: wrapperspb.String(old.Contract),
			Active:   wrapperspb.Bool(true),
			Risk:     risk,
			TenantContact: &property.TenantContact{
				FirstName:    wrap.StringValue(t.FirstName),
				LastName:     wrap.StringValue(t.LastName),
				EmailPrimary: wrap.StringValue(t.EmailPrimary),
				PhonePrimary: wrap.StringValue(t.PhonePrimary),
			},
		})
	}

	var guarantors []*property.Guarantor

	for _, g := range old.Guarantors {
		guarantors = append(guarantors, &property.Guarantor{
			FirstName:    wrap.StringValue(g.FirstName),
			LastName:     wrap.StringValue(g.LastName),
			EmailPrimary: wrap.StringValue(g.EmailPrimary),
			PhonePrimary: wrap.StringValue(g.PhonePrimary),
		})
	}

	_, err = connections.PropertyService.ServiceRented(bgctx, &property.RentedRequest{
		Id:         old.Property,
		Offer:      old.ID.Hex(),
		Landlord:   old.LandlordUser(),
		ListingId:  old.PropertyListing,
		Tenants:    &property.TenantArrayValue{Value: tenants},
		Guarantors: &property.GuarantorArrayValue{Value: guarantors},
		TermStart:  timestamppb.New(movein),
		TermEnd:    timestamppb.New(movein.AddDate(0, int(util.UnPtrInt32(old.Conditions.TermMonths)), 0)),
		TenancyId:  tn.GetId(),
	})
	if err != nil {
		log.WithError(err).Error("setting property to rented")
	}

	if err = mixpanel.TrackLandlordPropertyRented(bgctx, old); err != nil {
		log.WithError(err).Error("failed to track PropertyRented mixpanel event")
	}

	if err = mixpanel.TrackTenantFirstRentPaymentCompleted(bgctx, old); err != nil {
		log.WithError(err).Error("failed to track FirstRentpaymentCompleted mixpanel event")
	}
}

// TrySetPropertyRentedCreateInsurances for offer with essential/secured rent collection plan
func TrySetPropertyRentedCreateInsurances(ctx context.Context, old *model.OfferModel, isImitationMode bool) error {
	if old.RentInsurancesGenerated {
		return errors.New("rent insurances already generated")
	}

	startDate := util.UnPtrTime(old.Conditions.MoveinDate)

	if isImitationMode {
		nowDate := time.Now().Truncate(24 * time.Hour)
		dateIn30days := nowDate.AddDate(0, 0, 30).Truncate(24 * time.Hour)
		startDate = startDate.Truncate(24 * time.Hour)

		if startDate.Before(nowDate) || dateIn30days.Before(startDate) {
			startDate = time.Now().Truncate(24 * time.Hour)
		}
	}

	if old.RentPlan == offer.RentPlan_RENT_PLAN_MANUAL {
		_, err := connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_OFFER_RENT_COLLECTION_MANUAL,
			Action: notification.Action_ACTION_COMPLETE,
			Targets: []*notification.Target{
				{
					User: old.TenantUser(),
					Role: notification.Role_ROLE_TENANT,
				},
				{
					User: old.LandlordUser(),
					Role: notification.Role_ROLE_LANDLORD,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
					EmbedValue: old.LandlordUser(),
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: old.Property,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
					EmbedValue: old.ID.Hex(),
				},
			},
		})
		if err != nil {
			log.WithError(err).Errorf("sending failed: nanual rent for tenant")
		}
	}

	if old.RentPlan != offer.RentPlan_RENT_PLAN_ESSENTIAL && old.RentPlan != offer.RentPlan_RENT_PLAN_SECURED {
		return errors.New("available rent plan to generate insurances is not selected")
	}

	var rentplan invoice.RentPlan

	switch old.RentPlan {
	case offer.RentPlan_RENT_PLAN_LEGACY:
		rentplan = invoice.RentPlan_RENT_PLAN_LEGACY
	case offer.RentPlan_RENT_PLAN_MANUAL:
		rentplan = invoice.RentPlan_RENT_PLAN_MANUAL
	case offer.RentPlan_RENT_PLAN_ESSENTIAL:
		rentplan = invoice.RentPlan_RENT_PLAN_ESSENTIAL
	case offer.RentPlan_RENT_PLAN_SECURED:
		rentplan = invoice.RentPlan_RENT_PLAN_SECURED
	default:
		rentplan = invoice.RentPlan_RENT_PLAN_UNKNOWN
	}

	details := &invoice.ProductDetails{
		Time: &invoice.ProductDetailsTime{
			Start: timestamppb.New(startDate),
			End:   timestamppb.New(startDate.AddDate(1, 0, 0)),
		},
		Property: &invoice.ProductDetailsProperty{
			Id:      wrapperspb.String(old.Property),
			Listing: wrapperspb.String(old.PropertyListing),
		},
		Contract: &invoice.ProductDetailsContract{
			Id: wrapperspb.String(old.ID.Hex()),
			Conditions: &invoice.RentConditions{
				Rent:     wrapperspb.Double(util.UnPtrFloat64(old.Conditions.Price)),
				RentPlan: rentplan,
			},
			Tenant: wrapperspb.String(old.TenantUser()),
		},
	}

	var products []*invoice.Product

	if old.RentPlan == offer.RentPlan_RENT_PLAN_ESSENTIAL {
		lei, err := extras.GetServiceSnapshot(
			ctx,
			connections.ExtrasService,
			extras.ProductLandlordLegalExpensesInsurance,
		)
		if err != nil {
			log.WithError(err).Error("rented insurances: getting LEI service")

			return err
		}

		products = append(products, &invoice.Product{
			ServiceId: lei.GetServiceId(),
			Service: &extras.ServiceSnapshotStatus{
				Service:   lei,
				UpdatedAt: timestamppb.Now(),
				Status:    extras.ServiceStatus_ACTIVE,
			},
			Details: details,
		})
	}

	if old.RentPlan == offer.RentPlan_RENT_PLAN_SECURED {
		rgi, err := extras.GetServiceSnapshot(
			ctx,
			connections.ExtrasService,
			extras.ProductLandlordRentGuaranteeInsurance,
		)
		if err != nil {
			log.WithError(err).Error("rented insurances: getting RGI service")

			return err
		}

		hei, err := extras.GetServiceSnapshot(
			ctx,
			connections.ExtrasService,
			extras.ProductLandlordHomeEmergencyInsurance,
		)
		if err != nil {
			log.WithError(err).Error("rented insurances: getting HEI service")

			return err
		}

		lei, err := extras.GetServiceSnapshot(
			ctx,
			connections.ExtrasService,
			extras.ProductLandlordLegalExpensesInsurance,
		)
		if err != nil {
			log.WithError(err).Error("rented insurances: getting LEI service")

			return err
		}

		products = append(
			products,
			&invoice.Product{
				ServiceId: hei.GetServiceId(),
				Service: &extras.ServiceSnapshotStatus{
					Service:   hei,
					UpdatedAt: timestamppb.Now(),
					Status:    extras.ServiceStatus_ACTIVE,
				},
				Details: details,
			},
			&invoice.Product{
				ServiceId: rgi.GetServiceId(),
				Service: &extras.ServiceSnapshotStatus{
					Service:   rgi,
					UpdatedAt: timestamppb.Now(),
					Status:    extras.ServiceStatus_ACTIVE,
				},
				Details: details,
			},
			&invoice.Product{
				ServiceId: lei.GetServiceId(),
				Service: &extras.ServiceSnapshotStatus{
					Service:   lei,
					UpdatedAt: timestamppb.Now(),
					Status:    extras.ServiceStatus_ACTIVE,
				},
				Details: details,
			},
		)
	}

	order, err := connections.InvoiceService.OrderSave(ctx, &invoice.OrderSaveRequest{
		Userid:            old.LandlordUser(),
		Products:          products,
		Clientid:          "",
		OrderType:         invoice.OrderType_REGULAR_ORDER,
		Periodic:          false,
		CorrelationType:   invoice.CorrelationTypeOffer,
		CorrelationId:     old.ID.Hex(),
		CorrelationTarget: invoice.CorrelationTargetLandlordServices,
	})
	if err != nil {
		log.WithError(err).Error("rented insurances: saving order")

		return err
	}

	commit, err := connections.InvoiceService.OrderCommit(ctx, &invoice.OrderCommitRequest{
		Userid:    old.LandlordUser(),
		Id:        order.GetOrder().GetId(),
		OrderType: invoice.OrderType_REGULAR_ORDER,
	})
	if err != nil {
		log.WithError(err).Error("rented insurances: committing order")

		return err
	}

	_, err = connections.InvoiceService.ServicePaymentCreate(ctx, &invoice.ServicePaymentCreateRequest{
		InvoiceId: commit.GetInvoice().GetInvoiceId(),
		Status:    invoice.PaymentStatus_PAYMENT_SUCCESS,
	})
	if err != nil {
		log.WithError(err).Error("rented insurances: creating payment")

		return err
	}

	err = model.OfferInsurancesGenerated(ctx, old.ID)
	if err != nil {
		log.WithError(err).Error("rented insurances: offer insurances generated save")
		old.RentInsurancesGenerated = false

		return err
	}

	old.RentInsurancesGenerated = true

	return nil
}

// SetPropertyRented attempts to set offer in rented status, if all setup payments are completed
func SetPropertyRented(ctx context.Context, old *model.OfferModel) error {
	if old.State == offer.OfferState_RENTPAID {
		return nil
	}

	if (old.RentPlan > offer.RentPlan_RENT_PLAN_MANUAL || old.RentPlan == offer.RentPlan_RENT_PLAN_LEGACY) &&
		(!old.PaidStatuses.TenantServices || !old.PaidStatuses.StandingOrder) {
		return nil
	}

	err := SetOfferState(ctx, old, offer.OfferState_RENTPAID)
	if err != nil {
		return err
	}

	if old.RentPlan == offer.RentPlan_RENT_PLAN_MANUAL {
		err = createHoldingDepositPayoutInvoice(ctx, old)
		if err != nil {
			return err
		}
	}

	bgctx := grpcctx.WithServiceUser(grpcctx.ValuesBackgroundContext(ctx))

	go trySetPropertyRentedTenancyUpdate(bgctx, old)

	err = GenerateSchedule(ctx, old)
	if err != nil {
		return err
	}

	offerRentPaymentEmbeds := []*notification.EventEmbed{
		{
			EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
			EmbedValue: old.Tenants[0].User,
		},
		{
			EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
			EmbedValue: old.Property,
		},
	}

	// TODO GetPDF 2 time (for tenant and for landlord) + retry while error
	resp, err := connections.SignatureService.SigningSignedDocument(ctx, &signature.SigningSignedDocumentRequest{
		SignatureRequestId: old.SignatureRequestID,
	})
	if err != nil {
		log.Println("Error getPDF: ", err)
	} else {
		offerRentPaymentEmbeds = append(offerRentPaymentEmbeds, &notification.EventEmbed{
			EmbedType:  messaging.EmbedType_EMBED_TYPE_FILE,
			EmbedValue: resp.GetSignedDocumentId(),
		})
	}

	_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_RENT_PAYMENT,
		Action: notification.Action_ACTION_COMPLETE,
		Targets: []*notification.Target{
			{
				User: old.LandlordUser(),
				Role: notification.Role_ROLE_LANDLORD,
			},
		},
		Embeds: offerRentPaymentEmbeds,
	})
	if err != nil {
		log.WithError(err).Errorf("sending failed: offer rent paid for landlord")
	}

	fileType := model.OfferDocsTypeSomf
	fileModel := &model.OfferFileModel{
		FileType:    fileType,
		Version:     fileType.GetVersion(),
		Mime:        fileType.GetMime(),
		Description: fileType.GetDescription(),
		Name:        fileType.GetName(),
	}

	err = fileModel.RenderSOMF(ctx, old, old.TenantUser())
	if err != nil {
		log.Println("Error render")
	}

	_, err = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_OFFER_RENT_PAYMENT,
		Action: notification.Action_ACTION_COMPLETE,
		Targets: []*notification.Target{
			{
				User: old.TenantUser(),
				Role: notification.Role_ROLE_TENANT,
			},
		},
		Embeds: offerRentPaymentEmbeds,
	})
	if err != nil {
		log.WithError(err).Errorf("sending failed: offer rent paid for tenant")
	}

	_, err = connections.NotificationService.SendNotification(
		ctx,
		&notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_OFFER,
			Action: notification.Action_ACTION_COMPLETE,
			Targets: []*notification.Target{
				{
					User: old.TenantUser(),
					Role: notification.Role_ROLE_TENANT,
				},
				{
					User: old.LandlordUser(),
					Role: notification.Role_ROLE_LANDLORD,
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: old.Property,
				},
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_OFFER,
					EmbedValue: old.ID.Hex(),
				},
			},
		},
	)
	if err != nil {
		log.WithError(err).Errorf("sending failed: %s", notification.EventType_EVENT_PAYMENT_READY_MOVEIN_TENANT)
	}

	if old.Conditions.DepositType == offer.DepositType_DEPOSIT_REPLACEMENT {
		_, _ = RenderCSV(ctx, old, true)
	}

	return nil
}

// GenerateSchedule attempts to generate a payment schedule for signed and set up offer
func GenerateSchedule(ctx context.Context, old *model.OfferModel) error {
	if old.PaidStatuses.StandingOrder != old.PaidStatuses.StandingOrderSetup ||
		old.PaidStatuses.DepositReplacement != old.PaidStatuses.DepositReplacementSetup ||
		old.ScheduleGenerated ||
		(old.State != offer.OfferState_RENTPAID && old.State != offer.OfferState_AWAITING_TENANCY) {
		return nil
	}

	_, err := connections.ExpensesService.GenerateRentRecords(ctx, &expenses.GenerateRentRecordsRequest{
		OfferId:                  old.ID.Hex(),
		Property:                 old.Property,
		Ivr:                      old.PropertyIvr,
		ListingId:                old.PropertyListing,
		Landlord:                 old.LandlordUser(),
		Tenant:                   old.TenantUser(),
		Manager:                  old.Manager,
		Viewing:                  old.Viewing,
		Rent:                     util.UnPtrFloat64(old.Conditions.Price),
		BreakClause:              util.UnPtrInt32(old.Conditions.BreakClause),
		RentPaymentDay:           util.UnPtrInt32(old.Conditions.RentPaymentDay),
		TermMonths:               util.UnPtrInt32(old.Conditions.TermMonths),
		MoveinDate:               timestamppb.New(util.UnPtrTime(old.Conditions.MoveinDate)),
		FirstRentPaymentDate:     timestamppb.New(util.UnPtrTime(old.Conditions.FirstRentPaymentDate)),
		RentCollectionFeePercent: old.RentPlan.Percent(),
		Legacy:                   old.RentPlan == offer.RentPlan_RENT_PLAN_LEGACY,
	})
	if err != nil {
		return err
	}

	old.ScheduleGenerated = true

	err = model.UpdateOffer(ctx, old)

	return err
}

// RenderCSV renders offer details as CSV file
func RenderCSV(ctx context.Context, old *model.OfferModel, notify bool) (fileID string, err error) {
	prop, err := connections.PropertyService.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
		Id: old.Property,
	})
	if err != nil {
		return
	}

	header := `"Agent name", "Landlord name", "Risk Address", "Duration of tenancy", "Start date of tenancy",` +
		` "End date of tenancy", "LPG Premium", "Annual Rent", "Tenant Name", "Policy Type"`
	agentName := fmt.Sprintf("%s", "Mashroom Ltd")
	landlordName := fmt.Sprintf(
		"%s %s",
		util.UnPtrString(old.Landlord.FirstName),
		util.UnPtrString(old.Landlord.LastName),
	)
	riskAddress := fmt.Sprintf(
		"%s %s %s %s %s",
		prop.GetProperty().GetLocation().GetBuilding().GetValue(),
		prop.GetProperty().GetLocation().GetFlat().GetValue(),
		prop.GetProperty().GetLocation().GetStreet().GetValue(),
		prop.GetProperty().GetLocation().GetCity().GetValue(),
		prop.GetProperty().GetLocation().GetPostcode().GetValue(),
	)
	rentTerm := util.UnPtrInt32(old.Conditions.TermMonths)
	moveInDate := old.Conditions.MoveinDate
	condPrice := util.UnPtrFloat64(old.Conditions.Price)
	moveOutDate := moveInDate.AddDate(0, int(rentTerm), 0)
	lpgPremium := (condPrice * 12.0) * 0.0019
	annualRent := condPrice * 12.0
	tenantName := fmt.Sprintf(
		"%s %s",
		util.UnPtrString(old.Tenants[0].FirstName),
		util.UnPtrString(old.Tenants[0].LastName),
	)
	body := fmt.Sprintf(
		"%s, %s, %s, %d, %s, %s, %.2f, %.2f, %s, %s",
		agentName,
		landlordName,
		strings.ReplaceAll(riskAddress, "  ", " "),
		rentTerm,
		moveInDate.Format("2006-01-02"),
		moveOutDate.Format("2006-01-02"),
		lpgPremium,
		annualRent,
		tenantName,
		"New Business",
	)
	text := header + "\n" + body

	now := time.Now()
	name := fmt.Sprintf(
		"New-Deposit-Replacement-Product-order-%04d-%02d-%02d.%s",
		now.Year(),
		now.Month(),
		now.Day(),
		"csv",
	)

	resp, err := document.UploadFile(
		ctx,
		connections.DocumentService,
		name,
		"text/csv",
		&document.FileAttribute{
			Data: map[string]string{
				document.FileAttributeGenerated.String(): "true",
			},
		},
		strings.NewReader(text),
	)
	if err != nil {
		return
	}

	if notify {
		_, _ = connections.NotificationService.SendNotification(ctx, &notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_OFFER_DEPOSIT_REPLACEMENT,
			Action: notification.Action_ACTION_CREATE,
			Targets: []*notification.Target{
				{
					Role: notification.Role_ROLE_ADMIN,
					Contacts: &notification.Contacts{
						Email: "orders@mashroom.com",
					},
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_FILE,
					EmbedValue: resp.GetId(),
				},
			},
		})
	}

	return resp.GetId(), nil
}
