package model

import (
	"context"
	"errors"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"

	document "gitlab.mashroom.com/mashroom/backend/document/api"

	tenancy "gitlab.mashroom.com/mashroom/backend/tenancy/api"

	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/common/generic/util"
	expenses "gitlab.mashroom.com/mashroom/backend/expenses/api"
	extras "gitlab.mashroom.com/mashroom/backend/extras/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/connections"
	property "gitlab.mashroom.com/mashroom/backend/property/api"
	schedule "gitlab.mashroom.com/mashroom/backend/schedule/api"
	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"
	"google.golang.org/protobuf/types/known/timestamppb"
)

func modelMigrationV1(ctx context.Context, db *mongo.Database, migrationID string) (int, error) {
	offers := db.Collection(MongoOffersCollection)

	cur, err := offers.Find(ctx, bson.M{
		"state": bson.M{
			"$in": bson.A{
				offer.OfferState_SIGNED,
				offer.OfferState_AWAITING_CONFIRM_RENT,
				offer.OfferState_RENTPAID,
			},
		},
		"moveout_scheduled": bson.M{
			"$ne": true,
		},
	}, &options.FindOptions{
		Sort: bson.D{{Key: "created_at", Value: -1}},
	})
	if err != nil {
		return -1, err
	}

	for cur.Next(ctx) {
		var m OfferModel

		err = cur.Decode(&m)
		if err != nil {
			log.WithError(err).Error("decoding offer")
			return -1, err
		}

		tenancyResp, err := connections.TenancyService.TenancyGet(ctx, &tenancy.TenancyGetRequest{
			Spec: &tenancy.TenancySpec{
				UserId: "",
				Details: &tenancy.TenancySpec_Select{
					Select: &tenancy.TenancySpecSelect{
						States: nil,
						By: &tenancy.TenancySpecSelect_OfferId{
							OfferId: m.ID.Hex(),
						},
					},
				},
			},
		})
		if err != nil {
			log.WithError(err).Errorf("getting tenancy for %v", m.ID.Hex())
			continue
		}

		t := tenancyResp.GetTenancy()

		if m.Conditions.MoveinDate == nil || m.Conditions.TermMonths == nil {
			continue
		}

		mout := m.Conditions.MoveinDate.AddDate(0, int(*m.Conditions.TermMonths), 0)

		moveout := timestamppb.New(mout)

		_, err = connections.ScheduleService.InvocationSchedule(ctx, &schedule.InvocationScheduleRequest{
			Handler: "tenancy_termination",
			Instant: moveout,
			Data: map[string]string{
				"tenancy_id": t.GetId(),
			},
			Retry: 1,
		})
		if err != nil {
			log.WithError(err).Errorf("scheduling for %v", m.ID.Hex())
			return -1, err
		}

		m.MoveoutScheduled = true

		_, err = offers.ReplaceOne(ctx, bson.M{"_id": m.ID}, m)
		if err != nil {
			log.WithError(err).Errorf("updating %v", m.ID.Hex())
			return -1, err
		}
	}

	return -1, nil
}

func modelMigrationV2(ctx context.Context, db *mongo.Database, migrationID string) (int, error) {
	offers := db.Collection(MongoOffersCollection)

	_, err := offers.UpdateMany(ctx, bson.M{}, bson.M{
		"$set": bson.M{
			"schedule_generated": false,
		},
	})

	if errors.Is(err, mongo.ErrNoDocuments) {
		err = nil
	}

	if err != nil {
		return -1, err
	}

	schedules := db.Collection(MongoScheduledPaymentsCollection)

	_, err = schedules.DeleteMany(ctx, bson.M{})

	if errors.Is(err, mongo.ErrNoDocuments) {
		err = nil
	}

	return -1, err
}

func modelMigrationV3(ctx context.Context, db *mongo.Database, migrationID string) (int, error) {
	offers := db.Collection(MongoOffersCollection)

	cur, err := offers.Find(ctx, bson.M{})
	if err != nil {
		return -1, err
	}

	defer func() {
		_ = cur.Close(ctx)
	}()

	for cur.Next(ctx) {
		var off OfferModel

		err = cur.Decode(&off)
		if err != nil {
			return -1, err
		}

		if off.PropertyListing == "" {
			var prop *property.GetResponse

			prop, err = connections.PropertyService.Get(ctx, &property.GetRequest{
				Id: off.Property,
			})
			if err != nil {
				log.WithError(err).Errorf("fetching property %s for offer %s", off.Property, off.ID.Hex())

				continue
			}

			listing := prop.GetProperty().GetListing().GetId().GetValue()

			_, err = offers.UpdateOne(ctx, bson.M{"_id": off.ID}, bson.M{"$set": bson.M{"property_listing": listing}})
			if err != nil {
				log.WithError(err).Errorf("updating offer %s", off.ID.Hex())
			}
		}
	}

	return -1, nil
}

func modelMigrationV4(ctx context.Context, db *mongo.Database, migrationID string) (int, error) {
	schedules := db.Collection(MongoScheduledPaymentsCollection)
	offers := db.Collection(MongoOffersCollection)

	offerids, err := schedules.Distinct(ctx, "offer", bson.M{})
	if err != nil {
		if errors.Is(err, mongo.ErrNoDocuments) {
			err = nil
		}

		return -1, err
	}

	for _, o := range offerids {
		offid, ok := o.(string)
		if !ok {
			continue
		}

		var id primitive.ObjectID

		id, err = primitive.ObjectIDFromHex(offid)
		if err != nil {
			log.WithError(err).Errorf("parsing offer id %s", offid)

			continue
		}

		var off OfferModel

		err = offers.FindOne(ctx, bson.M{"_id": id}).Decode(&off)
		if err != nil {
			log.WithError(err).Errorf("fetching offer by id %s", offid)

			continue
		}

		if off.Conditions == nil || off.Conditions.MoveinDate == nil || off.Conditions.TermMonths == nil {
			continue
		}

		movein := *off.Conditions.MoveinDate
		end := movein.AddDate(0, int(*off.Conditions.TermMonths), 0)

		_, err = schedules.UpdateMany(
			ctx,
			bson.M{
				"offer": offid,
				"id": bson.M{
					"$nin": bson.A{
						movein.Format("200601"),
						movein.AddDate(0, 1, 0).Format("200601"),
						end.AddDate(0, -1, 0).Format("200601"),
						end.Format("200601"),
					},
				},
			},
			bson.M{
				"$set": bson.M{
					"advance.eligible": true,
				},
			},
		)
		if err != nil {
			log.WithError(err).Errorf("updating advance eligiblity for offer by id %s", offid)
		}

		var count int64

		count, err = schedules.CountDocuments(ctx, bson.M{
			"offer": offid,
			"parts": bson.M{
				"$elemMatch": bson.M{
					"status":        offer.PaymentStatus_PAYMENT_SUCCESS,
					"services.type": extras.ServiceType_RENT,
				},
			},
			"id": bson.M{
				"$in": bson.A{
					movein.Format("200601"),
					movein.AddDate(0, 1, 0).Format("200601"),
				},
			},
		})
		if err != nil {
			log.WithError(err).Errorf("checking paid statuses for %s", offid)

			continue
		}

		if count != 2 {
			continue
		}

		_, err = schedules.UpdateMany(
			ctx,
			bson.M{"offer": offid},
			bson.M{
				"$set": bson.M{
					"advance.initialized": true,
				},
			},
		)

		if err != nil {
			log.WithError(err).Errorf("updating offer %s", offid)
		}
	}

	return -1, nil
}

func modelMigrationV5(ctx context.Context, db *mongo.Database, migrationID string) (int, error) {
	schedules := db.Collection(MongoScheduledPaymentsCollection)
	offers := db.Collection(MongoOffersCollection)

	offerids, err := schedules.Distinct(ctx, "offer", bson.M{})
	if err != nil {
		if errors.Is(err, mongo.ErrNoDocuments) {
			err = nil
		}

		return -1, err
	}

	for _, o := range offerids {
		offid, ok := o.(string)
		if !ok {
			continue
		}

		var id primitive.ObjectID

		id, err = primitive.ObjectIDFromHex(offid)
		if err != nil {
			log.WithError(err).Errorf("parsing offer id %s", offid)

			continue
		}

		var off OfferModel

		err = offers.FindOne(ctx, bson.M{"_id": id}).Decode(&off)
		if err != nil {
			log.WithError(err).Errorf("fetching offer by id %s", offid)

			continue
		}

		if off.Conditions == nil || off.Conditions.MoveinDate == nil || off.Conditions.TermMonths == nil {
			continue
		}

		movein := *off.Conditions.MoveinDate
		term := *off.Conditions.TermMonths
		brclause := int(util.UnPtrInt32(off.Conditions.BreakClause))

		var cur *mongo.Cursor

		cur, err = schedules.Find(ctx, bson.M{
			"offer": offid,
		})
		if err != nil {
			log.WithError(err).Errorf("getting schedule list for %s", offid)

			continue
		}

		for cur.Next(ctx) {
			var sched ScheduledPaymentModel

			err = cur.Decode(&sched)
			if err != nil {
				log.WithError(err).Errorf("getting schedule for %s", offid)

				continue
			}

			sched.Payout.DueDay = int(util.UnPtrInt32(off.Conditions.RentPaymentDay))
			sched.Payout.Amount = util.UnPtrFloat64(off.Conditions.Price)
			sched.Advance.Eligible = RenderScheduleAvailable(sched.Date, movein, int(term), brclause)

			_, err = schedules.ReplaceOne(ctx, bson.M{"id": sched.ID, "offer": sched.Offer}, sched)
			if err != nil {
				log.WithError(err).Errorf("updating schedule for %s - %s", offid, sched.ID)

				continue
			}
		}
	}

	return -1, nil
}

func modelMigrationV6(ctx context.Context, db *mongo.Database, migrationID string) (int, error) {
	offers := db.Collection(MongoOffersCollection)

	_, err := offers.UpdateMany(ctx, bson.M{}, bson.M{"$set": bson.M{"ivr_migrate": true}})

	return -1, err
}

func migratePropertyIVR(ctx context.Context, db *mongo.Database) error {
	offers := db.Collection(MongoOffersCollection)

	propertyids, err := offers.Distinct(ctx, "property", bson.M{"ivr_migrate": true})
	if err != nil {
		if errors.Is(err, mongo.ErrNoDocuments) {
			err = nil
		}

		return err
	}

	for _, p := range propertyids {
		propid, ok := p.(string)
		if !ok {
			log.Error("invalid property id")

			continue
		}

		var resp *property.ServiceIVRListByIDResponse

		resp, err = connections.PropertyService.ServiceIVRListByID(ctx, &property.ServiceIVRListByIDRequest{
			Properties: []string{propid},
		})
		if err != nil {
			log.WithError(err).Errorf("fetching ivr for property %s", propid)

			continue
		}

		if resp.GetIvrs() == nil {
			log.WithError(err).Errorf("no ivr for property %s", propid)

			continue
		}

		ivr := resp.GetIvrs()[propid]
		if len(ivr) == 0 {
			log.WithError(err).Errorf("no ivr for property %s", propid)

			continue
		}

		_, err = offers.UpdateMany(
			ctx,
			bson.M{"property": propid},
			bson.M{"$set": bson.M{"property_ivr": ivr, "ivr_migrate": false}},
		)
		if err != nil {
			log.WithError(err).Errorf("updating ivr for property %s", propid)
		}
	}

	return nil
}

func modelMigrationV7(ctx context.Context, db *mongo.Database, migrationID string) (int, error) {
	col := db.Collection(MongoOffersCollection)

	_, err := col.UpdateMany(
		ctx,
		bson.M{},
		bson.A{
			bson.M{
				"$set": bson.M{
					"guarantors": bson.M{
						"$filter": bson.M{
							"input": bson.M{
								"$map": bson.M{
									"input": "$tenants",
									"as":    "t",
									"in":    "$$t.guarantor",
								},
							},
							"as":   "t",
							"cond": bson.M{"$ne": bson.A{bson.M{"$ifNull": bson.A{"$$t.email_primary", nil}}, nil}},
						},
					},
				},
			},
		},
	)
	if err != nil {
		return -1, err
	}

	cur, err := col.Find(ctx, bson.M{"guarantors": bson.M{"$ne": bson.A{}}})
	if err != nil {
		return -1, err
	}

	defer func() {
		_ = cur.Close(ctx)
	}()

	for cur.Next(ctx) {
		var off OfferModel

		err = cur.Decode(&off)
		if err != nil {
			return -1, err
		}

		m := make(map[string]struct{})

		var tgt []*GuarantorOfferModel

		for _, g := range off.Guarantors {
			_, ok := m[util.UnPtrString(g.EmailPrimary)]
			if !ok {
				tgt = append(tgt, g)
				m[util.UnPtrString(g.EmailPrimary)] = struct{}{}
			}
		}

		_, err = col.UpdateOne(ctx, bson.M{"_id": off.ID}, bson.M{"$set": bson.M{"guarantors": tgt}})
		if err != nil {
			return -1, err
		}
	}

	return -1, err
}

func modelMigrationV8(ctx context.Context, db *mongo.Database, migrationID string) (int, error) {
	col := db.Collection(MongoOffersCollection)

	_, err := col.UpdateMany(
		ctx,
		bson.M{"tenants.guarantor_needed": true},
		bson.M{"$set": bson.M{"guarantor_required": true}},
	)

	return -1, err
}

func modelMigrationV9(ctx context.Context, db *mongo.Database, migrationID string) (int, error) {
	col := db.Collection(MongoOffersCollection)

	cur, err := col.Find(ctx, bson.M{"schedule_notifications": bson.M{"$ne": nil}})
	if err != nil {
		return -1, err
	}

	defer func() {
		_ = cur.Close(ctx)
	}()

	for cur.Next(ctx) {
		var off OfferModel

		err = cur.Decode(&off)
		if err != nil {
			return -1, err
		}

		notifs := make(map[ScheduleNotificationType]*ScheduleNotification)

		for k, n := range off.ScheduleNotifications {
			switch int(k) {
			case 186:
				// notification.EventType_EVENT_OFFER_RECEIVED_NOT_RESPONDED_REMIND_LANDLORD_3HRS
				if notifs[ScheduleNotificationNotResponded3hrs] == nil {
					notifs[ScheduleNotificationNotResponded3hrs] = &ScheduleNotification{
						StartedAt:  n.StartedAt,
						IsNotified: n.IsNotified,
					}
				}
			case 187, 189:
				// notification.EventType_EVENT_OFFER_RECEIVED_NOT_RESPONDED_REMIND_LANDLORD_1WEEK_LEFT
				// notification.EventType_EVENT_OFFER_RECEIVED_NOT_RESPONDED_REMIND_TENANT_1WEEK_LEFT
				if notifs[ScheduleNotificationWeekLeftToMovein] == nil {
					notifs[ScheduleNotificationWeekLeftToMovein] = &ScheduleNotification{
						StartedAt: n.StartedAt,
					}
				}

				next := notifs[ScheduleNotificationWeekLeftToMovein]

				next.IsNotified = next.IsNotified || n.IsNotified
			}
		}

		_, err = col.UpdateOne(ctx, bson.M{"_id": off.ID}, bson.M{"$set": bson.M{"schedule_notifications": notifs}})
		if err != nil {
			return -1, err
		}
	}

	return -1, err
}

func modelMigrationV10(ctx context.Context, db *mongo.Database, migrationID string) (int, error) {
	offers := db.Collection(MongoOffersCollection)

	_, err := offers.UpdateMany(ctx, bson.M{
		"state": bson.M{
			"$in": bson.A{
				offer.OfferState_SIGNING,
				offer.OfferState_SIGNED,
				offer.OfferState_RENTPAID,
				offer.OfferState_TERMINATED,
				offer.OfferState_FINISHED,
				offer.OfferState_EXPIRED,
			},
		},
		"contract": bson.M{
			"$exists": true,
			"$ne":     "",
		},
	}, bson.M{"$set": bson.M{"contract_migrate": true}})

	return -1, err
}

func modelMigrationV11(ctx context.Context, db *mongo.Database, migrationID string) (int, error) {
	offers := db.Collection(MongoOffersCollection)

	_, err := offers.UpdateMany(ctx, bson.M{
		"state": bson.M{
			"$in": bson.A{
				offer.OfferState_SIGNED,
				offer.OfferState_RENTPAID,
				offer.OfferState_TERMINATED,
				offer.OfferState_FINISHED,
				offer.OfferState_EXPIRED,
				offer.OfferState_DENIED,
				offer.OfferState_CANCELED,
				offer.OfferState_ERROR,
				offer.OfferState_EARLY_TERMINATION,
			},
		},
	}, bson.M{"$set": bson.M{"rent_plan": offer.RentPlan_RENT_PLAN_LEGACY}})

	return -1, err
}

func modelMigrationV12(ctx context.Context, db *mongo.Database, migrationID string) (int, error) {
	offersHistory := db.Collection(MongoOfferHistoryCollection)
	offerID := primitive.NewObjectID().Hex()
	_, err := offersHistory.InsertOne(ctx, &OfferHistoryModel{
		CreatedAt:  time.Now(),
		UpdatedAt:  time.Now(),
		Action:     "update",
		Conditions: &OfferConditionsModel{},
		OfferID:    offerID,
		Deleted:    false,
		IsLocked:   false,
		Sync:       OfferSync{},
	},
	)

	if err != nil {
		return -1, err
	}

	return -1, nil
}

func migrateContractStates(ctx context.Context, db *mongo.Database) error {
	offers := db.Collection(MongoOffersCollection)

	cur, err := offers.Find(ctx, bson.M{"contract_migrate": true})
	if err != nil {
		return err
	}

	defer func() {
		_ = cur.Close(ctx)
	}()

	var m OfferModel

	for cur.Next(ctx) {
		err = cur.Decode(&m)
		if err != nil {
			continue
		}

		if m.Conditions == nil {
			continue
		}

		months := util.UnPtrInt32(m.Conditions.TermMonths)
		expiration := util.UnPtrTime(m.Conditions.MoveinDate).AddDate(0, int(months), 0)

		var state document.DocumentState

		if m.State == offer.OfferState_SIGNING {
			state = document.DocumentState_UNKNOWN_DOCUMENT_STATE
		} else {
			state = document.DocumentState_ACTIVE
		}

		if expiration.Before(time.Now()) {
			state = document.DocumentState_EXPIRED
		}

		_, err = connections.DocumentService.ServiceUpdateStatusExpiration(
			ctx,
			&document.ServiceUpdateStatusExpirationRequest{
				FileId:     m.Contract,
				State:      state,
				Expiration: timestamppb.New(expiration),
			},
		)
		if err != nil {
			log.WithError(err).WithField("contract", m.Contract).Error("migrating expiration date and state")

			continue
		}

		_, _ = offers.UpdateOne(
			ctx,
			bson.M{"_id": m.ID},
			bson.M{"$set": bson.M{"contract_migrate": false}},
		)
	}

	return nil
}

func migrateOfferManagers(ctx context.Context, db *mongo.Database) (count int64, err error) {
	offers := db.Collection(MongoOffersCollection)

	cur, err := offers.Find(ctx, bson.M{"$or": bson.A{
		bson.M{"manager": ""},
		bson.M{"manager": nil},
		bson.M{"manager": bson.M{"$exists": false}},
	}})
	if err != nil {
		return -1, err
	}

	defer func() {
		_ = cur.Close(ctx)
	}()

	var m OfferModel

	for cur.Next(ctx) {
		count++

		err = cur.Decode(&m)
		if err != nil {
			continue
		}

		var ver *listing.VersionGetResponse

		ver, err = connections.ListingService.VersionGet(grpcctx.WithServiceUser(ctx), &listing.VersionGetRequest{
			Spec: listing.VersionSpecListingID("", m.PropertyListing),
		})
		if err != nil {
			ver, err = connections.ListingService.VersionGet(grpcctx.WithServiceUser(ctx), &listing.VersionGetRequest{
				Spec: listing.VersionSpecPropertyID("", m.Property),
			})
			if err != nil {
				continue
			}
		}

		_, _ = offers.UpdateOne(
			ctx,
			bson.M{"_id": m.ID},
			bson.M{"$set": bson.M{"manager": ver.GetVersion().GetPeople().GetManager().GetValue()}},
		)
	}

	return
}

func migrateSchedulePayments(ctx context.Context, db *mongo.Database) (count int64, err error) {
	schedules := db.Collection(MongoScheduledPaymentsCollection)
	offers := db.Collection(MongoOffersCollection)

	cur, err := schedules.Find(ctx, bson.M{"$or": bson.A{
		bson.M{"migrated": false},
		bson.M{"migrated": nil},
		bson.M{"migrated": bson.M{"$exists": false}},
	}})
	if err != nil {
		return -1, err
	}

	defer func() {
		_ = cur.Close(ctx)
	}()

	var m ScheduledPaymentModel
	for cur.Next(ctx) {
		count++

		err = cur.Decode(&m)
		if err != nil {
			continue
		}

		id, err := primitive.ObjectIDFromHex(m.Offer)
		if err != nil {
			continue
		}

		var off OfferModel

		err = offers.FindOne(ctx, bson.M{"_id": id}).Decode(&off)
		if err != nil {
			continue
		}

		err = migrateExpenses(ctx, m, off.Property, off.PropertyListing)
		if err != nil {
			log.WithError(err).Errorf("migrateExpenses error %s - %s ", m.ID, m.Offer)
			continue
		}

		_, _ = schedules.UpdateOne(
			ctx,
			bson.M{"$and": bson.A{
				bson.M{"id": m.ID},
				bson.M{"offer": m.Offer},
			}},
			bson.M{"$set": bson.M{"migrated": true}},
		)
	}

	return
}

func migrateLandlordRentIncome(
	ctx context.Context,
	sch ScheduledPaymentModel,
	part *ScheduledPaymentPartsModel,
	propertyID, listingID string,
) error {
	rentDue, err := time.Parse("200601", sch.ID)
	if err != nil {
		return err
	}

	var paidAt time.Time

	if part.Status == offer.PaymentStatus_PAYMENT_SUCCESS {
		paidAt = part.Due
	}

	_, err = connections.ExpensesService.RecordCreate(grpcctx.WithServiceUser(ctx), &expenses.RecordCreateRequest{
		UserId:     sch.Landlord,
		PropertyId: propertyID,
		ListingId:  listingID,
		Type: &expenses.RecordCreateRequest_Income{
			Income: &expenses.Income{
				Income: &expenses.Income_Rent{
					Rent: &expenses.IncomeRent{
						AdvanceAvailable: sch.Advance.FeePercentage,
						Payouts: map[string]*expenses.IncomePayout{
							"": {
								InvoiceId:    sch.Payout.InvoiceID,
								PayoutStatus: LegacyTransferExpensesPayoutState(sch.Payout.State),
							},
						},
						Serial:                part.Serial,
						PaymentStatus:         LegacyTransferOfferPaymentStatusToExpenses(part.Status),
						PaymentPaidAt:         timestamppb.New(paidAt),
						FeePercentage:         0,
						FeeValue:              0,
						FeePercentageRentPlan: 0,
						FeeValueRentPlan:      0,
					},
				},
			},
		},
		Config: &expenses.Config{
			UserId: sch.Landlord,
			Type: &expenses.Config_Offer{
				Offer: &expenses.ConfigOffer{
					OfferId: sch.Offer,
					Role:    expenses.ConfigOfferRole_CONFIG_OFFER_ROLE_LANDLORD,
				},
			},
		},
		Description: "Property rental income",
		Amount:      sch.Payout.Amount,
		Time: timestamppb.New(time.Date(
			rentDue.Year(),
			rentDue.Month(),
			sch.Payout.DueDay,
			0,
			0,
			0,
			0,
			time.UTC,
		)),
	})
	if err != nil {
		return err
	}

	return nil
}

func migrateExpenses(
	ctx context.Context,
	sch ScheduledPaymentModel,
	propertyID, listingID string,
) error {
	for _, p := range sch.Parts {
		expenseType := expenses.ExpenseType_EXPENSE_TYPE_DRP

		var amount float64

		var paidAt time.Time

		description := "Property deposit replacement expense"

		for _, s := range p.Services {
			amount += s.Price

			if s.Type == extras.ServiceType_RENT {
				expenseType = expenses.ExpenseType_EXPENSE_TYPE_RENT
				description = "Property rent expense"
			}
		}

		if p.Status == offer.PaymentStatus_PAYMENT_SUCCESS {
			paidAt = p.Due
		}

		r := &expenses.RecordCreateRequest{
			UserId:     sch.Tenants[0],
			PropertyId: propertyID,
			ListingId:  listingID,
			Config: &expenses.Config{
				UserId: sch.Tenants[0],
				Type: &expenses.Config_Offer{
					Offer: &expenses.ConfigOffer{
						OfferId: sch.Offer,
						Role:    expenses.ConfigOfferRole_CONFIG_OFFER_ROLE_TENANT,
					},
				},
			},
			Description: description,
			Amount:      amount,
			Time:        timestamppb.New(p.Due),
		}

		switch expenseType {
		case expenses.ExpenseType_EXPENSE_TYPE_DRP:
			r.Type = &expenses.RecordCreateRequest_Expense{
				Expense: &expenses.Expense{
					Expense: &expenses.Expense_Drp{
						Drp: &expenses.ExpenseDrp{
							InvoiceId: p.InvoiceID,
							PaymentId: p.PaymentID,
							Serial:    p.Serial,
							Status:    LegacyTransferOfferPaymentStatusToExpenses(p.Status),
						},
					},
				},
			}
		case expenses.ExpenseType_EXPENSE_TYPE_RENT:
			r.Type = &expenses.RecordCreateRequest_Expense{
				Expense: &expenses.Expense{
					Expense: &expenses.Expense_Rent{
						Rent: &expenses.ExpenseRent{
							InvoiceId:  p.InvoiceID,
							PaymentId:  p.PaymentID,
							Serial:     p.Serial,
							Status:     LegacyTransferOfferPaymentStatusToExpenses(p.Status),
							LandlordId: sch.Landlord,
							Payouts: map[string]*expenses.ExpensePayout{
								"": {
									Status: LegacyTransferExpensesPayoutState(sch.Payout.State),
								},
							},
							PaymentPaidAt: timestamppb.New(paidAt),
						},
					},
				},
			}
		}

		_, err := connections.ExpensesService.RecordCreate(ctx, r)
		if err != nil {
			return err
		}

		if expenseType == expenses.ExpenseType_EXPENSE_TYPE_RENT {
			err = migrateLandlordRentIncome(ctx, sch, p, propertyID, listingID)
			if err != nil {
				return err
			}
		}
	}

	return nil
}
