package model

import (
	"context"
	"testing"

	"github.com/stretchr/testify/require"
	"google.golang.org/protobuf/types/known/wrapperspb"

	"gitlab.mashroom.com/mashroom/backend/common/generic/util"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
)

func TestOfferModel_UpdateOfferLandlords(t *testing.T) {
	type (
		fields struct {
			Landlord  *LandlordModel
			Landlords []*LandlordModel
		}

		args struct {
			ctx       context.Context
			landlords []*offer.Landlord
		}
	)

	tests := []struct {
		want    *OfferModel
		name    string
		args    args
		fields  fields
		wantErr bool
	}{
		{
			name:   "ok: empty landlord",
			fields: fields{},
			args: args{
				ctx: context.TODO(),
				landlords: []*offer.Landlord{
					{
						User:           "11",
						FirstName:      wrapperspb.String("first1"),
						LastName:       wrapperspb.String("last1"),
						EmailPrimary:   wrapperspb.String("email1"),
						PhonePrimary:   wrapperspb.String("+111111221"),
						BillingAddress: wrapperspb.String("address1"),
					},
				},
			},
			want:    &OfferModel{},
			wantErr: false,
		},
		{
			name: "ok: empty incoming landlords - nothing to do",
			fields: fields{
				Landlord: &LandlordModel{
					User:           "10",
					FirstName:      util.PtrString("1first"),
					LastName:       util.PtrString("1last"),
					EmailPrimary:   util.PtrString("1email"),
					PhonePrimary:   util.PtrString("+22222211111122"),
					BillingAddress: util.PtrString("1address"),
				},
				Landlords: []*LandlordModel{
					{
						User:           "11",
						FirstName:      util.PtrString("first"),
						LastName:       util.PtrString("last"),
						EmailPrimary:   util.PtrString("email"),
						PhonePrimary:   util.PtrString("+11111122"),
						BillingAddress: util.PtrString("address"),
					},
				},
			},
			args: args{
				ctx:       context.TODO(),
				landlords: nil,
			},
			want: &OfferModel{
				Landlord: &LandlordModel{
					User:           "10",
					FirstName:      util.PtrString("1first"),
					LastName:       util.PtrString("1last"),
					EmailPrimary:   util.PtrString("1email"),
					PhonePrimary:   util.PtrString("+22222211111122"),
					BillingAddress: util.PtrString("1address"),
				},
				Landlords: []*LandlordModel{
					{
						User:           "11",
						FirstName:      util.PtrString("first"),
						LastName:       util.PtrString("last"),
						EmailPrimary:   util.PtrString("email"),
						PhonePrimary:   util.PtrString("+11111122"),
						BillingAddress: util.PtrString("address"),
					},
				},
			},
			wantErr: false,
		},
		{
			name: "ok, update landlord",
			fields: fields{
				Landlord: &LandlordModel{
					User:           "10",
					FirstName:      util.PtrString("1first"),
					LastName:       util.PtrString("1last"),
					EmailPrimary:   util.PtrString("1email"),
					PhonePrimary:   util.PtrString("+22222211111122"),
					BillingAddress: util.PtrString("1address"),
				},
				Landlords: []*LandlordModel{
					{
						User:           "11",
						FirstName:      util.PtrString("first"),
						LastName:       util.PtrString("last"),
						EmailPrimary:   util.PtrString("email"),
						PhonePrimary:   util.PtrString("+11111122"),
						BillingAddress: util.PtrString("address"),
					},
				},
			},
			args: args{
				ctx: context.TODO(),
				landlords: []*offer.Landlord{
					{
						User:           "11",
						FirstName:      wrapperspb.String("first1"),
						LastName:       wrapperspb.String("last1"),
						EmailPrimary:   wrapperspb.String("email1"),
						PhonePrimary:   wrapperspb.String("+111111221"),
						BillingAddress: wrapperspb.String("address1"),
					},
				},
			},
			want: &OfferModel{
				Landlord: &LandlordModel{
					User:           "10",
					FirstName:      util.PtrString("1first"),
					LastName:       util.PtrString("1last"),
					EmailPrimary:   util.PtrString("1email"),
					PhonePrimary:   util.PtrString("+22222211111122"),
					BillingAddress: util.PtrString("1address"),
				},
				Landlords: []*LandlordModel{
					{
						User:           "11",
						FirstName:      util.PtrString("first1"),
						LastName:       util.PtrString("last1"),
						EmailPrimary:   util.PtrString("email1"),
						PhonePrimary:   util.PtrString("+111111221"),
						BillingAddress: util.PtrString("address1"),
					},
				},
			},
			wantErr: false,
		},
		{
			name: "ok, new second landlord",
			fields: fields{
				Landlord: &LandlordModel{
					User:           "10",
					FirstName:      util.PtrString("1first"),
					LastName:       util.PtrString("1last"),
					EmailPrimary:   util.PtrString("1email"),
					PhonePrimary:   util.PtrString("+22222211111122"),
					BillingAddress: util.PtrString("1address"),
				},
			},
			args: args{
				ctx: context.TODO(),
				landlords: []*offer.Landlord{
					{
						User:           "11",
						FirstName:      wrapperspb.String("first1"),
						LastName:       wrapperspb.String("last1"),
						EmailPrimary:   wrapperspb.String("email1"),
						PhonePrimary:   wrapperspb.String("+111111221"),
						BillingAddress: wrapperspb.String("address1"),
					},
				},
			},
			want: &OfferModel{
				Landlord: &LandlordModel{
					User:           "10",
					FirstName:      util.PtrString("1first"),
					LastName:       util.PtrString("1last"),
					EmailPrimary:   util.PtrString("1email"),
					PhonePrimary:   util.PtrString("+22222211111122"),
					BillingAddress: util.PtrString("1address"),
				},
				Landlords: []*LandlordModel{
					{
						User:           "11",
						FirstName:      util.PtrString("first1"),
						LastName:       util.PtrString("last1"),
						EmailPrimary:   util.PtrString("email1"),
						PhonePrimary:   util.PtrString("+111111221"),
						BillingAddress: util.PtrString("address1"),
					},
				},
			},
			wantErr: false,
		},
		{
			name: "error: landlord already exist by names",
			fields: fields{
				Landlord: &LandlordModel{
					User:           "11",
					FirstName:      util.PtrString("first11"),
					LastName:       util.PtrString("last11"),
					EmailPrimary:   util.PtrString("email11"),
					PhonePrimary:   util.PtrString("+1111112233"),
					BillingAddress: util.PtrString("address11"),
				},
				Landlords: []*LandlordModel{
					{
						User:           "11",
						FirstName:      util.PtrString("first"),
						LastName:       util.PtrString("last"),
						EmailPrimary:   util.PtrString("email"),
						PhonePrimary:   util.PtrString("+11111122"),
						BillingAddress: util.PtrString("address"),
					},
				},
			},
			args: args{
				ctx: context.TODO(),
				landlords: []*offer.Landlord{
					{
						User:           "11",
						FirstName:      wrapperspb.String("first11"),
						LastName:       wrapperspb.String("last11"),
						EmailPrimary:   wrapperspb.String("email1"),
						PhonePrimary:   wrapperspb.String("+111111221"),
						BillingAddress: wrapperspb.String("address1"),
					},
				},
			},
			want: &OfferModel{
				Landlord: &LandlordModel{
					User:           "11",
					FirstName:      util.PtrString("first11"),
					LastName:       util.PtrString("last11"),
					EmailPrimary:   util.PtrString("email11"),
					PhonePrimary:   util.PtrString("+1111112233"),
					BillingAddress: util.PtrString("address11"),
				},
				Landlords: []*LandlordModel{
					{
						User:           "11",
						FirstName:      util.PtrString("first"),
						LastName:       util.PtrString("last"),
						EmailPrimary:   util.PtrString("email"),
						PhonePrimary:   util.PtrString("+11111122"),
						BillingAddress: util.PtrString("address"),
					},
				},
			},
			wantErr: true,
		},
		{
			name: "error: landlord already exist by email",
			fields: fields{
				Landlord: &LandlordModel{
					User:           "11",
					FirstName:      util.PtrString("first11"),
					LastName:       util.PtrString("last11"),
					EmailPrimary:   util.PtrString("email11"),
					PhonePrimary:   util.PtrString("+1111112233"),
					BillingAddress: util.PtrString("address11"),
				},
				Landlords: []*LandlordModel{
					{
						User:           "11",
						FirstName:      util.PtrString("first"),
						LastName:       util.PtrString("last"),
						EmailPrimary:   util.PtrString("email"),
						PhonePrimary:   util.PtrString("+11111122"),
						BillingAddress: util.PtrString("address"),
					},
				},
			},
			args: args{
				ctx: context.TODO(),
				landlords: []*offer.Landlord{
					{
						User:           "11",
						FirstName:      wrapperspb.String("first1"),
						LastName:       wrapperspb.String("last1"),
						EmailPrimary:   wrapperspb.String("email11"),
						PhonePrimary:   wrapperspb.String("+111111221"),
						BillingAddress: wrapperspb.String("address1"),
					},
				},
			},
			want: &OfferModel{
				Landlord: &LandlordModel{
					User:           "11",
					FirstName:      util.PtrString("first11"),
					LastName:       util.PtrString("last11"),
					EmailPrimary:   util.PtrString("email11"),
					PhonePrimary:   util.PtrString("+1111112233"),
					BillingAddress: util.PtrString("address11"),
				},
				Landlords: []*LandlordModel{
					{
						User:           "11",
						FirstName:      util.PtrString("first"),
						LastName:       util.PtrString("last"),
						EmailPrimary:   util.PtrString("email"),
						PhonePrimary:   util.PtrString("+11111122"),
						BillingAddress: util.PtrString("address"),
					},
				},
			},
			wantErr: true,
		},
		{
			name: "error: landlord already exist by phone",
			fields: fields{
				Landlord: &LandlordModel{
					User:           "11",
					FirstName:      util.PtrString("first11"),
					LastName:       util.PtrString("last11"),
					EmailPrimary:   util.PtrString("email11"),
					PhonePrimary:   util.PtrString("+1111112233"),
					BillingAddress: util.PtrString("address11"),
				},
				Landlords: []*LandlordModel{
					{
						User:           "11",
						FirstName:      util.PtrString("first"),
						LastName:       util.PtrString("last"),
						EmailPrimary:   util.PtrString("email"),
						PhonePrimary:   util.PtrString("+11111122"),
						BillingAddress: util.PtrString("address"),
					},
				},
			},
			args: args{
				ctx: context.TODO(),
				landlords: []*offer.Landlord{
					{
						User:           "11",
						FirstName:      wrapperspb.String("first1"),
						LastName:       wrapperspb.String("last1"),
						EmailPrimary:   wrapperspb.String("email11"),
						PhonePrimary:   wrapperspb.String("+1111112233"),
						BillingAddress: wrapperspb.String("address1"),
					},
				},
			},
			want: &OfferModel{
				Landlord: &LandlordModel{
					User:           "11",
					FirstName:      util.PtrString("first11"),
					LastName:       util.PtrString("last11"),
					EmailPrimary:   util.PtrString("email11"),
					PhonePrimary:   util.PtrString("+1111112233"),
					BillingAddress: util.PtrString("address11"),
				},
				Landlords: []*LandlordModel{
					{
						User:           "11",
						FirstName:      util.PtrString("first"),
						LastName:       util.PtrString("last"),
						EmailPrimary:   util.PtrString("email"),
						PhonePrimary:   util.PtrString("+11111122"),
						BillingAddress: util.PtrString("address"),
					},
				},
			},
			wantErr: true,
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			m := &OfferModel{
				Landlord:  tt.fields.Landlord,
				Landlords: tt.fields.Landlords,
			}

			_ = m.updateOfferLandlords(context.Background(), tt.args.landlords)

			require.Equal(t, tt.want, m)
		})
	}
}
