package model

import (
	"context"
	"time"

	cardpayment "gitlab.mashroom.com/mashroom/backend/cardpayment/api"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	extras "gitlab.mashroom.com/mashroom/backend/extras/api"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	"gitlab.mashroom.com/mashroom/backend/offer/internal/connections"
	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
	"google.golang.org/protobuf/types/known/timestamppb"
)

// OfferService is data model of ...
type OfferService struct {
	ID            string               `bson:"id"`
	Code          string               `bson:"code"`
	Title         string               `bson:"title"`
	Description   string               `bson:"description"`
	Price         float64              `bson:"price"`
	TaxType       extras.TaxType       `bson:"tax_type"`
	TaxAmount     float64              `bson:"tax_amount"`
	TaxPercentage float64              `bson:"tax_percentage"`
	Type          extras.ServiceType   `bson:"type"`
	Status        extras.ServiceStatus `bson:"status"`
	Role          extras.ServiceRole   `bson:"role"`
	Primary       bool                 `bson:"primary"`
}

// OfferReceipt is data model of ...
type OfferReceipt struct {
	UpdatedAt             time.Time           `bson:"updated_at"`
	CreatedAt             time.Time           `bson:"created_at"`
	Kind                  string              `bson:"kind"`
	InvoiceMigrationError string              `bson:"invoice_migration_error"`
	Offer                 string              `bson:"offer"`
	ID                    string              `bson:"id"`
	User                  string              `bson:"user"`
	TransferID            string              `bson:"transfer_id"`
	CsrfToken             string              `bson:"csrf_token"`
	CardpaymentID         string              `bson:"cardpayment_id"`
	Services              []*OfferService     `bson:"services"`
	CheckTenants          []string            `bson:"check_tenants"`
	InvoiceNumber         uint64              `bson:"invoice_number"`
	Status                offer.PaymentStatus `bson:"status"`
	Type                  offer.PaymentType   `bson:"type"`
	InvoiceMigrated       bool                `bson:"invoice_migrated"`
}

func migrateReceiptsToInvoice(ctx context.Context, db *mongo.Database) error {
	o := db.Collection(MongoOffersCollection)

	c := db.Collection(MongoReceiptCollection)
	cur, err := c.Find(ctx, bson.M{"invoice_migrated": bson.M{"$ne": true}})
	if err != nil {
		return err
	}

	for cur.Next(ctx) {
		var rcpt OfferReceipt

		err = cur.Decode(&rcpt)
		if err != nil {
			log.WithError(err).Error("fetching receipt")
			continue
		}

		id, err := primitive.ObjectIDFromHex(rcpt.Offer)
		if err != nil {
			log.WithError(err).WithField("rcpt", rcpt).Error("decoding offer id")
			continue
		}

		var off OfferModel

		err = o.FindOne(ctx, bson.M{"_id": id}).Decode(&off)
		if err != nil {
			log.WithError(err).WithField("rcpt", rcpt).Error("fetching offer")

			rcpt.InvoiceMigrated = true
			rcpt.InvoiceMigrationError = err.Error()

			_, _ = c.ReplaceOne(ctx, bson.M{"id": rcpt.ID}, rcpt)

			continue
		}

		createdAt := timestamppb.New(rcpt.CreatedAt)
		updatedAt := timestamppb.New(rcpt.UpdatedAt)

		var services []*extras.ServiceSnapshotStatus

		var corrTarget string

		var amount float64

		for _, s := range rcpt.Services {
			services = append(services, &extras.ServiceSnapshotStatus{
				Service: &extras.ServiceSnapshot{
					Id:                "",
					ServiceId:         s.ID,
					Created:           nil,
					Updated:           nil,
					Code:              s.Code,
					Type:              s.Type,
					Role:              s.Role,
					Price:             s.Price,
					TaxType:           s.TaxType,
					TaxAmount:         s.TaxAmount,
					TaxPercentage:     s.TaxPercentage,
					Title:             s.Title,
					Description:       s.Description,
					Primary:           s.Primary,
					ReferenceProvider: "",
					ReferenceId:       "",
					TargetType:        "",
					TargetId:          "",
				},
				UpdatedAt: timestamppb.Now(),
				Admin:     "",
				Status:    s.Status,
			})

			switch {
			case s.Code == extras.ProductLandlordReferencingCheck.String():
				corrTarget = invoice.CorrelationTargetReferenceCheck
			case s.Code == extras.ProductTenantHoldingDeposit.String():
				corrTarget = invoice.CorrelationTargetHoldingDeposit
			case off.Landlord != nil && rcpt.User == off.Landlord.User:
				corrTarget = invoice.CorrelationTargetLandlordServices
			default:
				corrTarget = invoice.CorrelationTargetTenantServices
			}

			amount += s.Price + s.TaxAmount
		}

		var status invoice.InvoiceStatus

		var paystatus invoice.PaymentStatus

		switch rcpt.Status {
		case offer.PaymentStatus_PAYMENT_SUCCESS:
			status = invoice.InvoiceStatus_INVOICE_FULFILLED
			paystatus = invoice.PaymentStatus_PAYMENT_SUCCESS
		case offer.PaymentStatus_PAYMENT_PENDING:
			status = invoice.InvoiceStatus_INVOICE_OPEN
			paystatus = invoice.PaymentStatus_PAYMENT_PENDING
		case offer.PaymentStatus_PAYMENT_FAILED:
			status = invoice.InvoiceStatus_INVOICE_CANCELED
			paystatus = invoice.PaymentStatus_PAYMENT_CANCELED
		case offer.PaymentStatus_PAYMENT_CONFIRM:
			status = invoice.InvoiceStatus_INVOICE_OPEN
			paystatus = invoice.PaymentStatus_PAYMENT_CONFIRM
		}

		var payments []*invoice.Payment

		switch {
		case rcpt.CardpaymentID != "":
			payments = append(payments, &invoice.Payment{
				PaymentId: "",
				Reference: "",
				InvoiceId: "",
				User:      rcpt.User,
				Amount:    amount,
				Status:    paystatus,
				System: &invoice.PaymentSystemData{
					SystemReference: rcpt.CardpaymentID,
					Data: &invoice.PaymentSystemData_Stripe{
						Stripe: &invoice.PaymentSystemStripe{
							Request: &invoice.PaymentSystemStripeRequest{
								Type: cardpayment.PaymentType(rcpt.Type),
							},
							Response: &invoice.PaymentSystemStripeResponse{
								StripeId:     rcpt.CardpaymentID,
								ClientSecret: "",
							},
						},
					},
				},
				CreatedAt: createdAt,
				UpdatedAt: updatedAt,
			})
		case rcpt.CsrfToken != "":
			payments = append(payments, &invoice.Payment{
				PaymentId: "",
				Reference: "",
				InvoiceId: "",
				User:      rcpt.User,
				Amount:    amount,
				Status:    paystatus,
				System: &invoice.PaymentSystemData{
					SystemReference: rcpt.CsrfToken,
					Data: &invoice.PaymentSystemData_Tokenio{
						Tokenio: &invoice.PaymentSystemTokenio{
							Request: &invoice.PaymentSystemTokenioRequest{
								Description: "",
								Ivr:         off.PropertyIvr,
								Purpose:     0,
							},
							Response: &invoice.PaymentSystemTokenioResponse{
								Url:  "",
								Csrf: rcpt.CsrfToken,
							},
						},
					},
				},
				CreatedAt: createdAt,
				UpdatedAt: updatedAt,
			})
		}

		inv := &invoice.Invoice{
			Property:          off.Property,
			Ivr:               off.PropertyIvr,
			User:              rcpt.User,
			CorrelationType:   invoice.CorrelationTypeOffer,
			CorrelationId:     rcpt.Offer,
			CorrelationTarget: corrTarget,
			Reference:         "",
			Services:          services,
			Amount:            amount,
			Status:            status,
			Render:            "",
			CreatedAt:         createdAt,
			UpdatedAt:         updatedAt,
			Payments:          payments,
			Properties:        nil,
		}

		_, err = connections.InvoiceService.ServiceInvoiceImport(ctx, &invoice.ServiceInvoiceImportRequest{
			Invoice: []*invoice.Invoice{
				inv,
			},
			NotifyCallback: connections.SelfRemote,
		})
		if err != nil {
			log.WithError(err).WithField("inv", inv).Error("importing invoice")
			continue
		}

		rcpt.InvoiceMigrated = true

		_, err = c.ReplaceOne(ctx, bson.M{"id": rcpt.ID}, rcpt)
		if err != nil {
			log.WithError(err).WithField("inv", inv).Error("updating receipt")
		}
	}

	return err
}
