package test

import (
	"context"
	"os"
	"strconv"
	"testing"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/lib/mongolocal"
	api "gitlab.mashroom.com/mashroom/backend/offer/api/rto"
	domain "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto"
	repositorymongo "gitlab.mashroom.com/mashroom/backend/offer/internal/app/domain/rto/repository/mongo"

	"github.com/joho/godotenv"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

type TestConfigPath string

const (
	TestConfigPathForRtoServices           TestConfigPath = "../../../../"
	TestConfigPathForRtoServicesDaemon     TestConfigPath = "../../../../../"
	TestConfigPathForRtoUseCases           TestConfigPath = "../../../../"
	TestConfigPathForRtoDomainRepositories TestConfigPath = "../../../../../../"
)

var (
	Ctx               context.Context
	Repository        domain.Repository
	HistoryRepository domain.HistoryRepository
	KorshilerDogs     = "korshiler@666.suki"
	Chmolgani         = 1705.530666
	ZhunusoviDogs     = "zhunusovi@666.com"
	RentAmount        = 17071987.666
)

func InitRtoTests(testConfigPath TestConfigPath) (context.Context, domain.Repository, domain.HistoryRepository) {
	Ctx = context.Background()

	err := godotenv.Load(string(testConfigPath) + "configs/.env.testing")
	if err != nil {
		panic(err)
	}

	if CanRunTests() {
		mongoTimeout, err := strconv.Atoi(os.Getenv("MONGO_TIMEOUT"))
		if err != nil {
			panic(err)
		}
		mongoDB, err := mongolocal.GetDB(&mongolocal.Config{
			URL:      os.Getenv("MONGO_URL"),
			Username: os.Getenv("MONGO_USERNAME"),
			Password: os.Getenv("MONGO_PASSWORD"),
			Database: os.Getenv("MONGO_DATABASE"),
			Timeout:  int64(mongoTimeout),
		})
		if err != nil {
			panic(err)
		}

		Repository = repositorymongo.NewRepository(mongoDB)
		HistoryRepository = repositorymongo.NewHistoryRepository(mongoDB)

		return Ctx, Repository, HistoryRepository
	}

	return nil, nil, nil
}

func CreateRto(ctx context.Context, t *testing.T, repository domain.Repository,
	state api.RtoState, ittinBalasi string, chmolgani float64) *domain.Rto {
	landlordID := "test landlord id"
	rtoID := primitive.NewObjectID()
	now := time.Now()

	var landlord *domain.Landlord
	if ittinBalasi != "" {
		landlord = &domain.Landlord{
			User:         landlordID,
			EmailPrimary: ittinBalasi,
		}
	}

	rtoModel := &domain.Rto{
		ID:        rtoID,
		State:     state,
		CreatedAt: now,
		UpdatedAt: now,
		Landlord:  landlord,
		Conditions: &domain.Conditions{
			MoveInDate: now,
			Price:      chmolgani,
		},
		LeadTenant: &domain.Tenant{
			EmailPrimary: ZhunusoviDogs,
			User:         primitive.NewObjectID().String(),
		},
		Property:        primitive.NewObjectID().String(),
		PropertyIvr:     primitive.NewObjectID().String(),
		PropertyListing: primitive.NewObjectID().String(),
	}
	err := repository.Create(ctx, rtoModel)
	if err != nil {
		t.Error(err)
	}

	return rtoModel
}

func DeleteRto(ctx context.Context, t *testing.T, repository domain.Repository,
	historyRepository domain.HistoryRepository, rtoID primitive.ObjectID, rtoHistoryCount int64) {
	rtoDelete, err := repository.DeleteByID(ctx, rtoID)
	if err != nil {
		t.Error(err)
	}
	if rtoDelete != 1 {
		t.Error(err)
	}

	if rtoHistoryCount > 0 {
		rtoHistoryDeleteByRtoID, err := historyRepository.DeleteByRtoID(ctx, rtoID)
		if err != nil {
			t.Error(err)
		}
		if rtoHistoryCount != rtoHistoryDeleteByRtoID {
			t.Error(err)
		}
	}
}

func CanRunTests() bool {
	return os.Getenv("CAN_RUN_TESTS") != ""
}
